/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QAction>
#include <QDockWidget>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QPlainTextEdit>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QTabWidget>
#include <QVBoxLayout>
#include <QWidget>
#include <QPushButton>
#include <QToolBar>
#include <QDragEnterEvent>
#include <QDropEvent>
#include <QMimeData>
#include <QDateTime>
#include <QFileDialog>
#include <QSettings>
#include <QCloseEvent>
#include <QMenuBar>
#include <QMenu>
#include <QFileInfo>
#include <QTimer>
#include <QUrl>
#include <QFont>
#include <QDir>
#include <QRect>
#include <QPoint>

#include "../../core/messagebox.h"
#include "../../core/xmlfile.h"

#include "tolerancewidget.h"
#include "opengldisplaywidget.h"
#include "reportwidget.h"
#include "analysiswidget.h"
#include "pathwidget.h"
#include "selectiondialog.h"
#include "program_dmis.h"

#include "mcg.h"

static const int						RECENT_FILE_LIST_SIZE(20);
static const double						ZERO_EPSILON(0.001);

TMachineCheckingGauge::TMachineCheckingGauge(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QAction								*recent_file_action;
	QWidget								*central_widget;
	QToolBar							*toolbar;
	QGridLayout							*widget_layout;
	QHBoxLayout							*copyright_hlayout;
	QLabel								*logo_label;
	QLabel								*copyright_label;
	QDockWidget							*path_dock;
	QDockWidget							*tolerance_dock;
	QSpacerItem							*copyright_hspacer;
	QMenuBar							*main_menu;
	QMenu								*file_menu;
	QMenu								*open_recent_menu;
	QMenu								*view_menu;
	QMenu								*help_menu;
	int									cntr;
	QDir								home;
	QList<int>							splitter_sizes;
	QFont								copyright_font;

	copyright_font.setFamily(QStringLiteral("Verdana"));
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);

	home = QDir::home();

	if(!home.exists(".mcg"))
	{
		home.mkdir(".mcg");
	}
	
	home.cd(".mcg");
	
	d_data_path = home.absolutePath();
	
	d_settings = new QSettings(d_data_path + "/Settings.ini",QSettings::IniFormat,this);

	this->resize(1201,801);
	this->setAcceptDrops(true);
	
	toolbar = new QToolBar(this);
	toolbar->setIconSize(QSize(32,32));
	this->addToolBar(Qt::TopToolBarArea,toolbar);

	d_new_action = new QAction(this);
	d_new_action->setIcon(QIcon(":/toolbar/new_32x32.png"));
	
	d_open_action = new QAction(this);
	d_open_action->setIcon(QIcon(":/toolbar/open_32x32.png"));
	
	d_save_action = new QAction(this);
	d_save_action->setIcon(QIcon(":/toolbar/save_32x32.png"));
	
	d_export_action = new QAction(this);
	d_export_action->setIcon(QIcon(":/toolbar/export_32x32.png"));
	
	d_import_action = new QAction(this);
	d_import_action->setIcon(QIcon(":/toolbar/import_32x32.png"));
	
	d_quit_action = new QAction(this);
	
	d_report_zoomin_action = new QAction(this);
	d_report_zoomin_action->setIcon(QIcon(":/toolbar/zoom_in_32x32.png"));
	
	d_report_zoomout_action = new QAction(this);
	d_report_zoomout_action->setIcon(QIcon(":/toolbar/zoom_out_32x32.png"));
	
	d_report_output_action = new QAction(this);
	d_report_output_action->setIcon(QIcon(":/toolbar/report_32x32.png"));
	
	d_report_print_action = new QAction(this);
	d_report_print_action->setIcon(QIcon(":/toolbar/print_32x32.png"));

	d_about_action = new QAction(this);
	d_about_action->setIcon(QIcon(":/toolbar/about_32x32.png"));

	toolbar->addAction(d_new_action);
	toolbar->addAction(d_open_action);
	toolbar->addAction(d_save_action);
	toolbar->addSeparator();
	toolbar->addAction(d_export_action);
	toolbar->addAction(d_import_action);
	toolbar->addSeparator();
	toolbar->addAction(d_report_zoomin_action);
	toolbar->addAction(d_report_zoomout_action);
	toolbar->addAction(d_report_output_action);
	toolbar->addAction(d_report_print_action);
	toolbar->addSeparator();
	toolbar->addAction(d_about_action);

	main_menu = new QMenuBar(this);
	this->setMenuBar(main_menu);
	
	file_menu = new QMenu(main_menu);
	main_menu->addMenu(file_menu);
	
	open_recent_menu = new QMenu(main_menu);
	
	view_menu = new QMenu(main_menu);
	main_menu->addMenu(view_menu);
	
	help_menu = new QMenu(main_menu);
	main_menu->addMenu(help_menu);

	main_menu->addAction(file_menu->menuAction());
	main_menu->addAction(view_menu->menuAction());
	main_menu->addAction(help_menu->menuAction());
	
	file_menu->addAction(d_new_action);
	file_menu->addAction(d_open_action);
	file_menu->addMenu(open_recent_menu);
	file_menu->addAction(d_save_action);
	file_menu->addSeparator();
	file_menu->addAction(d_report_output_action);
	file_menu->addAction(d_report_print_action);
	file_menu->addSeparator();
	file_menu->addAction(d_export_action);
	file_menu->addAction(d_import_action);
	file_menu->addSeparator();
	file_menu->addAction(d_quit_action);
	
	help_menu->addAction(d_about_action);
	
	for(cntr = 0;cntr < RECENT_FILE_LIST_SIZE;++cntr)
	{
		recent_file_action = new QAction(this);
		recent_file_action->setVisible(false);
		
		connect(recent_file_action, &QAction::triggered,this,&TMachineCheckingGauge::Open_Recent);
		
		open_recent_menu->addAction(recent_file_action);
		d_recent_file_actions.push_back(recent_file_action);
	}
	
	path_dock = new QDockWidget(this);
	path_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::LeftDockWidgetArea,path_dock);

	view_menu->addAction(path_dock->toggleViewAction());

	d_path_widget = new TPathWidget();
	path_dock->setWidget(d_path_widget);

	
	tolerance_dock = new QDockWidget(this);
	tolerance_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::LeftDockWidgetArea,tolerance_dock);
		
	view_menu->addAction(tolerance_dock->toggleViewAction());

	d_tolerance_widget = new TToleranceWidget();
	tolerance_dock->setWidget(d_tolerance_widget);
	
	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);

	widget_layout = new QGridLayout(central_widget);
	
	copyright_hlayout = new QHBoxLayout();
	
	copyright_label = new QLabel(central_widget);
	copyright_label->setFont(copyright_font);
	copyright_label->setAlignment(Qt::AlignBottom|Qt::AlignLeading|Qt::AlignLeft);
	copyright_hlayout->addWidget(copyright_label);
	
	copyright_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	copyright_hlayout->addItem(copyright_hspacer);
	
	logo_label = new QLabel(central_widget);
	logo_label->setMinimumSize(100,40);
	logo_label->setMaximumSize(100,40);
	copyright_hlayout->addWidget(logo_label);
	
	widget_layout->addLayout(copyright_hlayout,1,0,1,1);

	d_tab_widget = new QTabWidget(central_widget);
	
	d_opengl_display_widget = new TOpenGLDisplayWidget();
	d_tab_widget->addTab(d_opengl_display_widget,QStringLiteral("Model"));
	
	d_analysis_widget = new TAnalysisWidget();
	d_tab_widget->addTab(d_analysis_widget,QStringLiteral("Analysis"));

	d_report_widget = new TReportWidget();
	d_tab_widget->addTab(d_report_widget,QStringLiteral("Report"));
	
	widget_layout->addWidget(d_tab_widget,0,0,1,1);
	
	d_msg_box = new TMessageBox(this);
	d_selection_dialog = new TSelectionDialog(this,Qt::WindowFlags(Qt::Tool));
	d_simulation_timer = new QTimer(this);

	// settings
	d_path_widget->Set_Horizontal_Arm_Checked(d_settings->value("Path_Horizontal_Arm",false).toBool());
	d_path_widget->Set_Arm_Length_Selection(static_cast<TPathWidget::TArmLength>(d_settings->value("Selected_MCG_Arm",2).toInt()));

	d_path_widget->Set_Elevation_Start(d_settings->value("Path_Elevation_Start",-40.0).toDouble());
	d_path_widget->Set_Elevation_End(d_settings->value("Path_Elevation_End",40.0).toDouble());

	d_path_widget->Set_Azimuth_Count(d_settings->value("Path_Azimuth_Count",8).toInt());
	d_path_widget->Set_Elevation_Count(d_settings->value("Path_Elevation_Count",3).toInt());
	d_path_widget->Set_Pattern_Count(d_settings->value("Path_Pattern_Count",2).toInt());
	d_path_widget->Set_Point_Count(d_settings->value("Path_Point_Count",2).toInt());
	
	d_path_widget->Set_Arm_Axis(d_settings->value("Path_Arm_Axis",1).toInt());
	d_file_path = d_settings->value("File_Path",QString()).toString();
	
	this->Load_Recent_File_List(d_data_path);
	
	this->Update_Recent_File_Actions();

	d_tolerance_widget->Set_Tolerance_Item_A(d_settings->value("Tolerance_A","0.012").toDouble());
	d_tolerance_widget->Set_Tolerance_Item_B(d_settings->value("Tolerance_B","0.006").toDouble());
	d_tolerance_widget->Set_Tolerance_Item_C(d_settings->value("Tolerance_C","1000").toDouble());
	d_tolerance_widget->Set_Tolerance_Item_D(d_settings->value("Tolerance_D","0").toDouble());
	
	d_tolerance_widget->Update();
	
	logo_label->setPixmap(QPixmap(QStringLiteral(":/images/logo.png")));

	this->setWindowTitle(QStringLiteral("Machine Checking Gauge"));
	this->setWindowIcon(QIcon(":/gui/mcg.png"));
	
	path_dock->setWindowTitle(QStringLiteral("Path Generator"));
	tolerance_dock->setWindowTitle(QStringLiteral("Tolerance"));
	
	file_menu->setTitle(QStringLiteral("File"));
	open_recent_menu->setTitle(QStringLiteral("Open Recent..."));
	view_menu->setTitle(QStringLiteral("View"));
	help_menu->setTitle(QStringLiteral("Help"));

	d_new_action->setText(QStringLiteral("New"));
	d_open_action->setText(QStringLiteral("Open"));
	d_save_action->setText(QStringLiteral("Save"));
	d_import_action->setText(QStringLiteral("Import"));
	d_export_action->setText(QStringLiteral("Export"));
	d_quit_action->setText(QStringLiteral("Quit"));
	d_report_zoomin_action->setText(QStringLiteral("Zoom In Report"));
	d_report_zoomout_action->setText(QStringLiteral("Zoom Out Report"));
	d_report_output_action->setText(QStringLiteral("Create Report"));
	d_report_print_action->setText(QStringLiteral("Print Report"));
	d_about_action->setText(QStringLiteral("About"));

	d_new_action->setToolTip(QStringLiteral("New"));
	d_open_action->setToolTip(QStringLiteral("Open"));
	d_save_action->setToolTip(QStringLiteral("Save"));
	d_import_action->setToolTip(QStringLiteral("Import Measurement"));
	d_export_action->setToolTip(QStringLiteral("Export Measurement"));
	d_report_zoomin_action->setToolTip(QStringLiteral("Zoom In Report"));
	d_report_zoomout_action->setToolTip(QStringLiteral("Zoom Out Report"));
	d_report_output_action->setToolTip(QStringLiteral("Create Report"));
	d_report_print_action->setToolTip(QStringLiteral("Print Report"));
	d_about_action->setToolTip(QStringLiteral("About"));
	copyright_label->setText(QStringLiteral("Copyright (C) 2023 Select Calibration Incorporated.  www.selectcalibration.ca"));

	
	// defaults
	d_data_changed = false;
	d_position_selection_dialog = true;
	d_report_zoomin_action->setEnabled(false);
	d_report_zoomout_action->setEnabled(false);
	d_export_action->setEnabled(false);
	d_new_action->setEnabled(false);
	d_save_action->setEnabled(false);
	d_report_print_action->setEnabled(false);
	d_report_output_action->setEnabled(false);

	connect(d_new_action,&QAction::triggered,this,&TMachineCheckingGauge::New);
	connect(d_open_action,&QAction::triggered,this,&TMachineCheckingGauge::Open);
	connect(d_save_action,&QAction::triggered,this,&TMachineCheckingGauge::Save);
	connect(d_import_action,&QAction::triggered,this,&TMachineCheckingGauge::Import);
	connect(d_export_action,&QAction::triggered,this,&TMachineCheckingGauge::Export);
	connect(d_quit_action,&QAction::triggered,this,&TMachineCheckingGauge::close);
	connect(d_about_action,&QAction::triggered,this,&TMachineCheckingGauge::Help_About);
	connect(d_report_zoomin_action,&QAction::triggered,d_report_widget,&TReportWidget::Zoom_In);
	connect(d_report_zoomout_action,&QAction::triggered,d_report_widget,&TReportWidget::Zoom_Out);
	connect(d_report_print_action,&QAction::triggered,d_report_widget,&TReportWidget::Print);
	connect(d_report_output_action,&QAction::triggered,this,&TMachineCheckingGauge::Output);

	connect(d_path_widget,&TPathWidget::Update_Path,this,&TMachineCheckingGauge::Update_Path);
	connect(d_tab_widget,&QTabWidget::currentChanged,this,&TMachineCheckingGauge::Tab_Index_Changed);
	connect(d_opengl_display_widget,&TOpenGLDisplayWidget::Model_Item_Clicked,this,&TMachineCheckingGauge::MCG_Selection);
	connect(d_tolerance_widget,&TToleranceWidget::Update_Tolerance,this,&TMachineCheckingGauge::Tolerance_Changed);
	
	connect(d_path_widget,&TPathWidget::Simulate_Position,this,&TMachineCheckingGauge::Draw_MCG_Item);
	connect(d_simulation_timer,&QTimer::timeout,this,&TMachineCheckingGauge::Simulation_Timeout);
	connect(d_path_widget,&TPathWidget::Start_Simulation,this,&TMachineCheckingGauge::Simulation_Start);
	connect(d_path_widget,&TPathWidget::Stop_Simulation,this,&TMachineCheckingGauge::Simulation_Stop);
	connect(d_path_widget,&TPathWidget::Generate_DMIS_Program,this,&TMachineCheckingGauge::Generate_DMIS_Program);
	connect(d_path_widget,&TPathWidget::Arm_Length_Changed,this,&TMachineCheckingGauge::Arm_Length_Changed);

	this->Update_Path();
	
	d_opengl_display_widget->Update(true);
}

TMachineCheckingGauge::~TMachineCheckingGauge(void)
{
}

void TMachineCheckingGauge::New(void)
{
	if(d_data_changed)
	{
		d_msg_box->setText("Unsaved Data");
		d_msg_box->setInformativeText("The current data has not been saved.  Do you want to continue?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setIcon(QMessageBox::Warning);
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		
		if(QMessageBox::No == d_msg_box->exec())
		{
			return;
		}
	}

	d_path_widget->setEnabled(true);
	
	d_measurement.Clear();
	
	this->Update_Display();
	this->Update_Path();
	d_data_changed = false;
	
	d_opengl_display_widget->Update(true);
	
	d_export_action->setEnabled(false);
	d_new_action->setEnabled(false);
	d_save_action->setEnabled(false);
	d_report_print_action->setEnabled(false);
	d_report_output_action->setEnabled(false);
	
	d_active_file_name.clear();
}

void TMachineCheckingGauge::Open(void)
{
	QString								file_name;
	
	if(d_data_changed)
	{
		d_msg_box->setText("Unsaved Data");
		d_msg_box->setInformativeText("The current data has not been saved.  Do you want to continue?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setIcon(QMessageBox::Warning);
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		
		if(QMessageBox::No == d_msg_box->exec())
		{
			return;
		}
	}

	file_name = QFileDialog::getOpenFileName(
					 this,
					 QStringLiteral("Select the MCG data file to open"),
					 d_file_path,
					 QStringLiteral("MCG Files (*.mcg)"));
	
	if(file_name.length())
	{
		this->Open_File(file_name);
		
		d_recent_files.push_front(file_name);
		d_recent_files.removeDuplicates();
		
		this->Update_Recent_File_Actions();
	}
}

void TMachineCheckingGauge::Open_Recent(void)
{
	QString								file_name;
	QAction								*recent_file_action;
	
	if(d_data_changed)
	{
		d_msg_box->setText("Unsaved Data");
		d_msg_box->setInformativeText("The current data has not been saved.  Do you want to continue?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setIcon(QMessageBox::Warning);
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		
		if(QMessageBox::No == d_msg_box->exec())
		{
			return;
		}
	}
 
	recent_file_action = qobject_cast<QAction*>(sender());
	
	if(recent_file_action)
	{
		file_name = recent_file_action->data().toString();		
		this->Open_File(file_name);
		
		d_recent_files.push_front(file_name);
		d_recent_files.removeDuplicates();
		
		this->Update_Recent_File_Actions();
	}
}

void TMachineCheckingGauge::Save(void)
{
	QString								file_name;
	
	file_name = QFileDialog::getSaveFileName(
					 this,
					 QStringLiteral("Select the MCG data file to create"),
					 d_active_file_name,
					 QStringLiteral("MCG Files (*.mcg)"));
	
	if(file_name.length())
	{
		this->Save_File(file_name);
		
		d_recent_files.push_front(file_name);
		d_recent_files.removeDuplicates();
		
		this->Update_Recent_File_Actions();
	}
}

void TMachineCheckingGauge::Import(void)
{
	QString								file_name;
	TMachineCheckingGauge::TFileType	file_type;
	
	if(d_data_changed)
	{
		d_msg_box->setText("Unsaved Data");
		d_msg_box->setInformativeText("The current data has not been saved.  Do you want to continue?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setIcon(QMessageBox::Warning);
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		
		if(QMessageBox::No == d_msg_box->exec())
		{
			return;
		}
	}
	
	file_name = QFileDialog::getOpenFileName(
						 this,
						 QStringLiteral("Select the MCG measurement file to open"),
						 d_file_path,
						 QStringLiteral("MCG Measurement (*.xml *.txt *.dmo)"));
	
	if(file_name.length())
	{
		file_type = this->Identify_File_Type(file_name);
		
		switch(file_type)
		{
			case TMachineCheckingGauge::FILE_TYPE_MCG:	// shouldn't get here but just in case
				this->Open_File(file_name);
				
				d_recent_files.push_front(file_name);
				d_recent_files.removeDuplicates();
				
				this->Update_Recent_File_Actions();
				break;
				
			case TMachineCheckingGauge::FILE_TYPE_XML:
				this->Import_XML_Measurement(file_name);
				break;
				
			case TMachineCheckingGauge::FILE_TYPE_TXT:
				this->Import_Text_Measurement(file_name);
				break;
				
			case TMachineCheckingGauge::FILE_TYPE_DMO:
				this->Import_DMO_Measurement(file_name);
				break;
				
			default:
				d_msg_box->setText("Unrecognized Format");
				d_msg_box->setInformativeText("The format of the imported data is not recognized.");
				d_msg_box->setDetailedText(QString());
				d_msg_box->setIcon(QMessageBox::Critical);
				d_msg_box->setStandardButtons(QMessageBox::Ok);
				d_msg_box->setDefaultButton(QMessageBox::Ok);
				
				d_msg_box->exec();
				break;
		}
	}
}

void TMachineCheckingGauge::Export(void)
{
	QString								file_name;	
	
	file_name = QFileDialog::getSaveFileName(
						 this,
						 QStringLiteral("Select the MCG measurement file to create"),
						 d_file_path,
						 QStringLiteral("CSV Files (*.csv)"));
	
	if(file_name.length())
	{
		this->Export_Measurement(file_name);
	}
}

void TMachineCheckingGauge::Output(void)
{
	int									index;
	QString								file_name;
	
	file_name = QFileDialog::getSaveFileName(this,
						QStringLiteral("Enter the name of the PDF file to create."),
						d_file_path,
						QStringLiteral("PDF Files (*.pdf)"));

	if(file_name.length())
	{
		d_report_widget->Write(file_name);
		
		d_file_path = file_name;
		
		index = d_file_path.lastIndexOf('/');
		if(!(index < 0))
		{
			d_file_path.truncate(index);
		}
	}
}

void TMachineCheckingGauge::Help_About(void)
{
	d_msg_box->setText(QStringLiteral("Machine Checking Gauge"));
	d_msg_box->setInformativeText(QStringLiteral("Machine Checking Gauge - Version 4.0"));
	d_msg_box->setDetailedText(QString());
	d_msg_box->setStandardButtons(QMessageBox::Ok);
	d_msg_box->setDefaultButton(QMessageBox::Ok);
	d_msg_box->setIcon(QMessageBox::Information);
	
	d_msg_box->exec();
}

void TMachineCheckingGauge::Tab_Index_Changed(
	int									index)
{
	if(index == 2)	// report
	{
		d_report_zoomin_action->setEnabled(true);
		d_report_zoomout_action->setEnabled(true);
	}
	else
	{
		d_report_zoomin_action->setEnabled(false);
		d_report_zoomout_action->setEnabled(false);
	}
	
	if(index != 0)	// model
	{
		if(d_selection_dialog->isVisible())
		{
			d_selection_dialog->setVisible(false);
			
			d_opengl_display_widget->Clear_Selection();
		}
	}
}

void TMachineCheckingGauge::Update_Path(void)
{	
	double								arm_length;
	
	d_path_widget->End_Simulation();
	
	if(d_measurement.Measurement_Count() == 0)
	{
		arm_length = d_path_widget->Arm_Length();
		
		this->Arm_Length_Changed(arm_length);
	}
	else
	{
		d_opengl_display_widget->Clear_MCG_Path();
		
		d_motion_path.clear();
		d_simulation_index = -1;
	}
}

void TMachineCheckingGauge::Tolerance_Changed(void)
{
	this->Update_Display();
	this->Update_Path();
	
	if(d_selection_dialog->isVisible())
	{
		d_selection_dialog->setVisible(false);
	}
}

void TMachineCheckingGauge::MCG_Selection(
	const double						&x,
	const double						&y,
	const double						&z,
	const int							id)
{
	Q_UNUSED(x);
	Q_UNUSED(y);
	Q_UNUSED(z);
	
	QRect								rect;
	QPoint								pos;
	double								average;
	double								median;
	int									count;
	QString								text;
	TMeasurement::TMeasurementItem		meas;
	static const int					SELECTION_DIALOG_OFFSET(50);
	
	if(!(id < 0))
	{
		count = d_measurement.Measurement_Count();
		
		if(id < count)
		{
			meas = d_measurement.MCG_Measurement(id);
			average = d_measurement.Length_Average();
			median = d_measurement.Length_Median();
			
			d_selection_dialog->Clear_Text();
			
			d_selection_dialog->Add_Text(QString("Azimuth:         %1").arg(meas.azimuth,0,'f',1));
			d_selection_dialog->Add_Text(QString("Elevation:       %1").arg(meas.elevation,0,'f',1));
			d_selection_dialog->Add_Text(QString("Arm Length:      %1").arg(meas.length,0,'f',4));
			d_selection_dialog->Add_Text(QString("Arm Length Min:  %1").arg(meas.length_min,0,'f',4));
			d_selection_dialog->Add_Text(QString("Arm Length Max:  %1").arg(meas.length_max,0,'f',4));
			d_selection_dialog->Add_Text(QString("Arm Length SDev: %1").arg(meas.length_stddev,0,'f',4));
			d_selection_dialog->Add_Text(QString("Average Length:  %1").arg(average,0,'f',4));
			d_selection_dialog->Add_Text(QString("Median Length:   %1").arg(median,0,'f',4));
			d_selection_dialog->Add_Text(QString("Deviation:       %1").arg(meas.length - average,0,'f',4));
			d_selection_dialog->Add_Text_End();
			
			if(d_position_selection_dialog)
			{
				rect = this->geometry();
				
				pos = rect.topLeft();
				
				pos.setX(pos.x() + SELECTION_DIALOG_OFFSET);
				pos.setY(pos.y() + SELECTION_DIALOG_OFFSET);
				
				d_selection_dialog->move(pos);

				d_position_selection_dialog = false;
			}
			
			d_selection_dialog->show();
			
			d_opengl_display_widget->Clear_Selection();
			d_opengl_display_widget->Set_Selection(id);
		}
	}
}

void TMachineCheckingGauge::Simulation_Start(void)
{
	TVector3							position;
		
	position.Set(0,0,0);

	if(d_motion_path.size())
	{
		d_simulation_timer->start(50);
	}
}

void TMachineCheckingGauge::Simulation_Stop(void)
{
	d_simulation_timer->stop();
}

void TMachineCheckingGauge::Simulation_Timeout(void)
{
	TVector3							pnt;
	TVector3							center;
	TVector3							axis;
	double								arm_length;

	center.Set(0,0,0);
		
	if(d_motion_path.size())
	{
		if(!(static_cast<unsigned int>(d_simulation_index) < d_motion_path.size()))
		{
			d_simulation_index = 0;
		}
		
		pnt = d_motion_path[d_simulation_index];
	
		axis = pnt - center;
		
		if(axis.Length() > ZERO_EPSILON)
		{
			axis.Normal();
			
			arm_length = d_path_widget->Arm_Length();
					
			d_opengl_display_widget->Set_MCG(-1,center,axis,arm_length);
			d_opengl_display_widget->Update(false);
		}
		
		d_path_widget->Set_Simulator_Slider_Position(d_simulation_index);

		++d_simulation_index;		
	}
	else
	{
		d_simulation_timer->stop();
	}
}

void TMachineCheckingGauge::Generate_DMIS_Program(void)
{
	std::vector<TPathWidget::TPathPoint> path_points;
	TProgramDMIS						program_dmis;
	QString								file_name;
	double								arm_length;
	int									index;
	
	arm_length = d_path_widget->Arm_Length();
	
	file_name = QFileDialog::getSaveFileName(
						 this,
						 QStringLiteral("Enter the name of the DMIS program to create."),
						 d_file_path,
						 QStringLiteral("DMIS Files (*.DMS *.DMI)"));
	
	if(file_name.length())
	{
		path_points = d_path_widget->Path_Points();
		
		if(path_points.size())
		{
			program_dmis.Create_Program(file_name,path_points,arm_length);
			
			d_file_path = file_name;
			
			index = d_file_path.lastIndexOf('/');
			if(!(index < 0))
			{
				d_file_path.truncate(index);
			}
		}
	}
}

void TMachineCheckingGauge::Draw_MCG_Item(
	const int							index)
{	
	TVector3							pnt;
	TVector3							center;
	TVector3							axis;
	double								arm_length;

	center.Set(0,0,0);
	
	if(!(static_cast<unsigned int>(index) > d_motion_path.size()))
	{
		pnt = d_motion_path[index];
		
		axis = pnt - center;
		
		if(axis.Length() > ZERO_EPSILON)
		{
			axis.Normal();
			
			arm_length = d_path_widget->Arm_Length();
			
			d_opengl_display_widget->Set_MCG(-1,center,axis,arm_length);
			d_opengl_display_widget->Update(false);
		}		
	}
}

void TMachineCheckingGauge::Arm_Length_Changed(
	const double						&arm_length)
{
	TVector3							position;
	TVector3							center;
	TVector3							axis;
	TVector3							pnt;
	
	position.Set(0,0,0);
	
	if(d_measurement.Measurement_Count() == 0)
	{
		d_opengl_display_widget->Define_Machine_Volume(TVector3(-arm_length,-arm_length,-arm_length),TVector3(arm_length,arm_length,arm_length));
		
		d_motion_path = d_path_widget->Motion_Path(position,arm_length);
		d_opengl_display_widget->Set_MCG_Path(d_motion_path);
		
		if(d_motion_path.size())
		{
			d_simulation_index = 0;
			
			pnt = d_motion_path[d_simulation_index];
			
			axis = pnt - center;
			
			if(axis.Length() > ZERO_EPSILON)
			{
				axis.Normal();
				d_opengl_display_widget->Set_MCG(-1,center,axis,arm_length);
			}
		}
		else
		{
			d_simulation_index = -1;
		}
		
		d_opengl_display_widget->Update(false);
	}
}

void TMachineCheckingGauge::dragEnterEvent(
	QDragEnterEvent 					*drag_enter_event)
{
	if(drag_enter_event->mimeData()->hasFormat("text/uri-list"))
	{
		drag_enter_event->acceptProposedAction();
	}
}

void TMachineCheckingGauge::dropEvent(
	QDropEvent 							*drop_event)
{
	QList<QUrl>							urls;
	QString								file_name;
	TMachineCheckingGauge::TFileType	file_type;
	
	urls = drop_event->mimeData()->urls();
	
	if(urls.size() == 1)
	{
		if(d_data_changed)
		{
			d_msg_box->setText("Unsaved Data");
			d_msg_box->setInformativeText("The current data has not been saved.  Do you want to continue?");
			d_msg_box->setDetailedText(QString());
			d_msg_box->setIcon(QMessageBox::Warning);
			d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
			d_msg_box->setDefaultButton(QMessageBox::No);
			
			if(QMessageBox::No == d_msg_box->exec())
			{
				drop_event->ignore();
				return;
			}
		}

		file_name = (*urls.begin()).toLocalFile();
		
		file_type = this->Identify_File_Type(file_name);
		
		switch(file_type)
		{
			case TMachineCheckingGauge::FILE_TYPE_MCG:
				this->Open_File(file_name);
				
				d_recent_files.push_front(file_name);
				d_recent_files.removeDuplicates();
				
				this->Update_Recent_File_Actions();
				break;

			case TMachineCheckingGauge::FILE_TYPE_XML:
				this->Import_XML_Measurement(file_name);
				break;
				
			case TMachineCheckingGauge::FILE_TYPE_TXT:
				this->Import_Text_Measurement(file_name);
				break;
				
			case TMachineCheckingGauge::FILE_TYPE_DMO:
				this->Import_DMO_Measurement(file_name);
				break;
				
			default:
				d_msg_box->setText("Unrecognized Format");
				d_msg_box->setInformativeText("The format of the imported data is not recognized.");
				d_msg_box->setDetailedText(QString());
				d_msg_box->setIcon(QMessageBox::Critical);
				d_msg_box->setStandardButtons(QMessageBox::Ok);
				d_msg_box->setDefaultButton(QMessageBox::Ok);
				
				d_msg_box->exec();
				break;
		}
	}
	else
	{
		d_msg_box->setText("Multiple File Import not Supported");
		d_msg_box->setInformativeText("Only a single file can be added using drag and drop.");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setIcon(QMessageBox::Information);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		
		d_msg_box->exec();
	}
}

void TMachineCheckingGauge::closeEvent(
	QCloseEvent							*event)
{
	if(d_data_changed)
	{
		d_msg_box->setText("Unsaved Data");
		d_msg_box->setInformativeText("The current data has not been saved.  Do you want to close?");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setIcon(QMessageBox::Warning);
		d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
		d_msg_box->setDefaultButton(QMessageBox::No);
		
		if(QMessageBox::No == d_msg_box->exec())
		{
			event->ignore();
			return;
		}
	}
	
	this->Save_Recent_File_List(d_data_path);

	d_settings->setValue("Path_Horizontal_Arm", d_path_widget->Horizontal_Arm_Checked());
	d_settings->setValue("Selected_MCG_Arm", d_path_widget->Arm_Length_Selection());

	d_settings->setValue("Path_Elevation_Start", d_path_widget->Elevation_Start());
	d_settings->setValue("Path_Elevation_End", d_path_widget->Elevation_End());

	d_settings->setValue("Path_Azimuth_Count", d_path_widget->Azimuth_Count());
	d_settings->setValue("Path_Elevation_Count", d_path_widget->Elevation_Count());
	d_settings->setValue("Path_Pattern_Count", d_path_widget->Pattern_Count());
	d_settings->setValue("Path_Point_Count", d_path_widget->Point_Count());

	d_settings->setValue("Path_Arm_Axis", d_path_widget->Arm_Axis());
	
	d_settings->setValue("Tolerance_A", d_tolerance_widget->Tolerance_Item_A());
	d_settings->setValue("Tolerance_B", d_tolerance_widget->Tolerance_Item_B());
	d_settings->setValue("Tolerance_C", d_tolerance_widget->Tolerance_Item_C());
	d_settings->setValue("Tolerance_D", d_tolerance_widget->Tolerance_Item_D());

	d_settings->setValue("File_Path",d_file_path);

	event->accept();
}


void TMachineCheckingGauge::Open_File(
	 const QString						&file_name)
{
	int									index;
	
	if(file_name.length())
	{
		if(!d_measurement.Load_MCG_Data(file_name))
		{
			d_msg_box->setText("Load MCG Error");
			d_msg_box->setInformativeText("Cannot load the MCG file.");
			d_msg_box->setDetailedText(QString("ERR: %1").arg(d_measurement.Last_Error()));
			d_msg_box->setIcon(QMessageBox::Critical);
			d_msg_box->setStandardButtons(QMessageBox::Ok);
			d_msg_box->setDefaultButton(QMessageBox::Ok);
			
			d_msg_box->exec();
			return;
		}
		
		d_tolerance_widget->Set_Tolerance_Expression(d_measurement.Tolerance_Expression());
		
		d_active_file_name = file_name;
		
		d_data_changed = false;
		d_path_widget->setEnabled(false);
		
		this->Update_Display();
		this->Update_Path();
		
		d_file_path = d_active_file_name;
		
		index = d_file_path.lastIndexOf('/');
		if(!(index < 0))
		{
			d_file_path.truncate(index);
		}
		
		d_export_action->setEnabled(true);
		d_new_action->setEnabled(true);
		d_save_action->setEnabled(true);
		d_report_print_action->setEnabled(true);
		d_report_output_action->setEnabled(true);
	}
}

void TMachineCheckingGauge::Save_File(
	const QString						&file_name)
{
	int									index;
	
	d_measurement.Set_Tolerance_Expression(d_tolerance_widget->Tolerance_Expression());

	if(d_measurement.Save_MCG_Data(file_name))
	{
		d_active_file_name = file_name;
		d_data_changed = false;

		d_file_path = d_active_file_name;
		
		index = d_file_path.lastIndexOf('/');
		if(!(index < 0))
		{
			d_file_path.truncate(index);
		}
	}
	else
	{
		d_msg_box->setText("Save Error");
		d_msg_box->setInformativeText("Cannot save the MCG file.");
		d_msg_box->setDetailedText(QString("ERR: %1").arg(d_measurement.Last_Error()));
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		
		d_msg_box->exec();
		return;
	}
}

void TMachineCheckingGauge::Import_XML_Measurement(
	const QString						&file_name)
{
	int									index;
	
	if(!d_measurement.Import_XML_Measurement(file_name))
	{
		d_msg_box->setText("Import Error");
		d_msg_box->setInformativeText("Cannot import the selected file.");
		d_msg_box->setDetailedText(QString("ERR: %1").arg(d_measurement.Last_Error()));
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		
		d_msg_box->exec();
		return;
	}
	
	d_data_changed = true;
	d_path_widget->setEnabled(false);
	
	this->Update_Display();
	this->Update_Path();

	d_file_path = file_name;
	
	index = d_file_path.lastIndexOf('/');
	if(!(index < 0))
	{
		d_file_path.truncate(index);
	}
	
	d_export_action->setEnabled(true);
	d_new_action->setEnabled(true);
	d_save_action->setEnabled(true);
	d_report_print_action->setEnabled(true);
	d_report_output_action->setEnabled(true);
}

void TMachineCheckingGauge::Import_DMO_Measurement(
	const QString						&file_name)
{
	int									index;
	
	if(!d_measurement.Import_DMO_Measurement(file_name))
	{
		d_msg_box->setText("Import Error");
		d_msg_box->setInformativeText("Cannot import the selected file.");
		d_msg_box->setDetailedText(QString("ERR: %1").arg(d_measurement.Last_Error()));
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		
		d_msg_box->exec();
		return;
	}
	
	d_data_changed = true;
	d_path_widget->setEnabled(false);
	
	this->Update_Display();
	this->Update_Path();

	d_file_path = file_name;
	
	index = d_file_path.lastIndexOf('/');
	if(!(index < 0))
	{
		d_file_path.truncate(index);
	}
	
	d_export_action->setEnabled(true);
	d_new_action->setEnabled(true);
	d_save_action->setEnabled(true);
	d_report_print_action->setEnabled(true);
	d_report_output_action->setEnabled(true);
}

void TMachineCheckingGauge::Import_Text_Measurement(
	const QString						&file_name)
{
	int									index;
	
	if(!d_measurement.Import_Text_Measurement(file_name))
	{
		d_msg_box->setText("Import Error");
		d_msg_box->setInformativeText("Cannot import the selected file.");
		d_msg_box->setDetailedText(QString("ERR: %1").arg(d_measurement.Last_Error()));
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		
		d_msg_box->exec();
		return;
	}
	
	d_data_changed = true;
	d_path_widget->setEnabled(false);
	
	this->Update_Display();
	this->Update_Path();

	d_file_path = file_name;
	
	index = d_file_path.lastIndexOf('/');
	if(!(index < 0))
	{
		d_file_path.truncate(index);
	}
	
	d_export_action->setEnabled(true);
	d_new_action->setEnabled(true);
	d_save_action->setEnabled(true);
	d_report_print_action->setEnabled(true);
	d_report_output_action->setEnabled(true);
}

void TMachineCheckingGauge::Export_Measurement(
	const QString						&file_name)
{
	int									index;
	
	if(!d_measurement.Export_MCG_Measurements(file_name,d_tolerance_widget->Tolerance_Value(d_measurement.Length_Average())))
	{
		d_msg_box->setText("Export Error");
		d_msg_box->setInformativeText("Cannot export the MCG data.");
		d_msg_box->setDetailedText(QString("ERR: %1").arg(d_measurement.Last_Error()));
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		
		d_msg_box->exec();
		return;
	}
	
	d_file_path = file_name;
	
	index = d_file_path.lastIndexOf('/');
	if(!(index < 0))
	{
		d_file_path.truncate(index);
	}
}

void TMachineCheckingGauge::Update_Display(void)
{
	static const int					NUMBER_FREQUENCY_ITEMS = 11;
	int									set;
	int									count;
	int									cntr;
	TVector3							center;
	TVector3							vmin,vmax;
	TVector3							pivot_center;
	TVector3							vec;
	TVector2							pnt;
	TMeasurement::TMeasurementItem		meas;
	double								dval;
	int									flier_count;
	double								average_length;
	double								minimum_length;
	double								maximum_length;
	double								median_length;
	double								range;
	double								tolerance;
	double								std_deviation;
	double								error_x;
	double								error_y;
	double								error_z;
	bool								oot;
	bool								valid;
	double								measurement_window;
	int									cntr_frequency;
	int									frequency_max;
	int									frequency_array[NUMBER_FREQUENCY_ITEMS];
	
	count = d_measurement.Measurement_Count();
	
	d_opengl_display_widget->Clear_MCG();
	d_analysis_widget->Clear();
	d_report_widget->Clear_Report();

	if(!count)
	{
		return;
	}
	
	d_report_widget->Set_Date_Time(d_measurement.Date_Time().toString("dd MMM yyyy hh:mm:ss ap"));

	average_length = d_measurement.Length_Average();
	d_report_widget->Set_Average_Measurement_Length(average_length);

	tolerance = d_tolerance_widget->Tolerance_Value(average_length);
	d_report_widget->Set_Tolerance(tolerance);
	
	vmin = d_measurement.Volume_Corner_Min();
	vmax = d_measurement.Volume_Corner_Max();
	
	minimum_length = d_measurement.Length_Min();
	maximum_length = d_measurement.Length_Max();
	std_deviation = d_measurement.Length_Stdev();
	
	range = maximum_length - minimum_length;
	median_length = (maximum_length + minimum_length)/2.0;
	
	d_report_widget->Set_Measurement_Range(range);
	d_report_widget->Set_Minimum_Measured_Length(minimum_length);
	d_report_widget->Set_Maximum_Measured_Length(maximum_length);
	d_report_widget->Set_Median_Measured_Length(median_length);
	d_report_widget->Set_Stddev_Measured_Length(std_deviation);

	d_opengl_display_widget->Define_Machine_Volume(vmin,vmax);
	
	memset(frequency_array,0,sizeof(int) * NUMBER_FREQUENCY_ITEMS);
	
	d_analysis_widget->Add_Text(QStringLiteral("Measurement Volume"));
	d_analysis_widget->Add_Text(QStringLiteral("-----------------------------------------------------------------------"));
	d_analysis_widget->Add_Text(QString("          Start X: %1").arg(vmin.x,10,'f',4));
	d_analysis_widget->Add_Text(QString("            End X: %1").arg(vmax.x,10,'f',4));
	d_analysis_widget->Add_Text(QString("          Range X: %1").arg(vmax.x - vmin.x,10,'f',4));
	d_analysis_widget->Add_Text(QString("          Start Y: %1").arg(vmin.y,10,'f',4));
	d_analysis_widget->Add_Text(QString("            End Y: %1").arg(vmax.y,10,'f',4));
	d_analysis_widget->Add_Text(QString("          Range Y: %1").arg(vmax.y - vmin.y,10,'f',4));
	d_analysis_widget->Add_Text(QString("          Start Z: %1").arg(vmin.z,10,'f',4));
	d_analysis_widget->Add_Text(QString("            End Z: %1").arg(vmax.z,10,'f',4));
	d_analysis_widget->Add_Text(QString("          Range Z: %1\n").arg(vmax.z - vmin.z,10,'f',4));
	
	d_analysis_widget->Add_Text(QStringLiteral("Measurement Analysis"));
	d_analysis_widget->Add_Text(QStringLiteral("-----------------------------------------------------------------------"));

	dval = d_measurement.Temperature();
	d_report_widget->Set_Temperature(dval);
	d_analysis_widget->Add_Text(QString("      Temperature: %1 C").arg(dval,8,'f',2));

	vec = d_measurement.Probe_Offset();
	d_report_widget->Set_Probe_Offset(vec);
	d_analysis_widget->Add_Text(QString("     Probe Offset: %1, %2, %3").arg(vec.x,10,'f',4).arg(vec.y,10,'f',4).arg(vec.z,10,'f',4));

	dval = d_measurement.Probe_Error();
	d_report_widget->Set_Probe_Error(dval);
	d_analysis_widget->Add_Text(QString("      Probe Error: %1").arg(dval,10,'f',4));
	
	dval = d_measurement.Average_XY_Length();
	valid = d_measurement.Average_XY_Valid();
	if(valid)
	{
		d_analysis_widget->Add_Text(QString("Average Length XY: %1").arg(dval,10,'f',4));
	}
	else
	{
		d_analysis_widget->Add_Text(QStringLiteral("Average Length XY: Not Enough Data"));
	}
	d_report_widget->Set_Average_Length_XY(dval,valid);
	
	dval = d_measurement.Average_YZ_Length();
	valid = d_measurement.Average_YZ_Valid();
	if(valid)
	{
		d_analysis_widget->Add_Text(QString("Average Length YZ: %1").arg(dval,10,'f',4));
	}
	else
	{
		d_analysis_widget->Add_Text(QStringLiteral("Average Length YZ: Not Enough Data"));
	}
	d_report_widget->Set_Average_Length_YZ(dval,valid);
	
	dval = d_measurement.Average_ZX_Length();
	valid = d_measurement.Average_ZX_Valid();
	if(valid)
	{
		d_analysis_widget->Add_Text(QString("Average Length ZX: %1").arg(dval,10,'f',4));
	}
	else
	{
		d_analysis_widget->Add_Text(QStringLiteral("Average Length ZX: Not Enough Data"));
	}
	d_report_widget->Set_Average_Length_ZX(dval,valid);
	
	
	dval = d_measurement.Average_X_Length();
	valid = d_measurement.Average_X_Valid();
	if(valid)
	{
		d_analysis_widget->Add_Text(QString(" Average Length X: %1").arg(dval,10,'f',4));
		d_analysis_widget->Add_Text(QString("Relative Length X: %1").arg(dval - average_length,10,'f',4));
	}
	else
	{
		d_analysis_widget->Add_Text(QStringLiteral(" Average Length X: Not Enough Data"));
		d_analysis_widget->Add_Text(QStringLiteral("Relative Length X: Not Enough Data"));
	}

	error_x = dval - average_length;
	d_report_widget->Set_Relative_Scale_Error_X(error_x,valid);
	
	dval = d_measurement.Average_Y_Length();
	valid = d_measurement.Average_Y_Valid();
	if(valid)
	{
		d_analysis_widget->Add_Text(QString(" Average Length Y: %1").arg(dval,10,'f',4));
		d_analysis_widget->Add_Text(QString("Relative Length Y: %1").arg(dval - average_length,10,'f',4));
	}
	else
	{
		d_analysis_widget->Add_Text(QStringLiteral(" Average Length Y: Not Enough Data"));
		d_analysis_widget->Add_Text(QStringLiteral("Relative Length Y: Not Enough Data"));
	}
	
	error_y = dval - average_length;
	d_report_widget->Set_Relative_Scale_Error_Y(error_y,valid);
	
	dval = d_measurement.Average_Z_Length();
	valid = d_measurement.Average_Z_Valid();
	if(valid)
	{
		d_analysis_widget->Add_Text(QString(" Average Length Z: %1").arg(dval,10,'f',4));
		d_analysis_widget->Add_Text(QString("Relative Length Z: %1").arg(dval - average_length,10,'f',4));
	}
	else
	{
		d_analysis_widget->Add_Text(QStringLiteral(" Average Length Z: Not Enough Data"));
		d_analysis_widget->Add_Text(QStringLiteral("Relative Length Z: Not Enough Data"));
	}
	
	error_z = dval - average_length;
	d_report_widget->Set_Relative_Scale_Error_Z(error_z,valid);

	
	dval = d_measurement.Squareness_XY_Error();
	valid = d_measurement.Squareness_XY_Valid();
	if(valid)
	{
		d_analysis_widget->Add_Text(QString("    Squareness XY: %1 mm/m").arg(dval,10,'f',4));
	}
	else
	{
		d_analysis_widget->Add_Text(QStringLiteral("    Squareness XY: Not Enough Data"));
	}
	
	d_report_widget->Set_Squareness_Error_XY(dval,valid);
	
	dval = d_measurement.Squareness_YZ_Error();
	valid = d_measurement.Squareness_YZ_Valid();
	if(valid)
	{
		d_analysis_widget->Add_Text(QString("    Squareness YZ: %1 mm/m").arg(dval,10,'f',4));
	}
	else
	{
		d_analysis_widget->Add_Text(QStringLiteral("    Squareness YZ: Not Enough Data"));
	}
	
	d_report_widget->Set_Squareness_Error_YZ(dval,valid);
	
	dval = d_measurement.Squareness_ZX_Error();
	valid = d_measurement.Squareness_ZX_Valid();
	if(valid)
	{
		d_analysis_widget->Add_Text(QString("    Squareness ZX: %1 mm/m\n").arg(dval,10,'f',4));
	}
	else
	{
		d_analysis_widget->Add_Text(QStringLiteral("    Squareness ZX: Not Enough Data\n"));
	}
	
	d_report_widget->Set_Squareness_Error_ZX(dval,valid);
	

	d_analysis_widget->Add_Text(QStringLiteral("Measurement Statistics"));
	d_analysis_widget->Add_Text(QStringLiteral("-----------------------------------------------------------------------"));
	d_analysis_widget->Add_Text(QString("Measurement Count: %1").arg(count,10));
	d_analysis_widget->Add_Text(QString("   Minimum Length: %1").arg(minimum_length,10,'f',4));
	d_analysis_widget->Add_Text(QString("   Maximum Length: %1").arg(maximum_length,10,'f',4));
	d_analysis_widget->Add_Text(QString("   Average Length: %1").arg(average_length,10,'f',4));
	d_analysis_widget->Add_Text(QString("    Median Length: %1").arg(median_length,10,'f',4));
	d_analysis_widget->Add_Text(QString("    StdDev Length: %1").arg(std_deviation,10,'f',4));
	d_analysis_widget->Add_Text(QString("     Range Length: %1").arg(range,10,'f',4));
	d_analysis_widget->Add_Text(QString("        Tolerance: %1\n").arg(tolerance,10,'f',4));

	d_analysis_widget->Add_Text(QString("Estimated Range for 99% of Measurements: %1").arg(std_deviation * 3,0,'f',5));
	
	measurement_window = (maximum_length - minimum_length) / static_cast<double>(NUMBER_FREQUENCY_ITEMS - 1);
	
	flier_count = 0;
	center = d_measurement.MCG_Sphere_Center();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		meas = d_measurement.MCG_Measurement(cntr);
		
		if(meas.length > (median_length + tolerance/2.0) ||
		   meas.length < (median_length - tolerance/2.0))
		{
			oot = true;
		}
		else
		{
			oot = false;
		}
		
		if((meas.length < (average_length - std_deviation * 1.5)) || (meas.length > (average_length + std_deviation * 1.5)))
		{
			flier_count++;
		}
		
		d_opengl_display_widget->Add_MCG(cntr,center,meas.axis,meas.length,oot);
	}
	
	d_opengl_display_widget->Update(true);

	d_analysis_widget->Add_Text(QString("Number Outside of Estimated Range: %1\n").arg(flier_count));
	
	d_analysis_widget->Add_Text(QStringLiteral("Measurement Center"));
	d_analysis_widget->Add_Text(QStringLiteral("-----------------------------------------------------------------------"));
	
	pivot_center = d_measurement.MCG_Sphere_Center_Start();
	d_analysis_widget->Add_Text(QString("      Pivot Start: %1, %2, %3").arg(pivot_center.x,10,'f',4).arg(pivot_center.y,10,'f',4).arg(pivot_center.z,10,'f',4));
	d_report_widget->Set_Pivot_Sphere_Start(pivot_center);

	pivot_center = d_measurement.MCG_Sphere_Center_End();
	d_analysis_widget->Add_Text(QString("        Pivot End: %1, %2, %3").arg(pivot_center.x,10,'f',4).arg(pivot_center.y,10,'f',4).arg(pivot_center.z,10,'f',4));
	d_report_widget->Set_Pivot_Sphere_End(pivot_center);

	d_analysis_widget->Add_Text(QString("      Data Center: %1, %2, %3\n").arg(center.x,10,'f',4).arg(center.y,10,'f',4).arg(center.z,10,'f',4));
	d_report_widget->Set_Measurement_Sphere(center);

	
	
	d_analysis_widget->Add_Text(QStringLiteral("Individual Measurement Results"));
	d_analysis_widget->Add_Text(QStringLiteral("-----------------------------------------------------------------------\n"));
	d_analysis_widget->Add_Text(QStringLiteral("Azimuth Elevation  Length      Min        Max     Std Dev   Dev"));
	d_analysis_widget->Add_Text(QStringLiteral("-----------------------------------------------------------------------"));

	for(cntr = 0;cntr < count;++cntr)
	{
		meas = d_measurement.MCG_Measurement(cntr);
		
		if(meas.length > (median_length + tolerance/2.0) ||
		   meas.length < (median_length - tolerance/2.0))
		{
			d_analysis_widget->Add_Text(QString("%1 %2 %3 %4 %5 %6 %7  <-Out of tolerance.")
											.arg(meas.azimuth,7,'f',1)
											.arg(meas.elevation,7,'f',1)
											.arg(meas.length,10,'f',4)
											.arg(meas.length_min,10,'f',4)
											.arg(meas.length_max,10,'f',4)
											.arg(meas.length_stddev,8,'f',4)
											.arg(meas.length - average_length,8,'f',4));
		}
		else
		{
			d_analysis_widget->Add_Text(QString("%1 %2 %3 %4 %5 %6 %7")
											.arg(meas.azimuth,7,'f',1)
											.arg(meas.elevation,7,'f',1)
											.arg(meas.length,10,'f',4)
											.arg(meas.length_min,10,'f',4)
											.arg(meas.length_max,10,'f',4)
											.arg(meas.length_stddev,8,'f',4)
											.arg(meas.length - average_length,8,'f',4));
		}

		
		d_report_widget->Add_Measurement_Data(meas.elevation,meas.azimuth,meas.length,meas.length_max - meas.length_min,meas.length_stddev,meas.length - average_length);
		
		dval = minimum_length;
		
		for(cntr_frequency = 0;cntr_frequency < NUMBER_FREQUENCY_ITEMS;cntr_frequency++)
		{
			if(meas.length > (dval - (measurement_window/2.0)) && meas.length < (dval + (measurement_window/2.0)))
			{
				frequency_array[cntr_frequency]++;
				break;
			}
			
			dval += measurement_window;
		}
	}
	
	if(range > 0.000001)
	{
		set = d_analysis_widget->Create_Graph_Set(QColor(0,0,255));
		
		// Force one value to zero
		pnt.x = average_length;
		pnt.y = 0;
		
		d_analysis_widget->Add_Point_Graph(set,pnt);
		
		frequency_max = 0;
		
		for(cntr=0;cntr < NUMBER_FREQUENCY_ITEMS;++cntr)
		{
			pnt.x = minimum_length + static_cast<double>(cntr) * measurement_window;
			pnt.y = static_cast<double>(frequency_array[cntr]);
			
			if(frequency_array[cntr] > frequency_max)
			{
				frequency_max = frequency_array[cntr];
			}
			
			d_analysis_widget->Add_Point_Graph(set,pnt);
		}
		
		if(frequency_max)
		{
			set = d_analysis_widget->Create_Graph_Set(QColor(128,0,0),TGraphWidget::LINE_GRAPH);

			pnt.x = median_length - tolerance / 2;
			pnt.y = 0;
			d_analysis_widget->Add_Point_Graph(set,pnt);

			pnt.y = static_cast<double>(frequency_max);
			d_analysis_widget->Add_Point_Graph(set,pnt);

			set = d_analysis_widget->Create_Graph_Set(QColor(128,0,0),TGraphWidget::LINE_GRAPH);
			
			pnt.x = median_length + tolerance / 2;
			pnt.y = 0;
			d_analysis_widget->Add_Point_Graph(set,pnt);
			
			pnt.y = static_cast<double>(frequency_max);
			d_analysis_widget->Add_Point_Graph(set,pnt);
		}
		
		d_analysis_widget->Scale_To_Fit_Graph();
	}
	else
	{
		d_analysis_widget->Add_Text(QStringLiteral("\nWARN: There range of the data is too small to graph"));
	}

	d_analysis_widget->Add_Text_End();
	d_report_widget->Update_Report();
}

void TMachineCheckingGauge::Update_Recent_File_Actions(void)
{
	QList<QAction*>::iterator			action_iter;
	QStringList::const_iterator			iter;
	QString								file_name;
	
	for(action_iter = d_recent_file_actions.begin();action_iter != d_recent_file_actions.end();++action_iter)
	{
		(*action_iter)->setVisible(false);
	}
	
	while(d_recent_files.size() > RECENT_FILE_LIST_SIZE)
	{
		d_recent_files.pop_back();
	}
	
	for(iter = d_recent_files.begin(),action_iter = d_recent_file_actions.begin();iter != d_recent_files.end();++iter,++action_iter)
	{
		file_name = QFileInfo(*iter).fileName();
		
		(*action_iter)->setText(file_name);
		(*action_iter)->setData(*iter);
		(*action_iter)->setVisible(true);
	}
}

TMachineCheckingGauge::TFileType TMachineCheckingGauge::Identify_File_Type(
	const QString						&file_name) const
{
	TMachineCheckingGauge::TFileType	file_type(TMachineCheckingGauge::FILE_TYPE_UNKNOWN);
	QByteArray							file_data;
	QFile								file(file_name);
	QString								text;
	
	if(file.open(QIODevice::ReadOnly))
	{
		file_data = file.readAll();
		file.close();
		
		if(file_data.size() > 100)
		{
			file_data.resize(100);
			
			text = QString(file_data).trimmed();
			
			if(text.contains(QStringLiteral("<MCG_Data ")))
			{
				file_type = TMachineCheckingGauge::FILE_TYPE_MCG;
			}
			else if(text.contains(QStringLiteral("<MCG_Measurement_Data ")))
			{
				file_type = TMachineCheckingGauge::FILE_TYPE_XML;
			}
			else if(text.startsWith(QStringLiteral("FILNAM")))
			{
				file_type = TMachineCheckingGauge::FILE_TYPE_DMO;
			}
			else	// a big assumption...
			{
				file_type = TMachineCheckingGauge::FILE_TYPE_TXT;
			}
		}
	}
	
	return file_type;
}

void TMachineCheckingGauge::Load_Recent_File_List(
	const QString 						&file_path)
{
	QString								file_name;
	QString								text;
	TXmlFile							xml_file;
	QDomElement							element;
	int									version;
	int									cntr;
	
	file_name = file_path;
	
	if(file_name.endsWith('\\') || file_name.endsWith('/'))
	{
		file_name.chop(1);
	}
	
	file_name += QStringLiteral("/recent_file_list.xml");
	
	try
	{
		xml_file.Set_File_Name(file_name);
		
		version = 1;
		xml_file.Open("Recent_File_List",&version);
		
		element = xml_file.Get_Node(xml_file.Root_Node(),"Recent_File",true);
		
		cntr = 0;
		d_recent_files.clear();
		
		while(!element.isNull() && cntr < RECENT_FILE_LIST_SIZE)
		{
			text = xml_file.Read_Text_Node(&element,true);
			
			if(text.length())
			{
				d_recent_files.push_back(text);
			}
			
			element = xml_file.Get_Sibling_Node(&element,"Recent_File",true);
			
			++cntr;
		}
		
		d_recent_files.removeDuplicates();
	}
	catch(TXmlFileException XmlFileException)
	{
		
	}
}

void TMachineCheckingGauge::Save_Recent_File_List(
	const QString 						&file_path) const
{
	QString								file_name;
	QStringList::const_iterator			iter;
	TXmlFile							xml_file;
	QDomElement							element;
	
	file_name = file_path;
	
	if(file_name.endsWith('\\') || file_name.endsWith('/'))
	{
		file_name.chop(1);
	}
	
	file_name += QStringLiteral("/recent_file_list.xml");
	
	try
	{
		xml_file.Set_File_Name(file_name);
		xml_file.Reset("Recent_File_List",1);
		
		for(iter = d_recent_files.begin();iter != d_recent_files.end();++iter)
		{
			element = xml_file.Create_Node(xml_file.Root_Node(),"Recent_File");
			xml_file.Write_Text_Node(&element,(*iter));
		}

		xml_file.Write_File();
	}
	catch(TXmlFileException XmlFileException)
	{
		
	}
}

