/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QFile>
#include <QFileInfo>
#include <QTextStream>
#include <QRegularExpression>
#include <QRegularExpressionMatch>
#include <cmath>
#include <assert.h>

#include "../../core/xmlfile.h"
#include "../../core/bestfitsphere.h"

#include "measurement.h"

TMeasurement::TMeasurement(void)
{
	d_pattern_average = 0.0;
	d_pattern_stdev = 0.0;
	d_pattern_min_length = 0.0;
	d_pattern_max_length = 0.0;
	
	d_average_x_length_valid = false;
	d_average_y_length_valid = false;
	d_average_z_length_valid = false;
	
	d_average_xy_length_valid = false;
	d_average_yz_length_valid = false;
	d_average_zx_length_valid = false;
	
	d_squareness_xy_valid = false;
	d_squareness_yz_valid = false;
	d_squareness_zx_valid = false;
}

TMeasurement::~TMeasurement(void)
{

}

TVector3 TMeasurement::Volume_Corner_Min(void) const
{
	TVector3							pnt;
	
	pnt = d_measurement_sphere_position;
	
	pnt.x -= d_pattern_average;
	pnt.y -= d_pattern_average;
	pnt.z -= d_pattern_average;
	
	return pnt;
}

TVector3 TMeasurement::Volume_Corner_Max(void) const
{
	TVector3							pnt;
	
	pnt = d_measurement_sphere_position;
	
	pnt.x += d_pattern_average;
	pnt.y += d_pattern_average;
	pnt.z += d_pattern_average;
	
	return pnt;
}

TMeasurement::TMeasurementItem TMeasurement::MCG_Measurement(
	const int 							index) const
{
	assert(static_cast<unsigned int>(index) < d_measurement_items.size());
	
	return d_measurement_items[index];
}

bool TMeasurement::Save_MCG_Data(
	const QString						&file_name) const
{
	TXmlFile							xml_file;
	QDomElement							e0;
	QDomElement							e1;
	std::vector<TVector3>::const_iterator iter;
	
	try
	{
		xml_file.Set_File_Name(file_name);
		xml_file.Reset("MCG_Data",1);
		
		// date time
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Date_Time"));
		xml_file.Write_Text_Node(&e0,QString("%1").arg(d_date_time.toMSecsSinceEpoch()));
		
		// measurement sphere position
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Sphere_Position"));
		xml_file.Write_Text_Node(&e0,QString("%1,%2,%3").arg(d_measurement_sphere_position.x,0,'f',8).arg(d_measurement_sphere_position.y,0,'f',8).arg(d_measurement_sphere_position.z,0,'f',8));
		
		// pivot sphere start position
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Pivot_Sphere_Start_Position"));
		xml_file.Write_Text_Node(&e0,QString("%1,%2,%3").arg(d_pivot_sphere_position_start.x,0,'f',8).arg(d_pivot_sphere_position_start.y,0,'f',8).arg(d_pivot_sphere_position_start.z,0,'f',8));
		
		// pivot sphere end position
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Pivot_Sphere_End_Position"));
		xml_file.Write_Text_Node(&e0,QString("%1,%2,%3").arg(d_pivot_sphere_position_end.x,0,'f',8).arg(d_pivot_sphere_position_end.y,0,'f',8).arg(d_pivot_sphere_position_end.z,0,'f',8));
		
		// probe offset
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Probe_Offset"));
		xml_file.Write_Text_Node(&e0,QString("%1,%2,%3").arg(d_probe_offset.x,0,'f',8).arg(d_probe_offset.y,0,'f',8).arg(d_probe_offset.z,0,'f',8));
		
		// probe error
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Probe_Error"));
		xml_file.Write_Text_Node(&e0,QString("%1").arg(d_probe_error,0,'f',8));
		
		// temperature
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Temperature"));
		xml_file.Write_Text_Node(&e0,QString("%1").arg(d_temperature,0,'f',8));
		
		// tolerance
		if(d_tolerance_expression.length())
		{
			e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Tolerance_Expression"));
			xml_file.Write_Text_Node(&e0,d_tolerance_expression);
		}

		// measurement points
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Gauge_Points"));

		for(iter = d_points.begin();iter != d_points.end();++iter)
		{
			e1 = xml_file.Create_Node(&e0,QStringLiteral("Point_XYZ"));
			xml_file.Write_Text_Node(&e1,QString("%1,%2,%3").arg((*iter).x,0,'f',8).arg((*iter).y,0,'f',8).arg((*iter).z,0,'f',8));
		}
	
		xml_file.Write_File();
	}
	catch(TXmlFileException XmlFileException)
	{
		d_last_error = QString("error writing file '%1': %2").arg(file_name).arg(XmlFileException.ErrorText());
		return false;
	}
	
	return true;
}

bool TMeasurement::Export_MCG_Measurements(
	const QString						&file_name,
	const double						&tolerance) const
{
	QFile								file(file_name);
	TMeasurementItem						meas;
	double								average_length;
	int									cntr;
	
	average_length = this->Length_Average();
	
	if(file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		file.write("\"Measurement Analysis\"\n");
		file.write(QString("\"Temperature\",%1\n").arg(d_temperature,0,'f',6).toLatin1());
		file.write(QString("\"Probe Offset\",%1,%2,%3\n").arg(d_probe_offset.x,0,'f',6).arg(d_probe_offset.y,0,'f',6).arg(d_probe_offset.z,0,'f',6).toLatin1());
		file.write(QString("\"Probe Error\",%1\n").arg(d_probe_error,0,'f',6).toLatin1());

		if(d_average_xy_length_valid)
			file.write(QString("\"Average XY Length\",%1\n").arg(d_average_xy_length,0,'f',6).toLatin1());
		else
			file.write("\"Average XY Length\",\"Not Enough Data\"\n");
		
		if(d_average_yz_length_valid)
			file.write(QString("\"Average YZ Length\",%1\n").arg(d_average_yz_length,0,'f',6).toLatin1());
		else
			file.write("\"Average YZ Length\",\"Not Enough Data\"\n");
		
		if(d_average_zx_length_valid)
			file.write(QString("\"Average ZX Length\",%1\n").arg(d_average_zx_length,0,'f',6).toLatin1());
		else
			file.write("\"Average ZX Length\",\"Not Enough Data\"\n");
		
		if(d_average_x_length_valid)
		{
			file.write(QString("\"Average X Length\",%1\n").arg(d_average_x_length,0,'f',6).toLatin1());
			file.write(QString("\"Relative X Length\",%1\n").arg(d_average_x_length - average_length,0,'f',6).toLatin1());
		}
		else
		{
			file.write("\"Average X Length\",\"Not Enough Data\"\n");
			file.write("\"Relative X Length\",\"Not Enough Data\"\n");
		}
		
		if(d_average_y_length_valid)
		{
			file.write(QString("\"Average Y Length\",%1\n").arg(d_average_y_length,0,'f',6).toLatin1());
			file.write(QString("\"Relative Y Length\",%1\n").arg(d_average_y_length - average_length,0,'f',6).toLatin1());
		}
		else
		{
			file.write("\"Average Y Length\",\"Not Enough Data\"\n");
			file.write("\"Relative Y Length\",\"Not Enough Data\"\n");
		}
		
		if(d_average_z_length_valid)
		{
			file.write(QString("\"Average Z Length\",%1\n").arg(d_average_z_length,0,'f',6).toLatin1());
			file.write(QString("\"Relative Z Length\",%1\n").arg(d_average_z_length - average_length,0,'f',6).toLatin1());
		}
		else
		{
			file.write("\"Average Z Length\",\"Not Enough Data\"\n");
			file.write("\"Relative Z Length\",\"Not Enough Data\"\n");
		}
		
		if(d_squareness_xy_valid)
			file.write(QString("\"Squareness XY\",%1\n").arg(d_squareness_xy_error,0,'f',6).toLatin1());
		else
			file.write("\"Squareness XY\",\"Not Enough Data\"\n");
		
		if(d_squareness_yz_valid)
			file.write(QString("\"Squareness YZ\",%1\n").arg(d_squareness_yz_error,0,'f',6).toLatin1());
		else
			file.write("\"Squareness YZ\",\"Not Enough Data\"\n");
		
		if(d_squareness_zx_valid)
			file.write(QString("\"Squareness ZX\",%1\n").arg(d_squareness_zx_error,0,'f',6).toLatin1());
		else
			file.write("\"Squareness ZX\",\"Not Enough Data\"\n");

		file.write("\n\"Measurement Statistics\"\n");
		file.write(QString("\"Measurement Count\",%1\n").arg(this->Measurement_Count()).toLatin1());
		file.write(QString("\"Minimum Length\",%1\n").arg(this->Length_Min(),0,'f',6).toLatin1());
		file.write(QString("\"Maximum Length\",%1\n").arg(this->Length_Max(),0,'f',6).toLatin1());
		file.write(QString("\"Range Length\",%1\n").arg(this->Length_Max() - this->Length_Min(),0,'f',6).toLatin1());
		file.write(QString("\"Average Length\",%1\n").arg(average_length,0,'f',6).toLatin1());
		file.write(QString("\"Median Length\",%1\n").arg((this->Length_Max() + this->Length_Min())/2.0,0,'f',6).toLatin1());
		file.write(QString("\"StdDev Length\",%1\n").arg(this->Length_Stdev(),0,'f',6).toLatin1());
		file.write(QString("\"Estimated Range for 99% of Measurements\",%1\n").arg(this->Length_Stdev() * 3.0,0,'f',6).toLatin1());
		file.write(QString("\"Tolerance\",%1\n").arg(tolerance,0,'f',6).toLatin1());

		file.write("\n\"Measurement Center\"\n");
		file.write(QString("\"Pivot Start\",%1,%2,%3\n").arg(d_pivot_sphere_position_start.x,0,'f',6).arg(d_pivot_sphere_position_start.y,0,'f',6).arg(d_pivot_sphere_position_start.z,0,'f',6).toLatin1());
		file.write(QString("\"Pivot End\",%1,%2,%3\n").arg(d_pivot_sphere_position_end.x,0,'f',6).arg(d_pivot_sphere_position_end.y,0,'f',6).arg(d_pivot_sphere_position_end.z,0,'f',6).toLatin1());
		file.write(QString("\"Data Center\",%1,%2,%3\n").arg(d_measurement_sphere_position.x,0,'f',6).arg(d_measurement_sphere_position.y,0,'f',6).arg(d_measurement_sphere_position.z,0,'f',6).toLatin1());

		file.write("\n\"Individual Measurement Results\"\n");
		file.write("\"Point Number\",\"Azimuth\",\"Elevation\",\"Length\",\"Min\",\"Max\",\"Std Dev\",\"Dev\"\n");

		for(cntr = 0;cntr < this->Measurement_Count();++cntr)
		{
			meas = this->MCG_Measurement(cntr);
			
			file.write(QString("%1,%2,%3,%4,%5,%6,%7,%8\n")
					   .arg(cntr + 1)
					   .arg(meas.azimuth,0,'f',6)
					   .arg(meas.elevation,0,'f',6)
					   .arg(meas.length,0,'f',6)
					   .arg(meas.length_min,0,'f',6)
					   .arg(meas.length_max,0,'f',6)
					   .arg(meas.length_stddev,0,'f',6)
					   .arg(meas.length - average_length,0,'f',6)
					   .toLatin1());

		}
		
		file.close();
		return true;
	}
	
	return false;
}

void TMeasurement::Clear(void)
{
	d_measurement_items.clear();
	d_points.clear();
	
	d_temperature = 20.0;
	d_probe_error = 0.0;
	d_probe_offset = TVector3(0.0,0.0,0.0);
	d_pivot_sphere_position_start = TVector3(0.0,0.0,0.0);
	d_pivot_sphere_position_end = TVector3(0.0,0.0,0.0);
	d_measurement_sphere_position = TVector3(0.0,0.0,0.0);
	
	d_pattern_average = 0.0;
	d_pattern_stdev = 0.0;
	d_pattern_min_length = 0.0;
	d_pattern_max_length = 0.0;
	
	d_average_x_length_valid = false;
	d_average_y_length_valid = false;
	d_average_z_length_valid = false;
	
	d_average_xy_length_valid = false;
	d_average_yz_length_valid = false;
	d_average_zx_length_valid = false;
	
	d_squareness_xy_valid = false;
	d_squareness_yz_valid = false;
	d_squareness_zx_valid = false;

}

bool TMeasurement::Import_XML_Measurement(
	const QString						&file_name)
{
	TXmlFile							xmlfile;
	int									version;
	
	d_date_time = QDateTime::currentDateTime();
	
	try
	{
		xmlfile.Set_File_Name(file_name);
		xmlfile.Open("MCG_Measurement_Data",&version);
		
		switch(version)
		{
			case 1:
				this->Import_Measurement_V1(&xmlfile);
				break;
		}
		
		if(!this->Process_MCG_Data())
		{
			return false;
		}
		
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("error processing file '%1': %2").arg(file_name).arg(xml_exception.ErrorText());
		return false;
	}
	
	return true;
}

bool TMeasurement::Import_DMO_Measurement(
	const QString						&file_name)
{
	QFile								file;
	QByteArray							file_data;
	QRegularExpression					feature_result("^[FA(]{3}([a-z,A-Z_,0-9]{1,12})[)]");
	QRegularExpression					tolerance_result("^[TA(]{3}([a-z,A-Z_,0-9]{1,12})[)]");
	QRegularExpressionMatch				expression_match;
	QStringList							list;
	QString								line;
	QString								text;
	QString								feature_name;
	TVector3							pnt;
	bool								status[3];

	d_date_time = QDateTime::currentDateTime();

	file.setFileName(file_name);
	if(!file.open(QIODevice::ReadOnly))
	{
		d_last_error = QString("Cannot load file '%1'.").arg(file_name);
		return false;
	}
	
	file_data = file.readAll();
	file.close();
	
	// defaults
	d_temperature = 20.0;
	d_probe_offset = TVector3(0,0,-200);	// assumed probe offset.  not in DMO file
	d_probe_error = 0.0;
	d_pivot_sphere_position_start = TVector3(0,0,0);
	d_pivot_sphere_position_end = TVector3(0,0,0);
	d_points.clear();
	
	QTextStream							stream(file_data);
	
	while(!stream.atEnd())
	{
		line = stream.readLine().simplified();
		
		if(line.endsWith('$'))
		{
			line.chop(1);
			text += line;
			continue;
		}
		else
		{
			text += line;
		}
		
		
		if(text.length())
		{
			expression_match = feature_result.match(text);
			if(expression_match.hasMatch())
			{
				list = expression_match.capturedTexts();
				
				if(list.size() == 2)
				{
					feature_name = list[1];
				}
				
				list = text.split(',');
				
				if(list.size() > 7)	// minumum expected entries
				{
					// last entries X,Y,Z,D,I,J,K
					if(feature_name.compare(QStringLiteral("MCG_SPH_1"),Qt::CaseInsensitive) == 0)
					{
						pnt.x = list[list.size() - 7].toDouble(&status[0]);
						pnt.y = list[list.size() - 6].toDouble(&status[1]);
						pnt.z = list[list.size() - 5].toDouble(&status[2]);
						
						if(status[0] && status[1] && status[2])
						{
							d_pivot_sphere_position_start = pnt;
						}
					} // last entries X,Y,Z,D,I,J,K
					else if(feature_name.compare(QStringLiteral("MCG_SPH_2"),Qt::CaseInsensitive) == 0)
					{
						pnt.x = list[list.size() - 7].toDouble(&status[0]);
						pnt.y = list[list.size() - 6].toDouble(&status[1]);
						pnt.z = list[list.size() - 5].toDouble(&status[2]);
						
						if(status[0] && status[1] && status[2])
						{
							d_pivot_sphere_position_end = pnt;
						}
					} // last entries X,Y,Z,I,J,K
					else if(feature_name.startsWith(QStringLiteral("MCG_PNT_"),Qt::CaseInsensitive))
					{
						pnt.x = list[list.size() - 6].toDouble(&status[0]);
						pnt.y = list[list.size() - 5].toDouble(&status[1]);
						pnt.z = list[list.size() - 4].toDouble(&status[2]);
						
						if(status[0] && status[1] && status[2])
						{
							d_points.push_back(pnt);
						}
					}
				}
			}
			
			
			expression_match = tolerance_result.match(text);
			if(expression_match.hasMatch())
			{
				list = expression_match.capturedTexts();
				
				if(list.size() == 2)
				{
					feature_name = list[1];
				}
				
				list = text.split(',');
				
				if(list.size() > 1)	// minumum expected entries
				{
					// line entries TOL/TYPE, VALUE, IN/OUT_FLAG
					if(feature_name.startsWith(QStringLiteral("MCG_SPH_1"),Qt::CaseInsensitive))
					{
						d_probe_error = list[1].toDouble();
					}
				}
			}
		}
		
		text.clear();
	};
	
	if(!this->Process_MCG_Data())
	{
		return false;
	}
	
	return true;
}

bool TMeasurement::Import_Text_Measurement(
	const QString						&file_name)
{
	QFile								file;
	QByteArray							file_data;
	QString								text;
	QStringList							list;
	TVector3							pnt;
	bool								status[3];
	int									index;
	double								dval;
	
	d_date_time = QDateTime::currentDateTime();

	file.setFileName(file_name);
	if(!file.open(QIODevice::ReadOnly))
	{
		d_last_error = QString("Cannot load file '%1'.").arg(file_name);
		return false;
	}
	
	file_data = file.readAll();
	file.close();
	
	// defaults
	d_temperature = 20.0;
	d_probe_offset = TVector3(0,0,-200);	// default probe offset if not set
	d_probe_error = 0.0;
	d_pivot_sphere_position_start = TVector3(0,0,0);
	d_pivot_sphere_position_end = TVector3(0,0,0);
	d_points.clear();
	
	QTextStream							stream(file_data);
	
	while(!stream.atEnd())
	{
		text = stream.readLine().simplified();
		text.replace(',',' ');
		
		if(text.length())
		{
			if(text.startsWith(QStringLiteral("temperature:"),Qt::CaseInsensitive))
			{
				index = text.lastIndexOf(':');
				
				if(!(index < 0))
				{
					text = text.mid(index+1);
					
					dval = text.toDouble(&status[0]);
					
					if(status[0])
					{
						d_temperature = dval;
					}
				}
			}
			else if(text.startsWith(QStringLiteral("probe_offset:"),Qt::CaseInsensitive))
			{
				index = text.lastIndexOf(':');
				
				if(!(index < 0))
				{
					text = text.mid(index+1);
					
					list = text.split(' ',QString::SkipEmptyParts);
					
					if(list.size() == 3)
					{
						pnt.x = list[0].toDouble(&status[0]);
						pnt.y = list[1].toDouble(&status[1]);
						pnt.z = list[2].toDouble(&status[2]);
						
						if(status[0] && status[1] && status[2])
						{
							d_probe_offset = pnt;
						}
					}
					else
					{
						d_last_error = QStringLiteral("Probe offset expecting three values");
						return false;
					}
					
				}
			}
			else if(text.startsWith(QStringLiteral("probe_error:"),Qt::CaseInsensitive))
			{
				index = text.lastIndexOf(':');
				
				if(!(index < 0))
				{
					text = text.mid(index+1);
					
					dval = text.toDouble(&status[0]);
					
					if(status[0])
					{
						d_probe_error = dval;
					}
				}
			}
			else if(text.startsWith(QStringLiteral("pivot_sphere_start:"),Qt::CaseInsensitive))
			{
				index = text.lastIndexOf(':');
				
				if(!(index < 0))
				{
					text = text.mid(index+1);
					
					list = text.split(' ',QString::SkipEmptyParts);
					
					if(list.size() == 3)
					{
						pnt.x = list[0].toDouble(&status[0]);
						pnt.y = list[1].toDouble(&status[1]);
						pnt.z = list[2].toDouble(&status[2]);
						
						if(status[0] && status[1] && status[2])
						{
							d_pivot_sphere_position_start = pnt;
						}
					}
					else
					{
						d_last_error = QStringLiteral("Pivot sphere initial position expecting three values");
						return false;
					}
					
				}
			}
			else if(text.startsWith(QStringLiteral("pivot_sphere_end:"),Qt::CaseInsensitive))
			{
				index = text.lastIndexOf(':');
				
				if(!(index < 0))
				{
					text = text.mid(index+1);
					
					list = text.split(' ',QString::SkipEmptyParts);
					
					if(list.size() == 3)
					{
						pnt.x = list[0].toDouble(&status[0]);
						pnt.y = list[1].toDouble(&status[1]);
						pnt.z = list[2].toDouble(&status[2]);
						
						if(status[0] && status[1] && status[2])
						{
							d_pivot_sphere_position_end = pnt;
						}
					}
					else
					{
						d_last_error = QStringLiteral("Pivot sphere final position expecting three values");
						return false;
					}
				}
			}
			else
			{
				list = text.split(' ',QString::SkipEmptyParts);
				
				if(list.size() == 3)
				{
					pnt.x = list[0].toDouble(&status[0]);
					pnt.y = list[1].toDouble(&status[1]);
					pnt.z = list[2].toDouble(&status[2]);
					
					if(status[0] && status[1] && status[2])
					{
						d_points.push_back(pnt);
					}
				}
				else
				{
					d_last_error = QStringLiteral("Gauge point expecting three values");
					return false;
				}
			}
		}
	};
		
	if(!this->Process_MCG_Data())
	{
		return false;
	}
	
	return true;
}

bool TMeasurement::Load_MCG_Data(
	const QString						&file_name)
{
	TXmlFile							xmlfile;
	int									version;
	
	try
	{
		xmlfile.Set_File_Name(file_name);
		xmlfile.Open("MCG_Data",&version);
		
		switch(version)
		{
			case 1:
				this->Load_MCG_V1(&xmlfile);
				break;
		}
		
		if(!this->Process_MCG_Data())
		{
			return false;
		}
		
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("error loading file '%1': %2").arg(file_name).arg(xml_exception.ErrorText());
		return false;
	}
	
	return true;
}

bool TMeasurement::Process_MCG_Data(void)
{
	std::vector<TVector3>::iterator		iter;
	std::vector<TMeasurementItem>::const_iterator mea_iter;
	double								elevation;
	double								azimuth;
	double								length;
	double								dev,sum_squared;
	TVector3							axis;
	TVector3							sphere_center;
	TBestFitSphere						best_fit_sphere;
	bool								init;

	d_measurement_items.clear();
	
	d_average_x_length_valid = false;
	d_average_y_length_valid = false;
	d_average_z_length_valid = false;
	
	d_average_xy_length_valid = false;
	d_average_yz_length_valid = false;
	d_average_zx_length_valid = false;
	
	d_squareness_xy_valid = false;
	d_squareness_yz_valid = false;
	d_squareness_zx_valid = false;
	
	if(d_points.size() < 4)
	{
		d_last_error = QStringLiteral("Insufficient number of measurement points.");
		return false;
	}

	if(best_fit_sphere.Fit(d_points.size(),&d_points[0]))
	{
		sphere_center = best_fit_sphere.Position();
		
	}
	else
	{
		d_last_error = QStringLiteral("Cannot fit sphere with measurement points");
		return false;
	}
	
	d_measurement_sphere_position = d_pivot_sphere_position_start;
	d_measurement_sphere_position += sphere_center;
	
	// adjust all points relative to calculated sphere center
	for(iter = d_points.begin();iter != d_points.end();++iter)
	{
		(*iter) -= sphere_center;
	}
	
	// calculate data from points
	this->Analyze_Points();
	
	
	for(iter = d_points.begin();iter != d_points.end();++iter)
	{
		if(this->Calculate_Azimuth_Elevation((*iter),&azimuth,&elevation,&axis))
		{
			this->Update_MCG_Measurement_Point(axis,azimuth,elevation,(*iter));
		}
	}
	
	this->Calculate_MCG_Measurement_Data();
	
	d_pattern_average = 0.0;
	d_pattern_stdev = 0.0;
	d_pattern_min_length = 0.0;
	d_pattern_max_length = 0.0;
	init = true;

	for(mea_iter = d_measurement_items.begin(); mea_iter != d_measurement_items.end();++mea_iter)
	{
		length = (*mea_iter).length;
		
		if(init)
		{
			d_pattern_min_length = length;
			d_pattern_max_length = length;
			
			init = false;
		}
		else
		{
			if(length < d_pattern_min_length)
			{
				d_pattern_min_length = length;
			}
			else if(length > d_pattern_max_length)
			{
				d_pattern_max_length = length;
			}
		}
		
		d_pattern_average += length;
	}
	
	if(d_measurement_items.size())
	{
		d_pattern_average /= static_cast<double>(d_measurement_items.size());
	}
	
	if(d_measurement_items.size() > 1)
	{
		sum_squared = 0.0;
		for(mea_iter = d_measurement_items.begin(); mea_iter != d_measurement_items.end();++mea_iter)
		{
			dev = (*mea_iter).length - d_pattern_average;
			
			sum_squared += (dev * dev);
		}
		
		d_pattern_stdev = sqrt(sum_squared / static_cast<double>(d_measurement_items.size() - 1));
	}

	
	return true;
}

void TMeasurement::Analyze_Points(void)
{
	std::vector<TVector3>::iterator		iter;
	std::vector<TVector3>				point_data_a;
	std::vector<TVector3>				point_data_b;
	std::vector<TVector3>				point_data;
	int									point_count;
	double								error;
	double								length_a,length_b;
	TVector3							vec;
	static const double					PARALLEL_ANGLE_TOLERANCE(25);
	static const double					PARALLEL_ANGLE_Z_TOLERANCE(65.0);	// anything within 65 deg of Z allowed (no direct Z measurements)
	static const double					COSINE45(0.7071067811865475727);
	static const double					ZERO_EPSILON(0.0001);
	
	// Get points along X axis
	point_data_a = this->Get_Parallel_Points(TVector3(1,0,0),PARALLEL_ANGLE_TOLERANCE);
	point_data_b = this->Get_Parallel_Points(TVector3(-1,0,0),PARALLEL_ANGLE_TOLERANCE);
	
	point_data.clear();
	point_data.reserve(point_data_a.size() + point_data_b.size());
	point_data.insert(point_data.end(), point_data_a.begin(), point_data_a.end());
	point_data.insert(point_data.end(), point_data_b.begin(), point_data_b.end());
	
	point_count = static_cast<int>(point_data.size());
	
	d_average_x_length = 0.0;
	
	if(point_count)
	{
		for(iter = point_data.begin();iter < point_data.end();++iter)
		{
			d_average_x_length += (*iter).Length();
		}
		
		d_average_x_length /= static_cast<double>(point_count);
		d_average_x_length_valid = true;
	}
	
	// Get points along Y axis
	point_data_a = this->Get_Parallel_Points(TVector3(0,1,0),PARALLEL_ANGLE_TOLERANCE);
	point_data_b = this->Get_Parallel_Points(TVector3(0,-1,0),PARALLEL_ANGLE_TOLERANCE);
	
	point_data.clear();
	point_data.reserve(point_data_a.size() + point_data_b.size());
	point_data.insert(point_data.end(), point_data_a.begin(), point_data_a.end());
	point_data.insert(point_data.end(), point_data_b.begin(), point_data_b.end());
	
	point_count = static_cast<int>(point_data.size());
	
	d_average_y_length = 0.0;
	
	if(point_count)
	{
		for(iter = point_data.begin();iter < point_data.end();++iter)
		{
			d_average_y_length += (*iter).Length();
		}
		
		d_average_y_length /= static_cast<double>(point_count);
		d_average_y_length_valid = true;
	}
	
	// Get points in XY plane
	point_data = this->Get_Plane_XY_Points(PARALLEL_ANGLE_TOLERANCE);
	
	point_count = static_cast<int>(point_data.size());

	d_average_xy_length = 0.0;
	
	if(point_count > 3)
	{
		for(iter = point_data.begin();iter < point_data.end();++iter)
		{
			d_average_xy_length += (*iter).Length();
		}
	
		d_average_xy_length /= static_cast<double>(point_count);
		d_average_xy_length_valid = true;
	}
	
	// Get points in YZ plane
	point_data = this->Get_Plane_YZ_Points(PARALLEL_ANGLE_TOLERANCE);
	
	point_count = static_cast<int>(point_data.size());
	
	d_average_yz_length = 0.0;
	
	if(point_count > 3)
	{
		for(iter = point_data.begin();iter < point_data.end();++iter)
		{
			d_average_yz_length += (*iter).Length();
		}
		
		d_average_yz_length /= static_cast<double>(point_count);
		d_average_yz_length_valid = true;
	}
	
	// Get points in YZ plane
	point_data = this->Get_Plane_ZX_Points(PARALLEL_ANGLE_TOLERANCE);
	
	point_count = static_cast<int>(point_data.size());
	
	d_average_zx_length = 0.0;
	
	if(point_count > 3)
	{
		for(iter = point_data.begin();iter < point_data.end();++iter)
		{
			d_average_zx_length += (*iter).Length();
		}
		
		d_average_zx_length /= static_cast<double>(point_count);
		d_average_zx_length_valid = true;
	}
	
	
	// Estimate Z scale length.  Must have valid XY data.
	if(d_average_xy_length_valid)
	{
		// Z Length calculation done by extrapolation from all measurements
		//  above or below the XY plane.
		// Z error increases by the square of the vector K component of the direction.
		
		point_data_a = this->Get_Parallel_Points(TVector3(0,0,1),PARALLEL_ANGLE_Z_TOLERANCE);
		point_data_b = this->Get_Parallel_Points(TVector3(0,0,-1),PARALLEL_ANGLE_Z_TOLERANCE);
		
		point_data.clear();
		point_data.reserve(point_data_a.size() + point_data_b.size());
		point_data.insert(point_data.end(), point_data_a.begin(), point_data_a.end());
		point_data.insert(point_data.end(), point_data_b.begin(), point_data_b.end());
		
		point_count = static_cast<int>(point_data.size());
		
		d_average_z_length = 0.0;
		
		if(point_count > 7)
		{
			
			for(iter = point_data.begin();iter < point_data.end();++iter)
			{
				error = (*iter).Length();
				
				if(error > ZERO_EPSILON)
				{
					vec = (*iter);
					vec.Normal();
					
					error -= d_average_xy_length;
					
					d_average_z_length += d_average_xy_length + error / (vec.k * vec.k);
				}
				else
				{
					point_count--;
				}
			}

			if(point_count > 0)
			{
				d_average_z_length /= static_cast<double>(point_count);
				d_average_z_length_valid = true;
			}
		}
	}
	
	// Get XY squareness.  Must have valid XY data.
	if(d_average_xy_length_valid)
	{
		point_data_a = this->Get_Parallel_Points(TVector3(COSINE45,COSINE45,0),PARALLEL_ANGLE_TOLERANCE);
		point_data_b = this->Get_Parallel_Points(TVector3(-COSINE45,-COSINE45,0),PARALLEL_ANGLE_TOLERANCE);
		
		point_data.clear();
		point_data.reserve(point_data_a.size() + point_data_b.size());
		point_data.insert(point_data.end(), point_data_a.begin(), point_data_a.end());
		point_data.insert(point_data.end(), point_data_b.begin(), point_data_b.end());
		
		point_count = static_cast<int>(point_data.size());
		
		d_squareness_xy_error = 0;
		
		length_a = 0.0;
		
		if(point_count)
		{
			for(iter = point_data.begin();iter < point_data.end();++iter)
			{
				vec = (*iter);
				vec.k = 0.0;

				if(vec.Length() > ZERO_EPSILON)
				{
					vec.Normal();
					
					error = (*iter).Length() - d_average_xy_length;
					
					error /= fabs(TVector3(COSINE45,COSINE45,0) ^ vec);
					
					length_a += (d_average_xy_length + error);
				}
				else
				{
					point_count--;
				}
			}
			
			if(point_count > 0)
			{
				length_a /= static_cast<double>(point_count);

				point_data_a = this->Get_Parallel_Points(TVector3(-COSINE45,COSINE45,0),PARALLEL_ANGLE_TOLERANCE);
				point_data_b = this->Get_Parallel_Points(TVector3(COSINE45,-COSINE45,0),PARALLEL_ANGLE_TOLERANCE);
				
				point_data.clear();
				point_data.reserve(point_data_a.size() + point_data_b.size());
				point_data.insert(point_data.end(), point_data_a.begin(), point_data_a.end());
				point_data.insert(point_data.end(), point_data_b.begin(), point_data_b.end());
				
				point_count = static_cast<int>(point_data.size());
				
				length_b = 0.0;
				
				if(point_count)
				{
					for(iter = point_data.begin();iter < point_data.end();++iter)
					{
						vec = (*iter);
						vec.k = 0.0;
						
						if(vec.Length() > ZERO_EPSILON)
						{
							vec.Normal();
							
							error = (*iter).Length() - d_average_xy_length;
							
							error /= fabs(TVector3(-COSINE45,COSINE45,0) ^ vec);
							
							length_b += (d_average_xy_length + error);
						}
						else
						{
							point_count--;
						}
					}
					
					if(point_count > 0)
					{
						length_b /= static_cast<double>(point_count);
						
						d_squareness_xy_valid = true;
						d_squareness_xy_error = 2000 * (length_a - length_b)/(length_a + length_b);
					}
				}
			}
			
		}
	}
	
	// Squareness YZ
	// This requires YZ Length
	
	if(d_average_yz_length_valid)
	{
		point_data_a = this->Get_Parallel_Points(TVector3(0,COSINE45,COSINE45),PARALLEL_ANGLE_TOLERANCE);
		point_data_b = this->Get_Parallel_Points(TVector3(0,-COSINE45,-COSINE45),PARALLEL_ANGLE_TOLERANCE);
		
		point_data.clear();
		point_data.reserve(point_data_a.size() + point_data_b.size());
		point_data.insert(point_data.end(), point_data_a.begin(), point_data_a.end());
		point_data.insert(point_data.end(), point_data_b.begin(), point_data_b.end());
		
		point_count = static_cast<int>(point_data.size());
		
		d_squareness_yz_error = 0;
		
		length_a = 0.0;
		
		if(point_count)
		{
			for(iter = point_data.begin();iter < point_data.end();++iter)
			{
				vec = (*iter);
				vec.i = 0.0;
				
				if(vec.Length() > ZERO_EPSILON)
				{
					vec.Normal();
					
					error = (*iter).Length() - d_average_yz_length;
					
					error /= fabs(TVector3(0,COSINE45,COSINE45) ^ vec);
					
					length_a += (d_average_yz_length + error);
				}
				else
				{
					point_count--;
				}
			}
			
			if(point_count > 0)
			{
				length_a /= static_cast<double>(point_count);
				
				point_data_a = this->Get_Parallel_Points(TVector3(0,-COSINE45,COSINE45),PARALLEL_ANGLE_TOLERANCE);
				point_data_b = this->Get_Parallel_Points(TVector3(0,COSINE45,-COSINE45),PARALLEL_ANGLE_TOLERANCE);
				
				point_data.clear();
				point_data.reserve(point_data_a.size() + point_data_b.size());
				point_data.insert(point_data.end(), point_data_a.begin(), point_data_a.end());
				point_data.insert(point_data.end(), point_data_b.begin(), point_data_b.end());
				
				point_count = static_cast<int>(point_data.size());
				
				length_b = 0.0;
				
				if(point_count)
				{
					for(iter = point_data.begin();iter < point_data.end();++iter)
					{
						vec = (*iter);
						vec.i = 0.0;
						
						if(vec.Length() > ZERO_EPSILON)
						{
							vec.Normal();
							
							error = (*iter).Length() - d_average_yz_length;
							
							error /= fabs(TVector3(0,-COSINE45,COSINE45) ^ vec);
							
							length_b += (d_average_yz_length + error);
						}
						else
						{
							point_count--;
						}
					}
					
					if(point_count > 0)
					{
						length_b /= static_cast<double>(point_count);
						
						d_squareness_yz_valid = true;
						d_squareness_yz_error = 2000 * (length_a - length_b)/(length_a + length_b);
					}
				}
			}
		}
	}
	
	// Squareness ZX
	// This requires ZX Length
	
	if(d_average_zx_length_valid)
	{
		point_data_a = this->Get_Parallel_Points(TVector3(COSINE45,0,COSINE45),PARALLEL_ANGLE_TOLERANCE);
		point_data_b = this->Get_Parallel_Points(TVector3(-COSINE45,0,-COSINE45),PARALLEL_ANGLE_TOLERANCE);
		
		point_data.clear();
		point_data.reserve(point_data_a.size() + point_data_b.size());
		point_data.insert(point_data.end(), point_data_a.begin(), point_data_a.end());
		point_data.insert(point_data.end(), point_data_b.begin(), point_data_b.end());
		
		point_count = static_cast<int>(point_data.size());
		
		d_squareness_zx_error = 0;
		
		length_a = 0.0;
		
		if(point_count)
		{
			for(iter = point_data.begin();iter < point_data.end();++iter)
			{
				vec = (*iter);
				vec.j = 0.0;
				
				if(vec.Length() > ZERO_EPSILON)
				{
					vec.Normal();
					
					error = (*iter).Length() - d_average_zx_length;
					
					error /= fabs(TVector3(COSINE45,0,COSINE45) ^ vec);
					
					length_a += (d_average_zx_length + error);
				}
				else
				{
					point_count--;
				}
			}
			
			if(point_count > 0)
			{
				length_a /= static_cast<double>(point_count);
				
				point_data_a = this->Get_Parallel_Points(TVector3(-COSINE45,0,COSINE45),PARALLEL_ANGLE_TOLERANCE);
				point_data_b = this->Get_Parallel_Points(TVector3(COSINE45,0,-COSINE45),PARALLEL_ANGLE_TOLERANCE);
				
				point_data.clear();
				point_data.reserve(point_data_a.size() + point_data_b.size());
				point_data.insert(point_data.end(), point_data_a.begin(), point_data_a.end());
				point_data.insert(point_data.end(), point_data_b.begin(), point_data_b.end());
				
				point_count = static_cast<int>(point_data.size());
				
				length_b = 0.0;
				
				if(point_count)
				{
					for(iter = point_data.begin();iter < point_data.end();++iter)
					{
						vec = (*iter);
						vec.j = 0.0;
						
						if(vec.Length() > ZERO_EPSILON)
						{
							vec.Normal();
							
							error = (*iter).Length() - d_average_zx_length;
							
							error /= fabs(TVector3(-COSINE45,0,COSINE45) ^ vec);
							
							length_b += (d_average_zx_length + error);
						}
						else
						{
							point_count--;
						}
					}
					
					if(point_count > 0)
					{
						length_b /= static_cast<double>(point_count);
						
						d_squareness_zx_valid = true;
						d_squareness_zx_error = 2000 * (length_a - length_b)/(length_a + length_b);
					}
				}
			}
		}
	}
}


bool TMeasurement::Calculate_Azimuth_Elevation(
	const TVector3						&pnt,
	double								* const azimuth,
	double								* const elevation,
	TVector3							* const axis) const
{
	TVector3							vec;
	TVector3							vec_xy;
	double								angle;
	static const double					ZERO_EPSILON(0.0001);
	static const double					TWO_PI(6.28318530717959);
	static const double					RADIAN_TO_DEGREE(57.29577951308233);
	static const double					ZERO_ANGLE_TOLERANCE(0.00872664626);
	
	assert(azimuth);
	assert(elevation);
	
	vec = pnt;
	
	if(vec.Length() < ZERO_EPSILON)
	{
		return false;
	}
	
	vec.Normal();
	
	*axis = vec;
	
	vec_xy = vec;
	vec_xy.z = 0;
	
	if(vec_xy.Length() < ZERO_EPSILON)
	{
		*azimuth = 0;
		
		if(vec.z < 0)
		{
			*elevation = -90.0;
		}
		else
		{
			*elevation = 90.0;
		}
		
		return true;
	}
	
	vec_xy.Normal();
	
	angle = acos(vec_xy.i);
	if(vec_xy.j < 0)
	{
		angle = TWO_PI - angle;
	}
	
	// check if angle is close to 0
	// preference is 0 and not 360
	if((TWO_PI - angle) < ZERO_ANGLE_TOLERANCE)
	{
		angle -= TWO_PI;
	}
	
	
	*azimuth = angle * RADIAN_TO_DEGREE;
	
	angle = asin(vec.k);
	*elevation = angle * RADIAN_TO_DEGREE;
	
	return true;
}

void TMeasurement::Update_MCG_Measurement_Point(
	const TVector3						&axis,
	const double						&azimuth,
	const double						&elevation,
	const TVector3						&pnt)
{
	std::vector<TMeasurementItem>::iterator iter;
	TMeasurementItem						measurement;
	TVector3							vec;
	static const double					AXIS_LIMIT(0.999847695156391);	// within 1 degree
	
	vec = pnt;
	
	if(!vec.Length())
	{
		return;
	}
	
	vec.Normal();
	
	// find existing entry based on the axis
	for(iter = d_measurement_items.begin(); iter != d_measurement_items.end();++iter)
	{
		if((vec ^ (*iter).axis) > AXIS_LIMIT)
		{
			(*iter).points.push_back(pnt);
			return;
		}
	}

	measurement.elevation = elevation;
	measurement.azimuth = azimuth;
	measurement.axis = axis;
	measurement.points.push_back(pnt);
	
	d_measurement_items.push_back(measurement);
}

void TMeasurement::Calculate_MCG_Measurement_Data(void)
{
	std::vector<TMeasurementItem>::iterator iter;
	std::vector<TVector3>::iterator		pnt_iter;
	bool								init(true);
	double								min(0);
	double								max(0);
	double								mean(0);
	double								dev(0);
	double								sum_squared(0);
	double								length(0);

	for(iter = d_measurement_items.begin(); iter != d_measurement_items.end();++iter)
	{
		init = true;
		mean = 0.0;
		for(pnt_iter = (*iter).points.begin(); pnt_iter != (*iter).points.end();++pnt_iter)
		{
			length = (*pnt_iter).Length();
			
			if(init)
			{
				min = max = length;
				init = false;
			}
			else
			{
				if(length < min)
				{
					min = length;
				}
				else if(length > max)
				{
					max = length;
				}
			}
			
			mean += length;
		}
		
		if((*iter).points.size())
		{
			mean /= static_cast<double>((*iter).points.size());
		}
		
		(*iter).length = mean;
		(*iter).length_min = min;
		(*iter).length_max = max;
		
		if((*iter).points.size() > 1)
		{
			sum_squared = 0.0;
			
			for(pnt_iter = (*iter).points.begin(); pnt_iter != (*iter).points.end();++pnt_iter)
			{
				length = (*pnt_iter).Length();
				
				dev = length - mean;
				sum_squared += (dev * dev);
			}
			
			(*iter).length_stddev = sqrt(sum_squared / static_cast<double>((*iter).points.size() - 1));
		}
		else
		{
			(*iter).length_stddev = 0.0;
		}
	}
}


void TMeasurement::Load_MCG_V1(
	TXmlFile							* const xml_file)
{
	QDomElement							e0;
	QDomElement							e1;
	QString								text;
	QStringList							list;
	TVector3							point;
	
	assert(xml_file);
	
	d_points.clear();
	
	d_measurement_sphere_position = TVector3(0.0,0.0,0.0);
	d_pivot_sphere_position_start = TVector3(0.0,0.0,0.0);
	d_pivot_sphere_position_end = TVector3(0.0,0.0,0.0);
	d_probe_offset = TVector3(0.0,0.0,0.0);
	d_probe_error = 0.0;
	d_temperature = 20.0;
	
	// Date_Time
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Date_Time"));
	d_date_time.setMSecsSinceEpoch(xml_file->Read_Text_Node(&e0).toLongLong());
	
	// Sphere_Position
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Sphere_Position"));
	text = xml_file->Read_Text_Node(&e0);
	list = text.split(',');
	
	if(list.size() == 3)
	{
		d_measurement_sphere_position.x = list[0].toDouble();
		d_measurement_sphere_position.y = list[1].toDouble();
		d_measurement_sphere_position.z = list[2].toDouble();
	}
	else
	{
		throw TXmlFileException(QStringLiteral("Sphere_Position data invalid"));
	}
	
	// Pivot_Sphere_Start_Position
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Pivot_Sphere_Start_Position"));
	text = xml_file->Read_Text_Node(&e0);
	list = text.split(',');
	
	if(list.size() == 3)
	{
		d_pivot_sphere_position_start.x = list[0].toDouble();
		d_pivot_sphere_position_start.y = list[1].toDouble();
		d_pivot_sphere_position_start.z = list[2].toDouble();
	}
	else
	{
		throw TXmlFileException(QStringLiteral("Pivot_Sphere_Start_Position data invalid"));
	}

	// Pivot_Sphere_End_Position
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Pivot_Sphere_End_Position"));
	text = xml_file->Read_Text_Node(&e0);
	list = text.split(',');
	
	if(list.size() == 3)
	{
		d_pivot_sphere_position_end.x = list[0].toDouble();
		d_pivot_sphere_position_end.y = list[1].toDouble();
		d_pivot_sphere_position_end.z = list[2].toDouble();
	}
	else
	{
		throw TXmlFileException(QStringLiteral("Pivot_Sphere_End_Position data invalid"));
	}
	
	// Probe_Offset
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Probe_Offset"));
	text = xml_file->Read_Text_Node(&e0);
	list = text.split(',');
	
	if(list.size() == 3)
	{
		d_probe_offset.x = list[0].toDouble();
		d_probe_offset.y = list[1].toDouble();
		d_probe_offset.z = list[2].toDouble();
	}
	else
	{
		throw TXmlFileException(QStringLiteral("Probe_Offset data invalid"));
	}
	
	// Probe_Error
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Probe_Error"));
	d_probe_error = xml_file->Read_Text_Node(&e0).toDouble();
	
	// Temperature
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Temperature"));
	d_temperature = xml_file->Read_Text_Node(&e0).toDouble();
	
	// Tolerance.  May not exist in file
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Tolerance_Expression"),true);
	if(!e0.isNull())
	{
		d_tolerance_expression = xml_file->Read_Text_Node(&e0);
	}
	else
	{
		d_tolerance_expression.clear();
	}
	
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Gauge_Points"));
	e1 = xml_file->Get_Node(&e0,QStringLiteral("Point_XYZ"),true);
	
	while(!e1.isNull())
	{
		// point
		text = xml_file->Read_Text_Node(&e1);
		list = text.split(',');
		
		if(list.size() == 3)
		{
			point.x = list[0].toDouble();
			point.y = list[1].toDouble();
			point.z = list[2].toDouble();
			
			d_points.push_back(point);
		}
		else
		{
			throw TXmlFileException(QStringLiteral("Point_XYZ data invalid"));
		}

		e1 = xml_file->Get_Sibling_Node(&e1,QStringLiteral("Point_XYZ"),true);
	};
}

void TMeasurement::Import_Measurement_V1(
	TXmlFile							* const xml_file)
{
	QDomElement							e0;
	QDomElement							e1;
	QString								text;
	QStringList							list;
	TVector3							point;
	double								units_conversion(1.0);
	
	assert(xml_file);
	
	d_points.clear();
	
	// defaults.  Entries may not exist in file
	d_pivot_sphere_position_start = TVector3(0,0,0);
	d_pivot_sphere_position_end = TVector3(0,0,0);
	d_measurement_sphere_position = TVector3(0,0,0);
	d_probe_error = 0.0;
	d_temperature = 20.0;
	d_probe_offset = TVector3(0,0,0);
	
	text = xml_file->Read_Attribute(xml_file->Root_Node(),QStringLiteral("Units"),true);
	// assume metric by default
	if(text.contains(QStringLiteral("IN"),Qt::CaseInsensitive))
	{
		units_conversion = 25.4;
	}
	
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Environment"),true);
	if(!e0.isNull())
	{
		e1 = xml_file->Get_Node(&e0,QStringLiteral("Temperature"));
		d_temperature = xml_file->Read_Text_Node(&e1).toDouble();
	}
	
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Probe"),true);
	if(!e0.isNull())
	{
		e1 = xml_file->Get_Node(&e0,QStringLiteral("Probe_Offset_XYZ"));
		text = xml_file->Read_Text_Node(&e1);
		
		list = text.split(',');
		
		if(list.size() == 3)
		{
			d_probe_offset.x = list[0].toDouble();
			d_probe_offset.y = list[1].toDouble();
			d_probe_offset.z = list[2].toDouble();
			
			d_probe_offset *= units_conversion;
		}
		else
		{
			throw TXmlFileException(QStringLiteral("Probe_Offset_XYZ data invalid"));
		}
		
		e1 = xml_file->Get_Node(&e0,QStringLiteral("Probe_Error"));
		d_probe_error = xml_file->Read_Text_Node(&e1).toDouble();
		d_probe_error *= units_conversion;
	}
	
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Pivot_Sphere_Initial"),true);
	if(!e0.isNull())
	{
		e1 = xml_file->Get_Node(&e0,QStringLiteral("Location_XYZ"));
		text = xml_file->Read_Text_Node(&e1);
		
		list = text.split(',');
		
		if(list.size() == 3)
		{
			d_pivot_sphere_position_start.x = list[0].toDouble();
			d_pivot_sphere_position_start.y = list[1].toDouble();
			d_pivot_sphere_position_start.z = list[2].toDouble();
			
			d_pivot_sphere_position_start *= units_conversion;
		}
		else
		{
			throw TXmlFileException(QStringLiteral("Pivot Sphere Initial Location_XYZ data invalid"));
		}
	}
	
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Pivot_Sphere_Final"),true);
	if(!e0.isNull())
	{
		text = xml_file->Read_Text_Node(&e1);
		
		list = text.split(',');
		
		if(list.size() == 3)
		{
			d_pivot_sphere_position_end.x = list[0].toDouble();
			d_pivot_sphere_position_end.y = list[1].toDouble();
			d_pivot_sphere_position_end.z = list[2].toDouble();
			
			d_pivot_sphere_position_end *= units_conversion;
		}
		else
		{
			throw TXmlFileException(QStringLiteral("Pivot Sphere Final Location_XYZ data invalid"));
		}
	}

	
	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Gauge_Points"));
	e1 = xml_file->Get_Node(&e0,QStringLiteral("Point_XYZ"));
	while(!e1.isNull())
	{
		text = xml_file->Read_Text_Node(&e1);
		
		list = text.split(',');
		
		if(list.size() == 3)
		{
			point.x = list[0].toDouble();
			point.y = list[1].toDouble();
			point.z = list[2].toDouble();
			
			point *= units_conversion;
			
			d_points.push_back(point);
		}
		else
		{
			throw TXmlFileException(QStringLiteral("Gauge_Points Point_XYZ data invalid"));
		}

		e1 = xml_file->Get_Sibling_Node(&e1,QStringLiteral("Point_XYZ"),true);
	}
}


std::vector<TVector3> TMeasurement::Get_Parallel_Points(
	const TVector3 						&axis,
	const double 						&angle_tolerance) const
{
	double								tolerance;
	double								product;
	TVector3							vec;
	std::vector<TVector3>::const_iterator iter;
	std::vector<TVector3>				return_points;
	static const double					ZERO_EPSILON(0.0001);
	
	tolerance = cos(fabs(angle_tolerance) / 57.29577951308232311);
	
	for(iter = d_points.begin();iter != d_points.end();++iter)
	{
		vec = (*iter);
		
		if(vec.Length() > ZERO_EPSILON)
		{
			vec.Normal();
			product = axis ^ vec;
			
			if(product > tolerance)
			{
				return_points.push_back(*iter);
			}
		}
	}
	
	return return_points;
}

std::vector<TVector3> TMeasurement::Get_Plane_XY_Points(
	const double 						&angle_tolerance) const
{
	double								tolerance;
	double								product;
	TVector3							plane_axis(0,0,1);
	TVector3							vec;
	std::vector<TVector3>::const_iterator iter;
	std::vector<TVector3>				return_points;
	static const double					ZERO_EPSILON(0.0001);

	tolerance = sin(fabs(angle_tolerance) / 57.29577951308232311);
	
	for(iter = d_points.begin();iter != d_points.end();++iter)
	{
		vec = (*iter);
		
		if(vec.Length() > ZERO_EPSILON)
		{
			vec.Normal();
			
			product = plane_axis ^ vec;
			
			if(product < tolerance && product > (-tolerance))
			{
				return_points.push_back(*iter);
			}
		}
		
	}
	
	return return_points;
}

std::vector<TVector3> TMeasurement::Get_Plane_YZ_Points(
	const double 						&angle_tolerance) const
{
	double								tolerance;
	double								product;
	TVector3							plane_axis(1,0,0);
	TVector3							vec;
	std::vector<TVector3>::const_iterator iter;
	std::vector<TVector3>				return_points;
	static const double					ZERO_EPSILON(0.0001);

	tolerance = sin(fabs(angle_tolerance) / 57.29577951308232311);
	
	for(iter = d_points.begin();iter != d_points.end();++iter)
	{
		vec = (*iter);
		
		if(vec.Length() > ZERO_EPSILON)
		{
			vec.Normal();
			
			product = plane_axis ^ vec;
			
			if(product < tolerance && product > (-tolerance))
			{
				return_points.push_back(*iter);
			}
		}
		
	}
	
	return return_points;
}

std::vector<TVector3> TMeasurement::Get_Plane_ZX_Points(
	const double 						&angle_tolerance) const
{
	double								tolerance;
	double								product;
	TVector3							plane_axis(0,1,0);
	TVector3							vec;
	std::vector<TVector3>::const_iterator iter;
	std::vector<TVector3>				return_points;
	static const double					ZERO_EPSILON(0.0001);

	tolerance = sin(fabs(angle_tolerance) / 57.29577951308232311);
	
	for(iter = d_points.begin();iter != d_points.end();++iter)
	{
		vec = (*iter);
		
		if(vec.Length() > ZERO_EPSILON)
		{
			vec.Normal();
			
			product = plane_axis ^ vec;
			
			if(product < tolerance && product > (-tolerance))
			{
				return_points.push_back(*iter);
			}
		}
		
	}
	
	return return_points;
}


