/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef MEASUREMENTHEADERFILE
#define MEASUREMENTHEADERFILE

#include <QString>
#include <QList>
#include <QUrl>
#include <QDateTime>
#include <vector>

#include "../../core/vector2.h"
#include "../../core/vector3.h"

class TXmlFile;

class TMeasurement
{
public:

struct TMeasurementItem
{
	double								azimuth;
	double								elevation;
	TVector3							axis;
	std::vector<TVector3>				points;
	double								length;
	double								length_min;
	double								length_max;
	double								length_stddev;
};

// CREATORS
	TMeasurement(void);
	~TMeasurement(void);

// ACCESSORS
	QString Last_Error(void) const {return d_last_error;}

	double Length_Min(void) const {return d_pattern_min_length;}
	double Length_Max(void) const {return d_pattern_max_length;}
	double Length_Average(void) const {return d_pattern_average;}
	double Length_Median(void) const {return (d_pattern_max_length + d_pattern_min_length) / 2.0;}
	double Length_Stdev(void) const {return d_pattern_stdev;}
	double Temperature(void) const {return d_temperature;}
	
	TVector3 Probe_Offset(void) const {return d_probe_offset;}
	double Probe_Error(void) const {return d_probe_error;}
	
	bool Average_X_Valid(void) const {return d_average_x_length_valid;}
	bool Average_Y_Valid(void) const {return d_average_y_length_valid;}
	bool Average_Z_Valid(void) const {return d_average_z_length_valid;}
	double Average_X_Length(void) const {return d_average_x_length;}
	double Average_Y_Length(void) const {return d_average_y_length;}
	double Average_Z_Length(void) const {return d_average_z_length;}
	
	bool Average_XY_Valid(void) const {return d_average_xy_length_valid;}
	bool Average_YZ_Valid(void) const {return d_average_yz_length_valid;}
	bool Average_ZX_Valid(void) const {return d_average_zx_length_valid;}
	double Average_XY_Length(void) const {return d_average_xy_length;}
	double Average_YZ_Length(void) const {return d_average_yz_length;}
	double Average_ZX_Length(void) const {return d_average_zx_length;}
	
	bool Squareness_XY_Valid(void) const {return d_squareness_xy_valid;}
	bool Squareness_YZ_Valid(void) const {return d_squareness_yz_valid;}
	bool Squareness_ZX_Valid(void) const {return d_squareness_zx_valid;}
	double Squareness_XY_Error(void) const {return d_squareness_xy_error;}
	double Squareness_YZ_Error(void) const {return d_squareness_yz_error;}
	double Squareness_ZX_Error(void) const {return d_squareness_zx_error;}

	TVector3 MCG_Sphere_Center(void) const {return d_measurement_sphere_position;}
	TVector3 MCG_Sphere_Center_Start(void) const {return d_pivot_sphere_position_start;}
	TVector3 MCG_Sphere_Center_End(void) const {return d_pivot_sphere_position_end;}
	
	QString Tolerance_Expression(void) const {return d_tolerance_expression;}
	void Set_Tolerance_Expression(const QString &s) {d_tolerance_expression = s;}
	
	TVector3 Volume_Corner_Min(void) const;
	TVector3 Volume_Corner_Max(void) const;

	int Measurement_Count(void) const {return static_cast<int>(d_measurement_items.size());}
	TMeasurementItem MCG_Measurement(const int index) const;
	
	std::vector<TMeasurementItem> MCG_Pattern(void) const {return d_measurement_items;}	
	QDateTime Date_Time(void) const {return d_date_time;}
	
	bool Save_MCG_Data(const QString &file_name) const;
	bool Export_MCG_Measurements(const QString &file_name,const double &tolerance) const;

// MANIPULATORS
	void Clear(void);
	
	bool Import_XML_Measurement(const QString &file_name);
	bool Import_DMO_Measurement(const QString &file_name);
	bool Import_Text_Measurement(const QString &file_name);
	
	bool Load_MCG_Data(const QString &file_name);

private:
	std::vector<TMeasurementItem>		d_measurement_items;
	std::vector<TVector3>				d_points;
	double								d_temperature;
	double								d_probe_error;
	TVector3							d_probe_offset;
	TVector3							d_pivot_sphere_position_start;
	TVector3							d_pivot_sphere_position_end;
	TVector3							d_measurement_sphere_position;
	QString								d_tolerance_expression;
	
	bool								d_average_x_length_valid;
	bool								d_average_y_length_valid;
	bool								d_average_z_length_valid;
	
	bool								d_average_xy_length_valid;
	bool								d_average_yz_length_valid;
	bool								d_average_zx_length_valid;
	
	bool								d_squareness_xy_valid;
	bool								d_squareness_yz_valid;
	bool								d_squareness_zx_valid;
	
	double								d_average_x_length;
	double								d_average_y_length;
	double								d_average_z_length;
	
	double								d_average_xy_length;
	double								d_average_yz_length;
	double								d_average_zx_length;
	
	double								d_squareness_xy_error;
	double								d_squareness_yz_error;
	double								d_squareness_zx_error;

	double								d_pattern_average;
	double								d_pattern_stdev;
	double								d_pattern_min_length;
	double								d_pattern_max_length;
	
	QDateTime							d_date_time;
	
	mutable QString						d_last_error;
	
	void Import_Measurement_V1(TXmlFile * const xml_file);
	void Load_MCG_V1(TXmlFile * const xml_file);

	bool Calculate_Azimuth_Elevation(const TVector3 &pnt,double * const azimuth,double * const elevation,TVector3 * const axis) const;
	void Update_MCG_Measurement_Point(const TVector3 &axis, const double &azimuth, const double &elevation,const TVector3 &pnt);
	void Calculate_MCG_Measurement_Data(void);
	
	bool Process_MCG_Data(void);
	void Analyze_Points(void);
	
	std::vector<TVector3> Get_Parallel_Points(const TVector3 &Axis,const double &angle_tolerance) const;
	std::vector<TVector3> Get_Plane_XY_Points(const double &angle_tolerance) const;
	std::vector<TVector3> Get_Plane_YZ_Points(const double &angle_tolerance) const;
	std::vector<TVector3> Get_Plane_ZX_Points(const double &angle_tolerance) const;


// NOT IMPLEMENTED
	TMeasurement(const TMeasurement&);
	TMeasurement& operator=(const TMeasurement&);
};

#endif
