/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QGridLayout>
#include <QLabel>
#include <QSizePolicy>
#include <QToolButton>
#include <QFrame>
#include <QSpacerItem>
#include <QLabel>
#include <QIcon>

#include "../../core/mat4.h"
#include "../../core/referencesystem.h"
#include "../../core/openglgeometryfactory.h"

#include "measurement.h"
#include "opengldisplay.h"

#include "opengldisplaywidget.h"

TOpenGLDisplayWidget::TOpenGLDisplayWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QFrame(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QSizePolicy							size_policy(QSizePolicy::Maximum,QSizePolicy::Preferred);
	QSpacerItem							*widget_vspacer;
	QString								gl_error_text;
	QString								gl_version_text;
	QString								gl_extensions_text;
	int									gl_major_version;
	
	// check version of OpenGL.  Display will not work if opengl version is less than 2
	gl_major_version = TOpenGLWidget::Get_OpenGL_Version_Major(&gl_version_text,&gl_extensions_text);
	
	size_policy.setHorizontalStretch(0);
	size_policy.setVerticalStretch(0);
	
	this->setFrameShape(QFrame::Box);
	this->setStyleSheet(QStringLiteral("background-color: rgb(255, 255, 255);"));
	
	widget_layout = new QGridLayout(this);

	if(gl_major_version > 1)
	{
		d_opengl_display = new TOpenGLDisplay(this);
		widget_layout->addWidget(d_opengl_display,0,0,6,1);
		
		d_gldisplay_error_label = 0;
	}
	else
	{
		d_gldisplay_error_label = new QLabel(this);
		d_gldisplay_error_label->setWordWrap(true);
		d_gldisplay_error_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
		widget_layout->addWidget(d_gldisplay_error_label,0,0,6,1);
		
		d_opengl_display = 0;
	}


	d_scale_to_fit_button = new QToolButton(this);
	d_scale_to_fit_button->setMaximumSize(24,24);
	d_scale_to_fit_button->setCheckable(false);
	widget_layout->addWidget(d_scale_to_fit_button,0,1,1,1);

	d_pan_button = new QToolButton(this);
	d_pan_button->setMaximumSize(24,24);
	d_pan_button->setCheckable(true);
	widget_layout->addWidget(d_pan_button,1,1,1,1);

	d_rotate_2d_button = new QToolButton(this);
	d_rotate_2d_button->setMaximumSize(24,24);
	d_rotate_2d_button->setCheckable(true);
	widget_layout->addWidget(d_rotate_2d_button,2,1,1,1);

	d_rotate_3d_button = new QToolButton(this);
	d_rotate_3d_button->setMaximumSize(24,24);
	d_rotate_3d_button->setCheckable(true);
	widget_layout->addWidget(d_rotate_3d_button,3,1,1,1);
	
	widget_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	widget_layout->addItem(widget_vspacer,4,1,1,1);

	d_scale_to_fit_button->setText(QStringLiteral("..."));
	d_pan_button->setText(QStringLiteral("..."));
	d_rotate_2d_button->setText(QStringLiteral("..."));
	d_rotate_3d_button->setText(QStringLiteral("..."));
	
	d_scale_to_fit_button->setIcon(QIcon(":/graphics/scale_to_fit_24x24.png"));
	d_pan_button->setIcon(QIcon(":/graphics/pan_24x24.png"));
	d_rotate_2d_button->setIcon(QIcon(":/graphics/rotate2d_24x24.png"));
	d_rotate_3d_button->setIcon(QIcon(":/graphics/rotate3d_24x24.png"));
	
	d_scale_to_fit_button->setToolTip(QStringLiteral("Scale to Fit"));
	d_pan_button->setToolTip(QStringLiteral("Pan Mode"));
	d_rotate_2d_button->setToolTip(QStringLiteral("Rotate 2D Mode"));
	d_rotate_3d_button->setToolTip(QStringLiteral("Rotate 3D Mode"));
	
	
	if(d_gldisplay_error_label)
	{
		gl_error_text = QStringLiteral("<html><head/><body>");
		gl_error_text += QStringLiteral("<p><span style=\" font-weight:700;\">ERR:  OpenGL version not supported.  No graphical display will be shown.</span></p>");
		gl_error_text += QStringLiteral("<p><span style=\" font-weight:500; text-decoration: underline;\">OpenGL Version Text:</span></p>");
		gl_error_text += QString("<p>%1</p>").arg(gl_version_text);
		gl_error_text += QStringLiteral("<p><span style=\" font-weight:500; text-decoration: underline;\">OpenGL Extensions:</span></p>");
		gl_error_text += QString("<p>%1</p>").arg(gl_extensions_text);
		gl_error_text += QStringLiteral("</body></html>");
		
		d_gldisplay_error_label->setText(gl_error_text);
	}
	
	d_pan_button->setChecked(true);
	
	connect(d_scale_to_fit_button,&QToolButton::clicked,this,&TOpenGLDisplayWidget::Scale_To_Fit);
	connect(d_pan_button,&QToolButton::toggled,this,&TOpenGLDisplayWidget::Pan);
	connect(d_rotate_2d_button,&QToolButton::toggled,this,&TOpenGLDisplayWidget::Rotate2d);
	connect(d_rotate_3d_button,&QToolButton::toggled,this,&TOpenGLDisplayWidget::Rotate3d);
	
	if(d_opengl_display)
	{
		connect(d_opengl_display,&TOpenGLDisplay::Model_Click_Point,this,&TOpenGLDisplayWidget::Model_Item_Clicked);
	}
}

TOpenGLDisplayWidget::~TOpenGLDisplayWidget(void)
{
}

void TOpenGLDisplayWidget::Define_Machine_Volume(
	const TVector3						&pnt_min,
	const TVector3						&pnt_max)
{
	double								dval;
	
	if(!d_opengl_display)
	{
		return;
	}
	
	d_max_xyz = pnt_max;
	d_min_xyz = pnt_min;
	
	if(d_min_xyz.x > d_max_xyz.x)
	{
		dval = d_max_xyz.x;
		d_max_xyz.x = d_min_xyz.x;
		d_min_xyz.x = dval;
	}
	
	if(d_min_xyz.y > d_max_xyz.y)
	{
		dval = d_max_xyz.y;
		d_max_xyz.y = d_min_xyz.y;
		d_min_xyz.y = dval;
	}
	
	if(d_min_xyz.z > d_max_xyz.z)
	{
		dval = d_max_xyz.z;
		d_max_xyz.z = d_min_xyz.z;
		d_min_xyz.z = dval;
	}

	d_opengl_display->Define_Machine_Volume(d_min_xyz - TVector3(50,50,50),d_max_xyz + TVector3(50,50,50));
}

void TOpenGLDisplayWidget::Set_MCG(
	const int							id,
	const TVector3						&center,
	const TVector3						&axis,
	const double						&length)
{
	TOpenGLDisplay::TMCGGauge			mcg;
	
	if(!d_opengl_display)
	{
		return;
	}
	
	mcg.id = id;
	mcg.error = false;
	mcg.selected = false;
	
	mcg.center = center;
	mcg.axis = axis;
	mcg.length = length;
	
	mcg.min_xyz = mcg.center - (mcg.axis * 50);
	mcg.max_xyz = mcg.center + (mcg.axis * (length + 50));
	
	d_opengl_display->Add_MCG(mcg,true);
}

void TOpenGLDisplayWidget::Add_MCG(
	const int							id,
	const TVector3						&center,
	const TVector3						&axis,
	const double						&length,
	const bool							oot)
{
	TOpenGLDisplay::TMCGGauge			mcg;
	
	if(!d_opengl_display)
	{
		return;
	}
	
	mcg.id = id;
	mcg.error = oot;
	mcg.selected = false;
	
	mcg.center = center;
	mcg.axis = axis;
	mcg.length = length;
	
	mcg.min_xyz = mcg.center - (mcg.axis * 50);
	mcg.max_xyz = mcg.center + (mcg.axis * (length + 50));
	
	d_opengl_display->Add_MCG(mcg,false);
}

void TOpenGLDisplayWidget::Set_MCG_Path(
	const std::vector<TVector3>			&path_points)
{
	if(d_opengl_display)
	{
		d_opengl_display->Set_MCG_Path(path_points);
	}
}

void TOpenGLDisplayWidget::Clear_MCG_Path(void)
{
	if(d_opengl_display)
	{
		d_opengl_display->Clear_MCG_Path();
	}
}

void TOpenGLDisplayWidget::Clear_Selection(void)
{
	if(d_opengl_display)
	{
		d_opengl_display->Clear_Selection();
	}
}

void TOpenGLDisplayWidget::Set_Selection(
	const int							id)
{
	if(d_opengl_display)
	{
		d_opengl_display->Set_Selection(id);
	}
}

void TOpenGLDisplayWidget::Clear_MCG(void)
{
	if(d_opengl_display)
	{
		d_opengl_display->Clear_Render_Items();
	}
}

void TOpenGLDisplayWidget::Update(
	const bool							scale_to_fit)
{	
	if(!d_opengl_display)
	{
		return;
	}
	
	d_opengl_display->update();

	if(scale_to_fit)
	{
		d_opengl_display->Scale_To_Fit();
	}
}

void TOpenGLDisplayWidget::Scale_To_Fit(void)
{
	if(d_opengl_display)
	{
		d_opengl_display->Scale_To_Fit();
	}
}

void TOpenGLDisplayWidget::Rotate2d(
	bool								toggled)
{
	bool								prev_state;
	
	if(toggled)
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(false);
		d_pan_button->blockSignals(prev_state);
	
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Rotate_2D();
		}
	}
	else
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(true);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
}

void TOpenGLDisplayWidget::Rotate3d(
	bool								toggled)
{
	bool								prev_state;
	
	if(toggled)
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(false);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Rotate_3D();
		}
	}
	else
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(true);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
}

void TOpenGLDisplayWidget::Pan(
	bool								toggled)
{
	bool								prev_state;
	
	if(toggled)
	{
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
	else
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(true);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
}

