/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QString>
#include <cmath>

#include "program_dmis.h"

TProgramDMIS::TProgramDMIS(void)
{
}

TProgramDMIS::~TProgramDMIS(void)
{
	d_file.close();
}

bool TProgramDMIS::Create_Program(
	const QString						&file_name,
	const std::vector<TPathWidget::TPathPoint> &path_points,
	const double						&arm_length)
{
	std::vector<TPathWidget::TPathPoint>::const_iterator iter;
	QString								feature_name;
	TVector3							pnt;
	int									unique_id;
	bool								add_comment_line;
	static const double					PIVOT_SPHERE_RADIUS(2.0);
	static const double					MOVE_CLEARANCE(25);

	if(d_file.isOpen())
	{
		d_file.close();
	}
	
	if(path_points.size() < 4)
	{
		return false;
	}
	
	d_file.setFileName(file_name);

	if(d_file.open(QIODevice::WriteOnly))
	{
		this->Create_Header(QString("MCG DMIS program for use with an arm length of %1 mm").arg(arm_length,0,'f',1));
		
		// manual alignment
		this->Write_Line(QStringLiteral("Result=PROMPT/TEXT,'Measure 5 points on the MCG pivot sphere.'"));
		this->Write_Line(QStringLiteral("$$"));
		
		this->Add_Sphere_5(QStringLiteral("MAN_SPH"),TVector3(0,0,0),PIVOT_SPHERE_RADIUS);
		this->Write_Line(QStringLiteral("D(MAN_CSY)=TRANS/XORIG,FA(MAN_SPH),YORIG,FA(MAN_SPH),ZORIG,FA(MAN_SPH)"));
		this->Write_Line(QStringLiteral("$$"));

		// pivot sphere initial position in DCC
		this->Write_Line(QStringLiteral("Result=PROMPT/TEXT,'Clear probe for DCC'"));
		this->Write_Line(QStringLiteral("$$"));
		
		this->Write_Line(QStringLiteral("MODE/PROG,MAN"));
		this->Write_Line(QStringLiteral("$$"));
		
		pnt.Set(0,0,20);
		this->Add_Move_Point(pnt);
		this->Write_Line(QStringLiteral("$$"));

		this->Add_Sphere_25(QStringLiteral("MCG_SPH_1"),TVector3(0,0,0),PIVOT_SPHERE_RADIUS);
		this->Write_Line(QStringLiteral("D(DCC_CSY)=TRANS/XORIG,FA(MCG_SPH_1),YORIG,FA(MCG_SPH_1),ZORIG,FA(MCG_SPH_1)"));
		this->Write_Line(QStringLiteral("$$"));
		
		pnt.Set(0,0,20);
		this->Add_Move_Point(pnt);
		
		// position for loading arm
		pnt = path_points[0].point_vector * (arm_length + MOVE_CLEARANCE);
		pnt.z += 20;
		this->Add_Move_Point(pnt);
		
		pnt.z -= 20;
		this->Add_Move_Point(pnt);
		this->Write_Line(QStringLiteral("$$"));

		// load arm
		this->Write_Line(QStringLiteral("MODE/MAN"));
		this->Write_Line(QStringLiteral("$$"));
		
		this->Write_Line(QString("Result=PROMPT/TEXT,'Load MCG arm with length of %1 mm'").arg(arm_length,0,'f',1));
		this->Write_Line(QStringLiteral("$$"));
		
		this->Write_Line(QStringLiteral("MODE/PROG,MAN"));
		this->Write_Line(QStringLiteral("$$"));

		// measure points
		unique_id = 1;
		add_comment_line = false;
		for(iter = path_points.begin();iter != path_points.end();++iter)
		{
			if((*iter).point_type == TPathWidget::MOVE_POINT)
			{
				pnt = (*iter).point_vector * (arm_length + MOVE_CLEARANCE);
				this->Add_Move_Point(pnt);
				
				add_comment_line = true;
			}
			else if((*iter).point_type == TPathWidget::MEASURE_POINT)
			{
				feature_name = QString("MCG_PNT_%1").arg(unique_id);
				++unique_id;
				
				if(add_comment_line)
				{
					this->Write_Line(QStringLiteral("$$"));
					add_comment_line = false;
				}
				
				pnt = (*iter).point_vector * arm_length;
				this->Add_Touch_Point(feature_name,pnt,(*iter).point_vector);
				
				this->Report_Feature_Position(feature_name);
			}
		}

		// unload arm
		this->Write_Line(QStringLiteral("MODE/MAN"));
		this->Write_Line(QStringLiteral("$$"));
		
		this->Write_Line(QStringLiteral("Result=PROMPT/TEXT,'Remove MCG arm.'"));
		this->Write_Line(QStringLiteral("$$"));
		
		// measure pivot sphere
		this->Write_Line(QStringLiteral("Result=PROMPT/TEXT,'Clear probe for DCC'"));
		this->Write_Line(QStringLiteral("$$"));
		
		this->Write_Line(QStringLiteral("MODE/PROG,MAN"));
		this->Write_Line(QStringLiteral("$$"));
		
		pnt.Set(0,0,20);
		this->Add_Move_Point(pnt);
		this->Write_Line(QStringLiteral("$$"));

		this->Add_Sphere_25(QStringLiteral("MCG_SPH_2"),TVector3(0,0,0),PIVOT_SPHERE_RADIUS);
		this->Write_Line(QStringLiteral("$$"));
		
		pnt.Set(0,0,20);
		this->Add_Move_Point(pnt);
		this->Write_Line(QStringLiteral("$$"));

		// report pivot sphere locations in the machine coordinate system
		this->Write_Line(QStringLiteral("RECALL/DA(MC_CSY)"));
		this->Write_Line(QStringLiteral("$$"));
		
		this->Report_Feature_Position(QStringLiteral("MCG_SPH_1"));
		this->Report_Feature_Position(QStringLiteral("MCG_SPH_2"));
		this->Report_Feature_Form(QStringLiteral("MCG_SPH_1"));
		this->Report_Feature_Form(QStringLiteral("MCG_SPH_2"));

		this->Write_Line(QStringLiteral("ENDFIL"));
		
		d_file.close();

		return true;
	}
	
	return false;
}

void TProgramDMIS::Add_Move_Point(
	const TVector3 						&pnt)
{
	this->Write_Line(QString("GOTO/%1,%2,%3").arg(pnt.x,0,'f',3).arg(pnt.y,0,'f',3).arg(pnt.z,0,'f',3));
}

void TProgramDMIS::Add_Touch_Point(
	const QString						&name,
	const TVector3 						&pnt,
	const TVector3 						&vec)
{
	
	this->Write_Line(QString("F(%1)=FEAT/POINT,CART,%2,%3,%4,%5,%6,%7")
						.arg(name)
						.arg(pnt.x,0,'f',4)
						.arg(pnt.y,0,'f',4)
						.arg(pnt.z,0,'f',4)
						.arg(vec.i,0,'f',7)
						.arg(vec.j,0,'f',7)
						.arg(vec.k,0,'f',7));
	
	this->Write_Line(QString("MEAS/POINT,F(%1),1").arg(name));
	this->Write_Line(QString("\tPTMEAS/CART,%1,%2,%3,%4,%5,%6")
						.arg(pnt.x,0,'f',4)
						.arg(pnt.y,0,'f',4)
						.arg(pnt.z,0,'f',4)
						.arg(vec.i,0,'f',7)
						.arg(vec.j,0,'f',7)
						.arg(vec.k,0,'f',7));
	
	this->Write_Line(QStringLiteral("ENDMES"));
	this->Write_Line(QStringLiteral("$$"));
}

void TProgramDMIS::Add_Sphere_5(
	const QString						&name,
	const TVector3						&pos,
	const double						&radius)
{
	TVector3							pnt;
	TVector3							vec;
	
	this->Write_Line(QString("F(%1)=FEAT/SPHERE,OUTER,CART,%2,%3,%4,%5,0.000000,0.000000,1.000000")
					 .arg(name)
					 .arg(pos.x,0,'f',4)
					 .arg(pos.y,0,'f',4)
					 .arg(pos.z,0,'f',4)
					 .arg(radius * 2.0,0,'f',4));
	
	this->Write_Line(QString("MEAS/SPHERE,F(%1), 5").arg(name));
	
	vec.Set(1,0,0);
	pnt = pos + vec * radius;
	this->Write_Line(QString("\tPTMEAS/CART,%1,%2,%3,%4,%5,%6")
					 .arg(pnt.x,0,'f',4)
					 .arg(pnt.y,0,'f',4)
					 .arg(pnt.z,0,'f',4)
					 .arg(vec.i,0,'f',7)
					 .arg(vec.j,0,'f',7)
					 .arg(vec.k,0,'f',7));
	
	vec.Set(0,1,0);
	pnt = pos + vec * radius;
	this->Write_Line(QString("\tPTMEAS/CART,%1,%2,%3,%4,%5,%6")
					 .arg(pnt.x,0,'f',4)
					 .arg(pnt.y,0,'f',4)
					 .arg(pnt.z,0,'f',4)
					 .arg(vec.i,0,'f',7)
					 .arg(vec.j,0,'f',7)
					 .arg(vec.k,0,'f',7));
	
	vec.Set(-1,0,0);
	pnt = pos + vec * radius;
	this->Write_Line(QString("\tPTMEAS/CART,%1,%2,%3,%4,%5,%6")
					 .arg(pnt.x,0,'f',4)
					 .arg(pnt.y,0,'f',4)
					 .arg(pnt.z,0,'f',4)
					 .arg(vec.i,0,'f',7)
					 .arg(vec.j,0,'f',7)
					 .arg(vec.k,0,'f',7));
	
	vec.Set(0,-1,0);
	pnt = pos + vec * radius;
	this->Write_Line(QString("\tPTMEAS/CART,%1,%2,%3,%4,%5,%6")
					 .arg(pnt.x,0,'f',4)
					 .arg(pnt.y,0,'f',4)
					 .arg(pnt.z,0,'f',4)
					 .arg(vec.i,0,'f',7)
					 .arg(vec.j,0,'f',7)
					 .arg(vec.k,0,'f',7));
	
	vec.Set(0,0,1);
	pnt = pos + vec * radius;
	this->Write_Line(QString("\tPTMEAS/CART,%1,%2,%3,%4,%5,%6")
					 .arg(pnt.x,0,'f',4)
					 .arg(pnt.y,0,'f',4)
					 .arg(pnt.z,0,'f',4)
					 .arg(vec.i,0,'f',7)
					 .arg(vec.j,0,'f',7)
					 .arg(vec.k,0,'f',7));
	
	this->Write_Line(QStringLiteral("ENDMES"));
	this->Write_Line(QStringLiteral("$$"));
}

void TProgramDMIS::Add_Sphere_25(
	const QString						&name,
	const TVector3						&pos,
	const double						&radius)
{
	TVector3							pnt;
	TVector3							vec;
	int									cntr;
	double								angle_xy;
	double								angle_z;
	double								increment_xy;
	double								increment_z;
	static const double					PI(3.141592653589793);

	this->Write_Line(QString("F(%1)=FEAT/SPHERE,OUTER,CART,%2,%3,%4,%5,0.000000,0.000000,1.000000")
					 .arg(name)
					 .arg(pos.x,0,'f',4)
					 .arg(pos.y,0,'f',4)
					 .arg(pos.z,0,'f',4)
					 .arg(radius * 2.0,0,'f',4));
	
	this->Write_Line(QString("MEAS/SPHERE,F(%1), 25").arg(name));
	
	increment_z = PI / 7.0;
	angle_z = 0;
	
	angle_xy = 0;
	increment_xy = PI / 6.0;
	
	for(cntr = 0;cntr < 12;++cntr)
	{
		vec.x = cos(angle_xy) * cos(angle_z);
		vec.y = sin(angle_xy) * cos(angle_z);
		vec.z = sin(angle_z);
		
		pnt = pos + vec * radius;
		
		this->Write_Line(QString("\tPTMEAS/CART,%1,%2,%3,%4,%5,%6")
						 .arg(pnt.x,0,'f',4)
						 .arg(pnt.y,0,'f',4)
						 .arg(pnt.z,0,'f',4)
						 .arg(vec.i,0,'f',7)
						 .arg(vec.j,0,'f',7)
						 .arg(vec.k,0,'f',7));
		
		angle_xy += increment_xy;
	}
	
	angle_z += increment_z;
	
	angle_xy = 0;
	increment_xy = PI / 4.0;
	
	for(cntr = 0;cntr < 8;++cntr)
	{
		vec.x = cos(angle_xy) * cos(angle_z);
		vec.y = sin(angle_xy) * cos(angle_z);
		vec.z = sin(angle_z);
		
		pnt = pos + vec * radius;
		
		this->Write_Line(QString("\tPTMEAS/CART,%1,%2,%3,%4,%5,%6")
						 .arg(pnt.x,0,'f',4)
						 .arg(pnt.y,0,'f',4)
						 .arg(pnt.z,0,'f',4)
						 .arg(vec.i,0,'f',7)
						 .arg(vec.j,0,'f',7)
						 .arg(vec.k,0,'f',7));
		
		angle_xy += increment_xy;
	}
	
	angle_z += increment_z;
	
	angle_xy = 0;
	increment_xy = PI / 2.0;
	
	for(cntr = 0;cntr < 4;++cntr)
	{
		vec.x = cos(angle_xy) * cos(angle_z);
		vec.y = sin(angle_xy) * cos(angle_z);
		vec.z = sin(angle_z);
		
		pnt = pos + vec * radius;
		
		this->Write_Line(QString("\tPTMEAS/CART,%1,%2,%3,%4,%5,%6")
						 .arg(pnt.x,0,'f',4)
						 .arg(pnt.y,0,'f',4)
						 .arg(pnt.z,0,'f',4)
						 .arg(vec.i,0,'f',7)
						 .arg(vec.j,0,'f',7)
						 .arg(vec.k,0,'f',7));
		
		angle_xy += increment_xy;
	}
	
	vec.Set(0,0,1);
	pnt = pos + vec * radius;
	this->Write_Line(QString("\tPTMEAS/CART,%1,%2,%3,%4,%5,%6")
					 .arg(pnt.x,0,'f',4)
					 .arg(pnt.y,0,'f',4)
					 .arg(pnt.z,0,'f',4)
					 .arg(vec.i,0,'f',7)
					 .arg(vec.j,0,'f',7)
					 .arg(vec.k,0,'f',7));
	
	this->Write_Line(QStringLiteral("ENDMES"));
	this->Write_Line(QStringLiteral("$$"));
}
									  
void TProgramDMIS::Report_Feature_Position(
	const QString						&name)
{
	this->Write_Line(QString("OUTPUT/F(%1),T(X_TOL),T(Y_TOL),T(Z_TOL)").arg(name));
	this->Write_Line(QString("OUTPUT/FA(%1),TA(X_TOL),TA(Y_TOL),TA(Z_TOL)").arg(name));
	this->Write_Line(QStringLiteral("$$"));
}

void TProgramDMIS::Report_Feature_Form(
	const QString						&name)
{
	this->Write_Line(QString("T(%1_F)=TOL/CIRLTY,0.1").arg(name));
	this->Write_Line(QString("OUTPUT/F(%1),T(%1_F)").arg(name));
	this->Write_Line(QString("OUTPUT/FA(%1),TA(%1_F)").arg(name));
	this->Write_Line(QStringLiteral("$$"));
}

void TProgramDMIS::Create_Header(
	const QString					&comment_text)
{
	this->Write_Line(QStringLiteral("$$ MCG DMIS file"));
	this->Write_Line(QStringLiteral("DMISMN/'mcg.dmis' 04.0"));
	this->Write_Line(QStringLiteral("FILNAM/'mcg.dmis'"));
	this->Write_Line(QStringLiteral("$$"));
	
	if(comment_text.length())
	{
		this->Write_Line(QString("$$ %1").arg(comment_text));
		this->Write_Line(QStringLiteral("$$"));
	}
	
	this->Write_Line(QStringLiteral("DECL/COMMON,INTGR,Result"));
	this->Write_Line(QStringLiteral("$$"));
	this->Write_Line(QStringLiteral("UNITS/MM,ANGDEC"));
	this->Write_Line(QStringLiteral("D(MC_CSY)=DATSET/MCS"));
	this->Write_Line(QStringLiteral("MODE/MAN"));
	this->Write_Line(QStringLiteral("$$"));
	this->Write_Line(QStringLiteral("$$ Prefer 5 decimal places display precision."));
	this->Write_Line(QStringLiteral("DECPL/ALL,5"));
	this->Write_Line(QStringLiteral("$$"));
	this->Write_Line(QStringLiteral("SNSLCT/S(MCG_PROBE)"));
	this->Write_Line(QStringLiteral("SNSLCT/SA(MCG_PROBE)"));
	this->Write_Line(QStringLiteral("$$"));
	this->Write_Line(QStringLiteral("T(X_TOL)=TOL/CORTOL,XAXIS,-0.1,0.1"));
	this->Write_Line(QStringLiteral("T(Y_TOL)=TOL/CORTOL,YAXIS,-0.1,0.1"));
	this->Write_Line(QStringLiteral("T(Z_TOL)=TOL/CORTOL,ZAXIS,-0.1,0.1"));
	this->Write_Line(QStringLiteral("$$"));
}

bool TProgramDMIS::Write_Line(
	const QString						&line)
{
	QString								line_text;
	QString								text;
	int									tab_count;
	static const int					MAX_LINE_LENGTH(78);
	
	if(!d_file.isOpen())
	{
		return false;
	}
	
	line_text = line;
	
	tab_count = line_text.count('\t');
	
	while(line_text.length() > (MAX_LINE_LENGTH - tab_count * 4))	// line limit is 80 chars including CR/LF.  assume tab = 4 chars
	{
		text = line_text;
		text.truncate(MAX_LINE_LENGTH - tab_count * 4 - 1);
		text.push_back('$');
		
		d_file.write(text.toLatin1());
		d_file.write("\r\n");

		line_text = line_text.mid(MAX_LINE_LENGTH - tab_count * 4 - 1);
	}
	
	if(line_text.size())
	{
		d_file.write(line_text.toLatin1());
		d_file.write("\r\n");
	}
	
	return true;
}
