/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QGridLayout>
#include <QLabel>
#include <QLineEdit>
#include <QSpacerItem>
#include <QGraphicsView>
#include <QGraphicsScene>
#include <QGraphicsTextItem>
#include <QGraphicsRectItem>
#include <QTransform>
//#include <QToolButton>
#include <QPrinter>
#include <QPrintDialog>
//#include <QFileDialog>
#include <QColor>
#include <QFont>
#include <QDateTime>
#include <assert.h>

#include "reportwidget.h"

TReportWidget::TReportWidget(
	const QWidget 						*parent,
	Qt::WindowFlags 					flags)
:	QWidget(const_cast<QWidget*>(parent),flags)
{
    QGridLayout 						*widget_layout;
	
	d_font_family_name = QStringLiteral("Helvetica");
	
#ifdef Q_OS_MAC
	d_font_size = 12;
	d_font_size_title = 14;
	d_font_size_header = 20;
	d_font_size_footer = 8;
#else
	d_font_size = 10;
	d_font_size_title = 12;
	d_font_size_header = 14;
	d_font_size_footer = 6;
#endif
	
    widget_layout = new QGridLayout(this);

	d_graphic_scene = new QGraphicsScene();
    d_graphic_scene->setItemIndexMethod(QGraphicsScene::NoIndex);
	
	d_graphic_view = new QGraphicsView(d_graphic_scene,this);
    d_graphic_view->setRenderHint(QPainter::Antialiasing);
    d_graphic_view->setCacheMode(QGraphicsView::CacheBackground);
    d_graphic_view->setDragMode(QGraphicsView::ScrollHandDrag);
 
    widget_layout->addWidget(d_graphic_view, 0, 0, 1, 1);

	// defaults
	d_page_size.setWidth(750);	// (1500);
	d_page_size.setHeight(970);	// (1941);
	d_page_margin = 0.1;
	d_display_scale_factor = 1.0;
	d_graphic_view_brush.setColor(QColor(164,164,164));
	d_graphic_view_brush.setStyle(Qt::SolidPattern);
	
	d_graphic_scene->setSceneRect(0, 0, d_page_size.width(), d_page_size.height());
	d_graphic_view->setBackgroundBrush(d_graphic_view_brush);
	d_graphic_view->setAcceptDrops(false);
	
    this->Initialize_Page(&d_first_page,QPoint(0,0));
    this->Setup_Base_Page(&d_first_page);
    this->Setup_Summary_Page(&d_first_page);
    
	d_graphic_view->scale(d_display_scale_factor,d_display_scale_factor);
}

TReportWidget::~TReportWidget(void)
{

}

void TReportWidget::Set_Date_Time(
	const QString 						&text)
{
	d_first_page.test_date_time_value->setPlainText(text);
}

void TReportWidget::Set_Temperature(
	const double 						&value)
{
	d_first_page.temperature_value->setPlainText(QString("%1 ˚C").arg(value,0,'f',2));
}

void TReportWidget::Set_Tolerance(
	const double 						&value)
{
	d_first_page.tolerance_value->setPlainText(QString("%1").arg(value,0,'f',4));
}

void TReportWidget::Set_Measurement_Range(
	const double 						&value)
{
	d_first_page.range_value->setPlainText(QString("%1").arg(value,0,'f',4));
}

void TReportWidget::Set_Pivot_Sphere_Start(
	const TVector3						&position)
{
	d_first_page.pivot_sphere_start_value->setPlainText(QString("%1,    %2,    %3").arg(position.x,0,'f',4).arg(position.y,0,'f',4).arg(position.z,0,'f',4));
}

void TReportWidget::Set_Pivot_Sphere_End(
	const TVector3						&position)
{
	d_first_page.pivot_sphere_end_value->setPlainText(QString("%1,    %2,    %3").arg(position.x,0,'f',4).arg(position.y,0,'f',4).arg(position.z,0,'f',4));
}

void TReportWidget::Set_Measurement_Sphere(
	const TVector3						&position)
{
	d_first_page.measured_sphere_value->setPlainText(QString("%1,    %2,    %3").arg(position.x,0,'f',4).arg(position.y,0,'f',4).arg(position.z,0,'f',4));
}

void TReportWidget::Set_Probe_Offset(
	const TVector3						&offset)
{
	d_first_page.probe_offset_value->setPlainText(QString("%1,    %2,    %3").arg(offset.x,0,'f',4).arg(offset.y,0,'f',4).arg(offset.z,0,'f',4));
}

void TReportWidget::Set_Probe_Error(
	const double 						&value)
{
	d_first_page.probe_error_value->setPlainText(QString("%1").arg(value,0,'f',4));
}

void TReportWidget::Set_Average_Measurement_Length(
	const double 						&value)
{
	d_first_page.average_length_value->setPlainText(QString("%1").arg(value,0,'f',4));
}

void TReportWidget::Set_Average_Length_XY(
	const double 						&value,
	const bool 							valid)
{
	if(valid)
	{
		d_first_page.average_length_xy_value->setPlainText(QString("%1").arg(value,0,'f',4));
	}
	else
	{
		d_first_page.average_length_xy_value->setPlainText(tr("Insufficient Data"));
	}
}

void TReportWidget::Set_Average_Length_YZ(
	const double 						&value,
	const bool 							valid)
{
	if(valid)
	{
		d_first_page.average_length_yz_value->setPlainText(QString("%1").arg(value,0,'f',4));
	}
	else
	{
		d_first_page.average_length_yz_value->setPlainText(tr("Insufficient Data"));
	}
}

void TReportWidget::Set_Average_Length_ZX(
	const double 						&value,
	const bool 							valid)
{
	if(valid)
	{
		d_first_page.average_length_zx_value->setPlainText(QString("%1").arg(value,0,'f',4));
	}
	else
	{
		d_first_page.average_length_zx_value->setPlainText(tr("Insufficient Data"));
	}
}

void TReportWidget::Set_Relative_Scale_Error_X(
	const double 						&value,
	const bool 							valid)
{
	if(valid)
	{
		d_first_page.relative_scale_error_x_value->setPlainText(QString("%1").arg(value,7,'f',4));
	}
	else
	{
		d_first_page.relative_scale_error_x_value->setPlainText(tr("Insufficient Data"));
	}
}

void TReportWidget::Set_Relative_Scale_Error_Y(
	const double 						&value,
	const bool 							valid)
{
	if(valid)
	{
		d_first_page.relative_scale_error_y_value->setPlainText(QString("%1").arg(value,7,'f',4));
	}
	else
	{
		d_first_page.relative_scale_error_y_value->setPlainText(tr("Insufficient Data"));
	}
}

void TReportWidget::Set_Relative_Scale_Error_Z(
	const double 						&value,
	const bool 							valid)
{
	if(valid)
	{
		d_first_page.relative_scale_error_z_value->setPlainText(QString("%1").arg(value,7,'f',4));
	}
	else
	{
		d_first_page.relative_scale_error_z_value->setPlainText(tr("Insufficient Data"));
	}
}


void TReportWidget::Set_Squareness_Error_XY(
	const double 						&value,
	const bool 							valid)
{
	if(valid)
	{
		d_first_page.squareness_xy_value->setPlainText(QString("%1 mm/m").arg(value,7,'f',4));
	}
	else
	{
		d_first_page.squareness_xy_value->setPlainText(tr("Insufficient Data"));
	}
}

void TReportWidget::Set_Squareness_Error_YZ(
	const double 						&value,
	const bool 							valid)
{
	if(valid)
	{
		d_first_page.squareness_yz_value->setPlainText(QString("%1 mm/m").arg(value,7,'f',4));
	}
	else
	{
		d_first_page.squareness_yz_value->setPlainText(tr("Insufficient Data"));
	}
}

void TReportWidget::Set_Squareness_Error_ZX(
	const double 						&value,
	const bool 							valid)
{
	if(valid)
	{
		d_first_page.squareness_zx_value->setPlainText(QString("%1 mm/m").arg(value,7,'f',4));
	}
	else
	{
		d_first_page.squareness_zx_value->setPlainText(tr("Insufficient Data"));
	}
}

void TReportWidget::Set_Minimum_Measured_Length(
	const double 						&value)
{
	d_first_page.minimum_length_value->setPlainText(QString("%1").arg(value,0,'f',4));
}

void TReportWidget::Set_Maximum_Measured_Length(
	const double 						&value)
{
	d_first_page.maximum_length_value->setPlainText(QString("%1").arg(value,0,'f',4));
}

void TReportWidget::Set_Median_Measured_Length(
	const double 						&value)
{
	d_first_page.median_length_value->setPlainText(QString("%1").arg(value,0,'f',4));
}

void TReportWidget::Set_Stddev_Measured_Length(
	const double 						&value)
{
	d_first_page.stddev_length_value->setPlainText(QString("%1").arg(value,0,'f',4));
}

void TReportWidget::Clear_Report(void)
{	
	std::vector<TMeasurementData>::iterator	iter;
		
	for(iter = d_measurement_data.begin();iter != d_measurement_data.end();++iter)
	{
		d_graphic_scene->removeItem((*iter).elevation_value);
		d_graphic_scene->removeItem((*iter).angle_value);
		d_graphic_scene->removeItem((*iter).length_value);
		d_graphic_scene->removeItem((*iter).length_range_value);
		d_graphic_scene->removeItem((*iter).length_stddev_value);
		d_graphic_scene->removeItem((*iter).length_dev_value);
		
		delete (*iter).elevation_value;
		delete (*iter).angle_value;
		delete (*iter).length_value;
		delete (*iter).length_range_value;
		delete (*iter).length_stddev_value;
		delete (*iter).length_dev_value;
	}
	
	d_measurement_data.clear();
	
	d_first_page.test_date_time_value->setPlainText(QString());
	d_first_page.temperature_value->setPlainText(QString());
	d_first_page.range_value->setPlainText(QString());
	d_first_page.range_value->setPlainText(QString());
	d_first_page.tolerance_value->setPlainText(QString());
	
	d_first_page.pivot_sphere_start_value->setPlainText(QString());
	d_first_page.pivot_sphere_end_value->setPlainText(QString());
	d_first_page.measured_sphere_value->setPlainText(QString());

	d_first_page.probe_offset_value->setPlainText(QString());
	d_first_page.probe_error_value->setPlainText(QString());
	
	d_first_page.average_length_xy_value->setPlainText(QString());
	d_first_page.average_length_yz_value->setPlainText(QString());
	d_first_page.average_length_zx_value->setPlainText(QString());
	d_first_page.relative_scale_error_x_value->setPlainText(QString());
	d_first_page.relative_scale_error_y_value->setPlainText(QString());
	d_first_page.relative_scale_error_z_value->setPlainText(QString());
	d_first_page.squareness_xy_value->setPlainText(QString());
	d_first_page.squareness_yz_value->setPlainText(QString());
	d_first_page.squareness_zx_value->setPlainText(QString());

	d_first_page.average_length_value->setPlainText(QString());
	d_first_page.minimum_length_value->setPlainText(QString());
	d_first_page.maximum_length_value->setPlainText(QString());
	d_first_page.median_length_value->setPlainText(QString());
	d_first_page.stddev_length_value->setPlainText(QString());
}

void TReportWidget::Add_Measurement_Data(
	const double 						&Elevation,
	const double 						&Angle,
	const double 						&Length,
	const double 						&RangeLength,
	const double 						&SDLength,
	const double						&DevLength)
{
	QFont								normal_font(d_font_family_name,d_font_size);
	TMeasurementData					measurement_data;
	
	measurement_data.elevation_value = d_graphic_scene->addText(QString("%1  ").arg(Elevation,0,'f',1),normal_font);
	measurement_data.angle_value = d_graphic_scene->addText(QString("%1  ").arg(Angle,0,'f',1),normal_font);
	measurement_data.length_value = d_graphic_scene->addText(QString("%1").arg(Length,0,'f',4),normal_font);
	measurement_data.length_range_value = d_graphic_scene->addText(QString("%1").arg(RangeLength,0,'f',4),normal_font);
	measurement_data.length_stddev_value = d_graphic_scene->addText(QString("%1").arg(SDLength,0,'f',4),normal_font);
	measurement_data.length_dev_value = d_graphic_scene->addText(QString("%1 ").arg(DevLength,0,'f',4),normal_font);
	
	measurement_data.elevation_value->setZValue(2);
	measurement_data.angle_value->setZValue(2);
	measurement_data.length_value->setZValue(2);
	measurement_data.length_range_value->setZValue(2);
	measurement_data.length_stddev_value->setZValue(2);
	measurement_data.length_dev_value->setZValue(2);
	
	d_measurement_data.push_back(measurement_data);
}

TReportWidget::TReportBase::TReportBase(void)
{
	
}

TReportWidget::TReportBase::~TReportBase(void)
{
	
}

TReportWidget::TReportFirstPage::TReportFirstPage(void)
: TReportWidget::TReportBase()
{
	
}

TReportWidget::TReportFirstPage::~TReportFirstPage(void)
{
	
}

TReportWidget::TReportMeasurementPage::TReportMeasurementPage(void)
: TReportWidget::TReportBase()
{
	
}

TReportWidget::TReportMeasurementPage::~TReportMeasurementPage(void)
{
	
}

void TReportWidget::Update_Report(void)
{
	std::vector<TReportMeasurementPage>::iterator	iter;
	TReportMeasurementPage				report_page;
	int									page_items_limit;
	int									page_measurement_count;
	int									cntr;
	int									page_count;
	
	for(iter = d_measurement_pages.begin();iter != d_measurement_pages.end();++iter)
	{
		d_graphic_scene->removeItem((*iter).title_text);
		d_graphic_scene->removeItem((*iter).page_text);
		d_graphic_scene->removeItem((*iter).program_information_text);
		d_graphic_scene->removeItem((*iter).units_text);
		d_graphic_scene->removeItem((*iter).header_line_rect);
		d_graphic_scene->removeItem((*iter).label_line_rect);
		d_graphic_scene->removeItem((*iter).footer_line_rect);
		d_graphic_scene->removeItem((*iter).page_background_rect);

		d_graphic_scene->removeItem((*iter).elevation_label);
		d_graphic_scene->removeItem((*iter).azimuth_label);
		d_graphic_scene->removeItem((*iter).length_label);
		d_graphic_scene->removeItem((*iter).length_range_label);
		d_graphic_scene->removeItem((*iter).length_stddev_label);
		d_graphic_scene->removeItem((*iter).length_deviation_label);
	}

	d_measurement_pages.clear();
	
	if(!d_measurement_data.empty())
	{
		// Page 2 exists automatically when there are points
		this->Initialize_Page(&report_page,QPoint(0,d_page_size.height()));//970));
		this->Setup_Base_Page(&report_page);
									
		// Determine the total number of pages.
		page_items_limit = static_cast<int>((report_page.page_content_rect.height() / d_line_height) - 6);
		page_measurement_count = static_cast<int>(d_measurement_data.size()) / page_items_limit;
		page_count = page_measurement_count + 2;
		
		d_graphic_scene->setSceneRect(0, 0, d_page_size.width(), d_page_size.height() * (page_count));
		
		d_first_page.page_text->setPlainText(tr("Page 1 of %1").arg(page_count));
		report_page.page_text->setPlainText(tr("Page 2 of %1").arg(page_count));
		
		d_measurement_pages.push_back(report_page);
		
		if(page_measurement_count > 0)
		{
			for(cntr = 0;cntr < page_measurement_count;cntr++)
			{
				this->Initialize_Page(&report_page,QPoint(0,d_page_size.height() * (cntr + 2)));
				this->Setup_Base_Page(&report_page);
											
				report_page.page_text->setPlainText(tr("page %1 of %2").arg(cntr + 3).arg(page_count));
	
				d_measurement_pages.push_back(report_page);
			}
		}
			
		for(cntr = 0, iter = d_measurement_pages.begin();iter != d_measurement_pages.end();++iter)
		{
			Setup_Measurement_Page(&(*iter),cntr,cntr + page_items_limit);
			
			cntr += page_items_limit;
		}
	}
	else
	{
		// One page only with no points
		d_graphic_scene->setSceneRect(0, 0, d_page_size.width(), d_page_size.height());
		d_first_page.page_text->setPlainText(tr("Page 1 of 1"));
	}
}


void TReportWidget::Zoom_In(void)
{
	double 								scale_factor;
	QTransform 							graphics_view_mat;
	
	if(d_display_scale_factor < 10)
	{
		scale_factor = d_display_scale_factor / 0.80;
		graphics_view_mat = d_graphic_view->transform();
		
		d_graphic_view->resetTransform();
		d_graphic_view->translate(graphics_view_mat.dx(), graphics_view_mat.dy());
		d_graphic_view->scale(scale_factor, scale_factor);
		
		d_display_scale_factor = scale_factor;
	}
}

void TReportWidget::Zoom_Out(void)
{
	double 								scale_factor;
	QTransform 							graphics_view_mat;
	
	if(d_display_scale_factor > 0.1)
	{
		scale_factor = d_display_scale_factor * 0.80;
		graphics_view_mat = d_graphic_view->transform();
		
		d_graphic_view->resetTransform();
		d_graphic_view->translate(graphics_view_mat.dx(), graphics_view_mat.dy());
		d_graphic_view->scale(scale_factor, scale_factor);
		
		d_display_scale_factor = scale_factor;
	}
}

void TReportWidget::Print(void)
{
	QPainter							painter;
	bool								final_page;
	int									page_cntr;
	QPrinter							printer(QPrinter::HighResolution);
	
	printer.setPaperSize(QPrinter::Letter);
	printer.setOrientation(QPrinter::Portrait);
	printer.setPageMargins(10.0,10.0,10.0,10.0,QPrinter::Millimeter);
	printer.setFullPage(true);
	
	
	QPrintDialog						print_dialog(&printer,this);
	
	if(print_dialog.exec() == QDialog::Accepted)
	{
		if(!painter.begin(&printer))
		{
			return;
		}
		
		final_page = false;
		page_cntr = 0;
		
		for(;;)
		{
			this->Render_Page(&painter,printer.pageRect(),++page_cntr,&final_page);
			
			if(final_page)
			{
				break;
			}
			
			printer.newPage();
		};
		
		painter.end();
	}
}

void TReportWidget::Write(
	const QString						&file_name)
{
	QPrinter 							printer(QPrinter::HighResolution);
	QPainter							painter;
	bool								final_page;
	int									page_cntr;

	printer.setOutputFormat(QPrinter::PdfFormat);
	printer.setOutputFileName(file_name);
	printer.setFullPage(true);
	
	if(!painter.begin(&printer))
	{
		return;
	}
	
	
	final_page = false;
	page_cntr = 0;
	
	for(;;)
	{
		this->Render_Page(&painter,printer.pageRect(),++page_cntr,&final_page);
		
		if(final_page)
		{
			break;
		}
		
		printer.newPage();
	};
	
	painter.end();
}


void TReportWidget::Initialize_Page(
	TReportWidget::TReportBase 			* const page,
	const QPoint						&top_left)
{
	QRectF								rect;
	int									margin;
	QPen								pen;
	QBrush								brush;
	
	pen.setStyle(Qt::SolidLine);
	pen.setWidth(0);
	pen.setColor(QColor(255,255,255));
	
	brush.setStyle(Qt::SolidPattern);
	brush.setColor(QColor(255,255,255));
	
	page->page_rect.moveTo(top_left);
	page->page_rect.setSize(d_page_size);
	
	rect = page->page_rect;
	
	margin = static_cast<int>(static_cast<double>(d_page_size.width()) * d_page_margin + 0.5);
	
	rect.adjust(margin / 4,margin / 4,-margin / 4,-margin / 4);
	page->page_background_rect = d_graphic_scene->addRect(rect,pen,brush);
	
	page->page_rect.adjust(margin,margin,-margin,-margin);
}

void TReportWidget::Setup_Base_Page(
	TReportWidget::TReportBase 			* const page)
{
	QFont								header_font(d_font_family_name,d_font_size_header);
	QFont								footer_font(d_font_family_name,d_font_size_footer);
	QRectF								text_rect;
	int									header_line_position;
	int									footer_line_position;
	QRect								line_rect;
	QPen								pen;
	QBrush								brush;
	
	pen.setStyle(Qt::SolidLine);
	pen.setWidth(0);
	pen.setColor(QColor(0,0,0));
	
	brush.setStyle(Qt::SolidPattern);
	brush.setColor(QColor(0,0,0));
	
	page->page_content_rect = page->page_rect;
	
	page->title_text = d_graphic_scene->addText(tr("Machine Checking Gauge Measurement Report"),header_font);
	page->title_text->setPos(page->page_content_rect.topLeft());
	page->title_text->setZValue(1);
	text_rect = page->title_text->boundingRect();
	
	header_line_position = page->page_content_rect.top() + static_cast<int>(text_rect.height());
	
	page->page_text = d_graphic_scene->addText(QStringLiteral("Page 1 of 1"),footer_font);
	text_rect = page->page_text->boundingRect();
		
	page->page_text->setPos(page->page_content_rect.bottomRight() - text_rect.bottomRight() - text_rect.topLeft());
	page->page_text->setZValue(1);
	
	page->program_information_text = d_graphic_scene->addText(QStringLiteral("Select Calibration Incorporated.  www.selectcalibration.ca"),footer_font);
	text_rect = page->program_information_text->boundingRect();
	
	page->program_information_text->setPos(page->page_content_rect.bottomLeft() - text_rect.bottomLeft() - text_rect.topLeft());
	page->program_information_text->setZValue(1);

	footer_line_position = page->page_content_rect.bottom() - static_cast<int>(text_rect.height());
	
	line_rect.setRect(page->page_content_rect.left(),header_line_position,page->page_content_rect.width(),2);
	page->header_line_rect = d_graphic_scene->addRect(line_rect,pen,brush);
	page->header_line_rect->setZValue(1);
	
	line_rect.setRect(page->page_content_rect.left(),footer_line_position-2,page->page_content_rect.width(),2);
	page->footer_line_rect = d_graphic_scene->addRect(line_rect,pen,brush);
	page->footer_line_rect->setZValue(1);
	
	page->units_text = d_graphic_scene->addText(tr("Reported values are in metric (mm) unless otherwise specified"),footer_font);
	text_rect = page->units_text->boundingRect();
	page->units_text->setPos(QPointF(page->page_content_rect.right() - text_rect.width(),footer_line_position - text_rect.height()));
					
	page->units_text->setZValue(1);
	
	page->page_content_rect.setTop(header_line_position + 2);
	page->page_content_rect.setBottom(footer_line_position);
}

void TReportWidget::Setup_Summary_Page(
	TReportFirstPage 					* const page)
{
	QFont								title_font(d_font_family_name,d_font_size_title);
	QFont								bold_font(d_font_family_name,d_font_size,QFont::Bold);
	QFont								normal_font(d_font_family_name,d_font_size);
	QRectF								text_rect;
	QRectF								image_rect;
	QRect								line_rect;
	double								line_position;
	double								dval;
	double								value_column_position;
	double								label_column_position;
	int									ival;
	int									line_width;
	int									test_results_position;
	int									location_results_position;
	int									probe_results_position;
	int									analysis_results_position;
	int									statistics_results_position;
	QPen								pen;
	QBrush								brush;

	pen.setStyle(Qt::SolidLine);
	pen.setWidth(0);
	pen.setColor(QColor(0,0,0));
	
	brush.setStyle(Qt::SolidPattern);
	brush.setColor(QColor(0,0,0));
	
	// first item added is one that uses a normal font
	// used to determine the line height

	page->test_date_time_label = d_graphic_scene->addText(QStringLiteral("Report Date and Time: "),normal_font);
	text_rect = page->test_date_time_label->boundingRect();
	d_line_height = text_rect.height();
	
	// make sure everything fits within 40 lines
	ival = page->page_content_rect.height();
	while(d_line_height > 0 && (ival / d_line_height) < 40)
	{
		d_line_height--;
	}

	page->column_left = page->page_content_rect.left();			// + static_cast<int>(d_line_height);
	page->column_right = page->page_content_rect.right();		//- static_cast<int>(d_line_height);
	line_position = page->page_content_rect.top() + d_line_height;
	line_width = page->column_right - page->column_left;
	
	label_column_position = page->column_left + d_line_height * 2;
	value_column_position = (page->column_left + page->column_right)/2;
	
	// test result title
	line_position += d_line_height;
	
	line_rect.setRect(page->column_left,line_position,line_width,1);
	page->test_results_rect = d_graphic_scene->addRect(line_rect,pen,brush);
	page->test_results_rect->setZValue(2);
	
	page->test_results_label = d_graphic_scene->addText(QStringLiteral("Measurement Results"),title_font);
	text_rect = page->test_results_label->boundingRect();
	page->test_results_label->setPos(QPointF(page->column_left,line_position - text_rect.height()));
	page->test_results_label->setZValue(2);

	line_position += d_line_height;
	test_results_position = line_position;

	page->test_date_time_label->setPos(QPointF(label_column_position,line_position));
	page->test_date_time_label->setZValue(2);
	
	text_rect = page->test_date_time_label->boundingRect();
	dval = page->test_date_time_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->temperature_label = d_graphic_scene->addText(QStringLiteral("Temperature: "),normal_font);
	page->temperature_label->setPos(QPointF(label_column_position,line_position + d_line_height));
	page->temperature_label->setZValue(2);
	
	text_rect = page->temperature_label->boundingRect();
	dval = page->temperature_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->tolerance_label = d_graphic_scene->addText(QStringLiteral("Tolerance: "),normal_font);
	page->tolerance_label->setPos(QPointF(label_column_position,line_position + d_line_height * 2));
	page->tolerance_label->setZValue(2);
	
	text_rect = page->tolerance_label->boundingRect();
	dval = page->tolerance_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->range_label = d_graphic_scene->addText(QStringLiteral("Measured Range: "),bold_font);
	page->range_label->setPos(QPointF(label_column_position,line_position  + d_line_height * 3));
	page->range_label->setZValue(2);
	
	text_rect = page->range_label->boundingRect();
	dval = page->range_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}

	line_position += d_line_height * 5;


	// test location title
	line_position += d_line_height;
	
	line_rect.setRect(page->column_left,line_position,line_width,1);
	page->test_position_rect = d_graphic_scene->addRect(line_rect,pen,brush);
	page->test_position_rect->setZValue(2);

	page->test_position_label = d_graphic_scene->addText(QStringLiteral("Measurement Location in Machine"),title_font);
	text_rect = page->test_position_label->boundingRect();
	page->test_position_label->setPos(QPointF(page->column_left,line_position - text_rect.height()));
	page->test_position_label->setZValue(2);
	
	line_position += d_line_height;
	location_results_position = line_position;

	page->pivot_sphere_start_label = d_graphic_scene->addText(QStringLiteral("Pivot Sphere Start: "),normal_font);
	page->pivot_sphere_start_label->setPos(QPointF(label_column_position,line_position));
	page->pivot_sphere_start_label->setZValue(2);
	
	text_rect = page->pivot_sphere_start_label->boundingRect();
	dval = page->pivot_sphere_start_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}

	page->pivot_sphere_end_label = d_graphic_scene->addText(QStringLiteral("Pivot Sphere End: "),normal_font);
	page->pivot_sphere_end_label->setPos(QPointF(label_column_position,line_position + d_line_height));
	page->pivot_sphere_end_label->setZValue(2);
	
	text_rect = page->pivot_sphere_end_label->boundingRect();
	dval = page->pivot_sphere_end_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->measured_sphere_label = d_graphic_scene->addText(QStringLiteral("Sphere Center: "),normal_font);
	page->measured_sphere_label->setPos(QPointF(label_column_position,line_position + d_line_height * 2));
	page->measured_sphere_label->setZValue(2);
	
	text_rect = page->measured_sphere_label->boundingRect();
	dval = page->pivot_sphere_end_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	line_position += d_line_height * 4;
	
	// probe information title
	line_position += d_line_height;
	
	line_rect.setRect(page->column_left,line_position,line_width,1);
	page->probe_results_rect = d_graphic_scene->addRect(line_rect,pen,brush);
	page->probe_results_rect->setZValue(2);
	
	page->probe_results_label = d_graphic_scene->addText(QStringLiteral("Probe Information"),title_font);
	text_rect = page->probe_results_label->boundingRect();
	page->probe_results_label->setPos(QPointF(page->column_left,line_position - text_rect.height()));
	page->probe_results_label->setZValue(2);

	line_position += d_line_height;
	probe_results_position = line_position;

	page->probe_offset_label = d_graphic_scene->addText(QStringLiteral("Probe Offset: "),normal_font);
	page->probe_offset_label->setPos(QPointF(label_column_position,line_position));
	page->probe_offset_label->setZValue(2);
	
	text_rect = page->probe_offset_label->boundingRect();
	dval = page->probe_offset_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->probe_error_label = d_graphic_scene->addText(QStringLiteral("Probe Error: "),normal_font);
	page->probe_error_label->setPos(QPointF(label_column_position,line_position + d_line_height));
	page->probe_error_label->setZValue(2);
	
	text_rect = page->probe_error_label->boundingRect();
	dval = page->probe_error_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	line_position += d_line_height * 3;

	// analysis result title
	line_position += d_line_height;
	
	line_rect.setRect(page->column_left,line_position,line_width,1);
	page->analysis_results_rect = d_graphic_scene->addRect(line_rect,pen,brush);
	page->analysis_results_rect->setZValue(2);

	page->analysis_results_label = d_graphic_scene->addText(QStringLiteral("Data Analysis"),title_font);
	text_rect = page->analysis_results_label->boundingRect();
	page->analysis_results_label->setPos(QPointF(page->column_left,line_position - text_rect.height()));
	page->analysis_results_label->setZValue(2);
	
	line_position += d_line_height;
	analysis_results_position = line_position;

	page->average_length_xy_label = d_graphic_scene->addText(QStringLiteral("Average Length XY: "),normal_font);
	page->average_length_xy_label->setPos(QPointF(label_column_position,line_position));
	page->average_length_xy_label->setZValue(2);
	
	text_rect = page->average_length_xy_label->boundingRect();
	dval = page->average_length_xy_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->average_length_yz_label = d_graphic_scene->addText(QStringLiteral("Average Length YZ: "),normal_font);
	page->average_length_yz_label->setPos(QPointF(label_column_position,line_position + d_line_height));
	page->average_length_yz_label->setZValue(2);
	
	text_rect = page->average_length_yz_label->boundingRect();
	dval = page->average_length_yz_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->average_length_zx_label = d_graphic_scene->addText(QStringLiteral("Average Length ZX: "),normal_font);
	page->average_length_zx_label->setPos(QPointF(label_column_position,line_position + d_line_height * 2));
	page->average_length_zx_label->setZValue(2);
	
	text_rect = page->average_length_zx_label->boundingRect();
	dval = page->average_length_zx_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->relative_scale_error_x_label = d_graphic_scene->addText(QStringLiteral("Relative Scale Error X: "),normal_font);
	page->relative_scale_error_x_label->setPos(QPointF(label_column_position,line_position + d_line_height * 3));
	page->relative_scale_error_x_label->setZValue(2);
	
	text_rect = page->relative_scale_error_x_label->boundingRect();
	dval = page->relative_scale_error_x_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->relative_scale_error_y_label = d_graphic_scene->addText(QStringLiteral("Relative Scale Error Y: "),normal_font);
	page->relative_scale_error_y_label->setPos(QPointF(label_column_position,line_position + d_line_height * 4));
	page->relative_scale_error_y_label->setZValue(2);
	
	text_rect = page->relative_scale_error_y_label->boundingRect();
	dval = page->relative_scale_error_y_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->relative_scale_error_z_label = d_graphic_scene->addText(QStringLiteral("Relative Scale Error Z: "),normal_font);
	page->relative_scale_error_z_label->setPos(QPointF(label_column_position,line_position + d_line_height * 5));
	page->relative_scale_error_z_label->setZValue(2);
	
	text_rect = page->relative_scale_error_z_label->boundingRect();
	dval = page->relative_scale_error_z_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	
	page->squareness_xy_label = d_graphic_scene->addText(QStringLiteral("Squareness Error XY: "),normal_font);
	page->squareness_xy_label->setPos(QPointF(label_column_position,line_position + d_line_height * 6));
	page->squareness_xy_label->setZValue(2);
	
	text_rect = page->squareness_xy_label->boundingRect();
	dval = page->squareness_xy_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->squareness_yz_label = d_graphic_scene->addText(QStringLiteral("Squareness Error YZ: "),normal_font);
	page->squareness_yz_label->setPos(QPointF(label_column_position,line_position + d_line_height * 7));
	page->squareness_yz_label->setZValue(2);
	
	text_rect = page->squareness_yz_label->boundingRect();
	dval = page->squareness_yz_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->squareness_zx_label = d_graphic_scene->addText(QStringLiteral("Squareness Error ZX: "),normal_font);
	page->squareness_zx_label->setPos(QPointF(label_column_position,line_position + d_line_height * 8));
	page->squareness_zx_label->setZValue(2);
	
	text_rect = page->squareness_zx_label->boundingRect();
	dval = page->squareness_zx_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	line_position += d_line_height * 10;

	// statistic result labels
	line_position += d_line_height;
	
	line_rect.setRect(page->column_left,line_position,line_width,1);
	page->statistics_results_rect = d_graphic_scene->addRect(line_rect,pen,brush);
	page->statistics_results_rect->setZValue(2);
	
	page->statistic_results_label = d_graphic_scene->addText(QStringLiteral("Data Statistics"),title_font);
	text_rect = page->statistic_results_label->boundingRect();
	page->statistic_results_label->setPos(QPointF(page->column_left,line_position - text_rect.height()));
	page->statistic_results_label->setZValue(2);

	line_position += d_line_height;
	statistics_results_position = line_position;

	page->average_length_label = d_graphic_scene->addText(QStringLiteral("Average Length: "),normal_font);
	page->average_length_label->setPos(QPointF(label_column_position,line_position));
	page->average_length_label->setZValue(2);
	
	text_rect = page->average_length_label->boundingRect();
	dval = page->average_length_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->minimum_length_label = d_graphic_scene->addText(QStringLiteral("Minimum Length: "),normal_font);
	page->minimum_length_label->setPos(QPointF(label_column_position,line_position + d_line_height));
	page->minimum_length_label->setZValue(2);
	
	text_rect = page->minimum_length_label->boundingRect();
	dval = page->minimum_length_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->maximum_length_label = d_graphic_scene->addText(QStringLiteral("Maximum Length: "),normal_font);
	page->maximum_length_label->setPos(QPointF(label_column_position,line_position + d_line_height * 2));
	page->maximum_length_label->setZValue(2);
	
	text_rect = page->maximum_length_label->boundingRect();
	dval = page->maximum_length_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->median_length_label = d_graphic_scene->addText(QStringLiteral("Median Length: "),normal_font);
	page->median_length_label->setPos(QPointF(label_column_position,line_position + d_line_height * 3));
	page->median_length_label->setZValue(2);
	
	text_rect = page->median_length_label->boundingRect();
	dval = page->median_length_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	page->stddev_length_label = d_graphic_scene->addText(QStringLiteral("Stddev Length: "),normal_font);
	page->stddev_length_label->setPos(QPointF(label_column_position,line_position + d_line_height * 4));
	page->stddev_length_label->setZValue(2);
	
	text_rect = page->stddev_length_label->boundingRect();
	dval = page->stddev_length_label->pos().x() + text_rect.right();
	if(dval > value_column_position)
	{
		value_column_position = dval;
	}
	
	value_column_position += (d_line_height * 2);
	
	// test result values
	line_position = test_results_position;
	
	page->test_date_time_value = d_graphic_scene->addText(QString(),normal_font);
	page->test_date_time_value->setPos(QPointF(value_column_position,line_position));
	page->test_date_time_value->setZValue(2);
	
	page->temperature_value = d_graphic_scene->addText(QString(),normal_font);
	page->temperature_value->setPos(QPointF(value_column_position,line_position + d_line_height));
	page->temperature_value->setZValue(2);
	
	page->tolerance_value = d_graphic_scene->addText(QString(),normal_font);
	page->tolerance_value->setPos(QPointF(value_column_position,line_position + d_line_height * 2));
	page->tolerance_value->setZValue(2);
	
	page->range_value = d_graphic_scene->addText(QString(),bold_font);
	page->range_value->setPos(QPointF(value_column_position,line_position + d_line_height * 3));
	page->range_value->setZValue(2);

	// test location values
	line_position = location_results_position;
	
	page->pivot_sphere_start_value = d_graphic_scene->addText(QString(),normal_font);
	page->pivot_sphere_start_value->setPos(QPointF(value_column_position,line_position));
	page->pivot_sphere_start_value->setZValue(2);
	
	page->pivot_sphere_end_value = d_graphic_scene->addText(QString(),normal_font);
	page->pivot_sphere_end_value->setPos(QPointF(value_column_position,line_position + d_line_height));
	page->pivot_sphere_end_value->setZValue(2);
	
	page->measured_sphere_value = d_graphic_scene->addText(QString(),normal_font);
	page->measured_sphere_value->setPos(QPointF(value_column_position,line_position + d_line_height * 2));
	page->measured_sphere_value->setZValue(2);
	
	// probe result value
	line_position = probe_results_position;
	
	page->probe_offset_value= d_graphic_scene->addText(QString(),normal_font);
	page->probe_offset_value->setPos(QPointF(value_column_position,line_position));
	page->probe_offset_value->setZValue(2);
	
	page->probe_error_value = d_graphic_scene->addText(QString(),normal_font);
	page->probe_error_value->setPos(QPointF(value_column_position,line_position + d_line_height));
	page->probe_error_value->setZValue(2);
	
	// analysis result values
	line_position = analysis_results_position;
	
	page->average_length_xy_value = d_graphic_scene->addText(QString(),normal_font);
	page->average_length_xy_value->setPos(QPointF(value_column_position,line_position));
	page->average_length_xy_value->setZValue(2);
	
	page->average_length_yz_value = d_graphic_scene->addText(QString(),normal_font);
	page->average_length_yz_value->setPos(QPointF(value_column_position,line_position + d_line_height));
	page->average_length_yz_value->setZValue(2);
	
	page->average_length_zx_value = d_graphic_scene->addText(QString(),normal_font);
	page->average_length_zx_value->setPos(QPointF(value_column_position,line_position + d_line_height * 2));
	page->average_length_zx_value->setZValue(2);
	
	page->relative_scale_error_x_value = d_graphic_scene->addText(QString(),normal_font);
	page->relative_scale_error_x_value->setPos(QPointF(value_column_position,line_position + d_line_height * 3));
	page->relative_scale_error_x_value->setZValue(2);
	
	page->relative_scale_error_y_value = d_graphic_scene->addText(QString(),normal_font);
	page->relative_scale_error_y_value->setPos(QPointF(value_column_position,line_position + d_line_height * 4));
	page->relative_scale_error_y_value->setZValue(2);
	
	page->relative_scale_error_z_value = d_graphic_scene->addText(QString(),normal_font);
	page->relative_scale_error_z_value->setPos(QPointF(value_column_position,line_position + d_line_height * 5));
	page->relative_scale_error_z_value->setZValue(2);
	
	page->squareness_xy_value = d_graphic_scene->addText(QString(),normal_font);
	page->squareness_xy_value->setPos(QPointF(value_column_position,line_position + d_line_height * 6));
	page->squareness_xy_value->setZValue(2);
	
	page->squareness_yz_value = d_graphic_scene->addText(QString(),normal_font);
	page->squareness_yz_value->setPos(QPointF(value_column_position,line_position + d_line_height * 7));
	page->squareness_yz_value->setZValue(2);
	
	page->squareness_zx_value = d_graphic_scene->addText(QString(),normal_font);
	page->squareness_zx_value->setPos(QPointF(value_column_position,line_position + d_line_height * 8));
	page->squareness_zx_value->setZValue(2);
	
	// statistic result values
	line_position = statistics_results_position;
	
	page->average_length_value = d_graphic_scene->addText(QString(),normal_font);
	page->average_length_value->setPos(QPointF(value_column_position,line_position));
	page->average_length_value->setZValue(2);
	
	page->minimum_length_value = d_graphic_scene->addText(QString(),normal_font);
	page->minimum_length_value->setPos(QPointF(value_column_position,line_position + d_line_height));
	page->minimum_length_value->setZValue(2);
	
	page->maximum_length_value = d_graphic_scene->addText(QString(),normal_font);
	page->maximum_length_value->setPos(QPointF(value_column_position,line_position + d_line_height * 2));
	page->maximum_length_value->setZValue(2);
	
	page->median_length_value = d_graphic_scene->addText(QString(),normal_font);
	page->median_length_value->setPos(QPointF(value_column_position,line_position + d_line_height * 3));
	page->median_length_value->setZValue(2);
	
	page->stddev_length_value = d_graphic_scene->addText(QString(),normal_font);
	page->stddev_length_value->setPos(QPointF(value_column_position,line_position + d_line_height * 4));
	page->stddev_length_value->setZValue(2);
}

void TReportWidget::Setup_Measurement_Page(
	TReportMeasurementPage 				* const page,
	const int							index_start,
	const int							index_end)
{
	QFont								title_font(d_font_family_name,d_font_size_title);
	double								line_position;
	int									cntr;
	int									data_count;
	int									column_positions[6];
	int									ival;
	TMeasurementData					measurement_data;
	QRect								line_rect;
	QRectF								image_rect;
	QRectF								text_rect;
	QPen								pen;
	QBrush								brush;

	pen.setStyle(Qt::SolidLine);
	pen.setWidth(0);
	pen.setColor(QColor(0,0,0));
	
	brush.setStyle(Qt::SolidPattern);
	brush.setColor(QColor(0,0,0));
	
	page->column_left = page->page_content_rect.left() + static_cast<int>(d_line_height);
	page->column_right = page->page_content_rect.right() - static_cast<int>(d_line_height);
	
	ival = page->column_right - page->column_left;
	ival /= 7;
	
	column_positions[0] = page->column_left + ival;
	column_positions[1] = column_positions[0] + ival;
	column_positions[2] = column_positions[1] + ival * 2;
	column_positions[3] = column_positions[2] + ival;
	column_positions[4] = column_positions[3] + ival;
	column_positions[5] = column_positions[4] + ival;
	
	line_position = page->page_content_rect.top() + d_line_height;
	
	page->elevation_label = d_graphic_scene->addText(tr("Elevation"),title_font);
	text_rect = page->elevation_label->boundingRect();
	page->elevation_label->setPos(QPointF(column_positions[0] - text_rect.width(),line_position));
	page->elevation_label->setZValue(2);

	page->azimuth_label = d_graphic_scene->addText(tr("Azimuth"),title_font);
	text_rect = page->azimuth_label->boundingRect();
	page->azimuth_label->setPos(QPointF(column_positions[1] - text_rect.width(),line_position));
	page->azimuth_label->setZValue(2);

	page->length_label = d_graphic_scene->addText(tr("Length "),title_font);
	text_rect = page->length_label->boundingRect();
	page->length_label->setPos(QPointF(column_positions[2] - text_rect.width(),line_position));
	page->length_label->setZValue(2);
	
	page->length_range_label = d_graphic_scene->addText(tr("Range"),title_font);
	text_rect = page->length_range_label->boundingRect();
	page->length_range_label->setPos(QPointF(column_positions[3] - text_rect.width(),line_position));
	page->length_range_label->setZValue(2);

	page->length_stddev_label = d_graphic_scene->addText(tr("Std. Dev"),title_font);
	text_rect = page->length_stddev_label->boundingRect();
	page->length_stddev_label->setPos(QPointF(column_positions[4] - text_rect.width(),line_position));
	page->length_stddev_label->setZValue(2);

	page->length_deviation_label = d_graphic_scene->addText(tr("Deviation"),title_font);
	text_rect = page->length_deviation_label->boundingRect();
	page->length_deviation_label->setPos(QPointF(column_positions[5] - text_rect.width(),line_position));
	page->length_deviation_label->setZValue(2);

	line_position += text_rect.height();
	
	line_rect.setRect(page->page_content_rect.left(),static_cast<int>(line_position),page->page_content_rect.width(),1);
	page->label_line_rect = d_graphic_scene->addRect(line_rect,pen,brush);
	page->label_line_rect->setZValue(1);
	
	line_position += d_line_height;
	

	if(index_end > static_cast<int>(d_measurement_data.size()))
	{
		data_count = static_cast<int>(d_measurement_data.size());
	}
	else
	{
		data_count = index_end;
	}
	
	for(cntr = index_start;cntr < data_count;++cntr)
	{
		measurement_data = d_measurement_data[cntr];
		
		text_rect = measurement_data.elevation_value->boundingRect();
		measurement_data.elevation_value->setPos(QPointF(column_positions[0] -  text_rect.width(),line_position));
		
		text_rect = measurement_data.angle_value->boundingRect();
		measurement_data.angle_value->setPos(QPointF(column_positions[1] - text_rect.width(),line_position));
		
		text_rect = measurement_data.length_value->boundingRect();
		measurement_data.length_value->setPos(QPointF(column_positions[2] - text_rect.width(),line_position));
		
		text_rect = measurement_data.length_range_value->boundingRect();
		measurement_data.length_range_value->setPos(QPointF(column_positions[3] - text_rect.width(),line_position));
		
		text_rect = measurement_data.length_stddev_value->boundingRect();
		measurement_data.length_stddev_value->setPos(QPointF(column_positions[4] - text_rect.width(),line_position));
		
		text_rect = measurement_data.length_dev_value->boundingRect();
		measurement_data.length_dev_value->setPos(QPointF(column_positions[5] - text_rect.width(),line_position));
		
		line_position += d_line_height;
	}
	return;
}

bool TReportWidget::Render_Page(
	QPainter							* const painter,
	const QRect 						&page_rect,
	const int 							page_number,
	bool								*const is_last_page)
{
	QRect								source_rect;
	QRectF								destination_rect;
	QRectF								scene_rect;
	double								source_aspect_ratio;
	double								destination_aspect_ratio;
	double								scaling_factor;
	double								size_difference;
	QTransform 							graphics_view_mat;
	int									margin;
	
	assert(is_last_page);
	assert(page_number > 0);
	assert((page_number - 2) < static_cast<int>(d_measurement_pages.size()));
	
	*is_last_page = true;
	
	if(page_number == 1)
	{
		source_rect = d_first_page.page_rect;
		d_first_page.page_background_rect->setVisible(false);
	}
	else
	{
		source_rect = d_measurement_pages[page_number - 2].page_rect;
		d_measurement_pages[page_number - 2].page_background_rect->setVisible(false);
	}
	
	margin = static_cast<int>(static_cast<double>(d_page_size.width()) * d_page_margin + 0.5);
	
	source_rect.adjust(-margin,-margin,margin,margin);
	source_rect.adjust(margin/4,margin/4,-margin/4,-margin/4);

	destination_rect = page_rect.adjusted(0,0,-page_rect.left() * 2,-page_rect.top() * 2);
	
	graphics_view_mat = d_graphic_view->transform();
	
	d_graphic_view->resetMatrix();
	d_graphic_view->scale(1.0, 1.0);
	
	source_rect.moveTo(d_graphic_view->mapFromScene(source_rect.topLeft()));
	
	source_aspect_ratio = static_cast<double>(source_rect.height()) /
	static_cast<double>(source_rect.width());
	
	destination_aspect_ratio = static_cast<double>(destination_rect.height()) /
	static_cast<double>(destination_rect.width());
	
	if(source_aspect_ratio > destination_aspect_ratio)
	{
		scaling_factor = destination_rect.height() / source_rect.height();
		
		size_difference = destination_rect.width();
		destination_rect.setWidth(source_rect.width() * scaling_factor);
		
		size_difference -= destination_rect.width();
		destination_rect.moveLeft(destination_rect.left() + size_difference / 2.0);
	}
	else
	{
		scaling_factor = destination_rect.width() / source_rect.width();
		
		size_difference = destination_rect.height();
		destination_rect.setHeight(source_rect.height() * scaling_factor);
		
		size_difference -= destination_rect.height();
		destination_rect.moveTop(destination_rect.top() + size_difference / 2.0);
	}
	
	d_graphic_view->setBackgroundBrush(QColor(255,255,255));
	d_graphic_view->render(painter,destination_rect,source_rect,Qt::IgnoreAspectRatio);
	d_graphic_view->setBackgroundBrush(d_graphic_view_brush);

	d_graphic_view->setTransform(graphics_view_mat);
	
	*is_last_page = (static_cast<unsigned int>(page_number - 1) == d_measurement_pages.size());
	
	if(page_number == 1)
	{
		d_first_page.page_background_rect->setVisible(true);
	}
	else
	{
		d_measurement_pages[page_number - 2].page_background_rect->setVisible(true);
	}
	
	return true;
}
