/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QVariant>
#include <QAction>
#include <QApplication>
#include <QButtonGroup>
#include <QFrame>
#include <QGridLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QHeaderView>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QSpacerItem>
#include <QTextEdit>
#include <QVBoxLayout>
#include <QCloseEvent>
#include <QRegularExpression>
#include <QRegularExpressionMatch>
#include <cmath>

#include "tolerancewidget.h"

TToleranceWidget::TToleranceWidget(
	const QWidget 						*Parent,
	Qt::WindowFlags 					Flags)
:QWidget(const_cast<QWidget*>(Parent),Flags)
{
	QGridLayout 						*widget_layout;
	QLabel 								*sample1_graph_label;
	QLabel 								*sample1_text_label;
	QLabel 								*sample2_graph_label;
	QLabel 								*sample2_text_label;
	QLabel 								*sample3_graph_label;
	QLabel 								*sample3_text_label;
	QHBoxLayout 						*tolerance_hlayout;
	QLabel 								*tolerance_label;
	QSpacerItem							*widget_vspacer;

	widget_layout = new QGridLayout(this);
	widget_layout->setSpacing(6);

	sample1_graph_label = new QLabel(this);
	sample1_graph_label->setMinimumSize(QSize(48, 48));
	sample1_graph_label->setMaximumSize(QSize(48, 48));
	sample1_graph_label->setPixmap(QPixmap(QString::fromUtf8(":/images/tolerance1.png")));
	sample1_graph_label->setScaledContents(false);
	widget_layout->addWidget(sample1_graph_label, 0, 0, 1, 1);

	sample1_text_label = new QLabel(this);
	widget_layout->addWidget(sample1_text_label, 0, 1, 1, 1);

	sample2_graph_label = new QLabel(this);
	sample2_graph_label->setMinimumSize(QSize(48, 48));
	sample2_graph_label->setMaximumSize(QSize(48, 48));
	sample2_graph_label->setPixmap(QPixmap(QString::fromUtf8(":/images/tolerance2.png")));
	widget_layout->addWidget(sample2_graph_label, 1, 0, 1, 1);

	sample2_text_label = new QLabel(this);
	widget_layout->addWidget(sample2_text_label, 1, 1, 1, 1);

	sample3_graph_label = new QLabel(this);
	sample3_graph_label->setMinimumSize(QSize(48, 48));
	sample3_graph_label->setMaximumSize(QSize(48, 48));
	sample3_graph_label->setPixmap(QPixmap(QString::fromUtf8(":/images/tolerance3.png")));
	widget_layout->addWidget(sample3_graph_label, 2, 0, 1, 1);

	sample3_text_label = new QLabel(this);
	widget_layout->addWidget(sample3_text_label, 2, 1, 1, 1);

	tolerance_hlayout = new QHBoxLayout();

	tolerance_label = new QLabel(this);
	tolerance_label->setAlignment(Qt::AlignRight | Qt::AlignTrailing | Qt::AlignVCenter);
	tolerance_hlayout->addWidget(tolerance_label);

	d_tolerance_edit = new QLineEdit(this);
	d_tolerance_edit->setAlignment(Qt::AlignCenter);
	tolerance_hlayout->addWidget(d_tolerance_edit);

	widget_layout->addLayout(tolerance_hlayout, 3, 0, 1, 2);
	
	widget_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	widget_layout->addItem(widget_vspacer, 4, 0, 1, 2);

	sample1_text_label->setText(QStringLiteral("E Mpl = A"));
	sample2_text_label->setText(QStringLiteral("E Mpl = A + BL/C"));
	sample3_text_label->setText(QStringLiteral("E Mpl = A + BL/C | D"));

	tolerance_label->setText(QStringLiteral("E Mpl ="));

	// Set Defaults
	d_tolerance_type = TToleranceWidget::TOLERANCE_UNKNOWN;
	d_tolerance_edit->setText("0.003 + 0.004L/1000 | 0.005");

	this->Tolerance_Changed();

	connect(d_tolerance_edit,&QLineEdit::textEdited,this,&TToleranceWidget::Tolerance_Changed);
}

TToleranceWidget::~TToleranceWidget(void)
{

}

QString TToleranceWidget::Tolerance_Expression(void) const
{
	return d_tolerance_edit->text();
}

double TToleranceWidget::Tolerance_Value(
	const double						&length) const
{
	double								tolerance(0.0);
	
	switch(d_tolerance_type)
	{
		case TToleranceWidget::TOLERANCE_UNKNOWN:
			break;
			
		case TToleranceWidget::TOLERANCE_BAND:
			tolerance = d_tolerance_a;
			break;
			
		case TToleranceWidget::TOLERANCE_FORMULA:
			tolerance = d_tolerance_a;
			
			if(d_tolerance_c > 0.1)
			{
				tolerance += d_tolerance_b * length / d_tolerance_c;
			}
			break;
			
		case TToleranceWidget::TOLERANCE_FORMULA_WITH_LIMIT:
			tolerance = d_tolerance_a;
			
			if(d_tolerance_c > 0.1)
			{
				tolerance += d_tolerance_b * length / d_tolerance_c;
			}
			
			if(tolerance > d_tolerance_d)
			{
				tolerance = d_tolerance_d;
			}
			break;
	}
		
	return tolerance;
}

double TToleranceWidget::Tolerance_Item_D_Position(
	const double 						&tol_a,
	const double 						&tol_b,
	const double 						&tol_c,
	const double 						&tol_d) const
{
	double								slope;
	double								pos(0);
	static const double					ZERO_EPSILON(0.0001);

	if(tol_d < tol_a)
	{
		return pos;
	}
	
	if(fabs(tol_c) < ZERO_EPSILON)
	{
		return pos;
	}

	slope = tol_b / tol_c;	// mm / m

	pos = (tol_d - tol_a) / slope;

	return pos;
}


void TToleranceWidget::Set_Tolerance_Expression(
	const QString						&s)
{
	bool								previous_state;
	
	if(s.length())
	{
		previous_state = d_tolerance_edit->blockSignals(true);
		d_tolerance_edit->setText(s);
		d_tolerance_edit->blockSignals(previous_state);
		
		this->Tolerance_Changed();
	}
}

void TToleranceWidget::Set_Tolerance_Item_A(
	const double						&value)
{
	d_tolerance_a = value;
}

void TToleranceWidget::Set_Tolerance_Item_B(
	const double						&value)
{
	d_tolerance_b = value;
}

void TToleranceWidget::Set_Tolerance_Item_C(
	const double						&value)
{
	d_tolerance_c = value;
}

void TToleranceWidget::Set_Tolerance_Item_D(
	const double						&value)
{
	d_tolerance_d = value;
}

void TToleranceWidget::Update(void)
{
	static const double					ZERO_EPSILON(0.0001);

	// determine type of tolerance to use based on current tolerance values
	if(d_tolerance_d > 0 && d_tolerance_c > 0.1)
	{
		d_tolerance_type = TToleranceWidget::TOLERANCE_FORMULA_WITH_LIMIT;
	}
	else if(d_tolerance_c > 0.1)
	{
		d_tolerance_type = TToleranceWidget::TOLERANCE_FORMULA;
	}
	else if(d_tolerance_a > ZERO_EPSILON)
	{
		d_tolerance_type = TToleranceWidget::TOLERANCE_BAND;
	}
	else
	{
		d_tolerance_type = TToleranceWidget::TOLERANCE_UNKNOWN;
	}

	// write current values to display.  May not be valid
	this->Update_Display();
	
	// analyze what is written.  may change tolerance type
	this->Update_Tolerance();
}

void TToleranceWidget::Tolerance_Changed(void)
{
	QRegularExpression					value("^([0-9.]{1,8})$");
	QRegularExpression					expression("^([0-9.]{1,8})[+]([0-9.]{1,8})[L][/]([0-9]{1,5})$");
	QRegularExpression					expression_with_limit("^([0-9.]{1,8})[+]([0-9.]{1,8})[L][/]([0-9]{1,5})[|]([0-9.]{1,8})$");
	QRegularExpressionMatch				expression_match;
	QStringList							expression_text_list;
	QString								text;
	
	d_tolerance_type = TToleranceWidget::TOLERANCE_UNKNOWN;

	text = d_tolerance_edit->text().simplified();
	text = text.remove(' ');
	
	if(!text.length())
	{
		d_tolerance_edit->setStyleSheet(QStringLiteral("background-color: rgb(255, 151, 160);"));
		return;
	}
	else
	{
		d_tolerance_edit->setStyleSheet(QString());
	}

	text = text.toUpper();
	
	expression_match = expression_with_limit.match(text);
	if(expression_match.hasMatch())
	{		
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 5)
		{			
			d_tolerance_type = TOLERANCE_FORMULA_WITH_LIMIT;
			
			d_tolerance_a = expression_text_list[1].toDouble();
			d_tolerance_b = expression_text_list[2].toDouble();
			d_tolerance_c = expression_text_list[3].toDouble();
			d_tolerance_d = expression_text_list[4].toDouble();
			
			if(d_tolerance_a > 1 || d_tolerance_b > 1 || d_tolerance_d > 1)
			{
				d_tolerance_edit->setStyleSheet(QStringLiteral("background-color: rgb(255, 151, 160);"));
			}
			else
			{
				emit Update_Tolerance();
			}
			
			return;
		}
	}
	
	expression_match = expression.match(text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 4)
		{			
			d_tolerance_type = TOLERANCE_FORMULA;
			
			d_tolerance_a = expression_text_list[1].toDouble();
			d_tolerance_b = expression_text_list[2].toDouble();
			d_tolerance_c = expression_text_list[3].toDouble();
			d_tolerance_d = 0;
			
			if(d_tolerance_a > 1 || d_tolerance_b > 1)
			{
				d_tolerance_edit->setStyleSheet(QStringLiteral("background-color: rgb(255, 151, 160);"));
			}
			else
			{
				emit Update_Tolerance();
			}
			
			return;
		}
	}
	
	expression_match = value.match(text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 2)
		{						
			d_tolerance_type = TOLERANCE_BAND;
			
			d_tolerance_a = expression_text_list[1].toDouble();
			d_tolerance_b = 0;
			d_tolerance_c = 0;
			d_tolerance_d = 0;
			
			if(d_tolerance_a > 1)
			{
				d_tolerance_edit->setStyleSheet(QStringLiteral("background-color: rgb(255, 151, 160);"));
			}
			else
			{
				emit Update_Tolerance();
			}
			
			return;
		}
	}

	d_tolerance_edit->setStyleSheet(QStringLiteral("background-color: rgb(255, 151, 160);"));
}

void TToleranceWidget::Update_Display(void)
{
	bool								prev_state;
	
	prev_state = d_tolerance_edit->blockSignals(true);

	switch(d_tolerance_type)
	{
		case TToleranceWidget::TOLERANCE_UNKNOWN:
			d_tolerance_edit->setText(QString());
			break;
			
		case TToleranceWidget::TOLERANCE_BAND:
			d_tolerance_edit->setText(QString("%1").arg(d_tolerance_a,0,'f',4));
			break;
			
		case TToleranceWidget::TOLERANCE_FORMULA:
			d_tolerance_edit->setText(QString("%1+%2L/%3")
									.arg(d_tolerance_a,0,'f',4)
									.arg(d_tolerance_b,0,'f',4)
									.arg(d_tolerance_c,0,'f',0));
			break;
			
		case TToleranceWidget::TOLERANCE_FORMULA_WITH_LIMIT:
			d_tolerance_edit->setText(QString("%1+%2L/%3 | %4")
									.arg(d_tolerance_a,0,'f',4)
									.arg(d_tolerance_b,0,'f',4)
									.arg(d_tolerance_c,0,'f',0)
									.arg(d_tolerance_d,0,'f',4));
			break;
	}
	
	d_tolerance_edit->blockSignals(prev_state);
}

